

#include <windows.h>
#include <d3d11.h>
#include <d3dx11.h>
#include <d3dcompiler.h>
#include <D3DX11.h>
#include <D3DX10math.h>

#define _USE_MATH_DEFINES
#include <math.h>

#include "Utility.h"

#include "BlitDepth.h"

struct ConstantBuffer
{
    D3DXVECTOR4		m_vScaleOffSrc;
    D3DXVECTOR4		m_vScaleOffDst;

    D3DXVECTOR4     m_RadiusSq;
    D3DXVECTOR4     m_Center;
};



void BlitDepth::SetConstants( ID3D11DeviceContext* immediateContext, ID3D11ShaderResourceView* baseTex )
{
    ConstantBuffer buff;
    buff.m_vScaleOffDst = D3DXVECTOR4(
        (m_DstRect.hi.x - m_DstRect.lo.x) * 2.0f,
        (m_DstRect.hi.y - m_DstRect.lo.y) * 2.0f,
        m_DstRect.lo.x * 2.0f - 1.0f,
        m_DstRect.lo.y * 2.0f - 1.0f);

    buff.m_vScaleOffSrc = D3DXVECTOR4(
        m_SrcRect.hi.x - m_SrcRect.lo.x,
        m_SrcRect.hi.y - m_SrcRect.lo.y,
        m_SrcRect.lo.x,
        m_SrcRect.lo.y);


    buff.m_RadiusSq = D3DXVECTOR4(m_fRadiusSq, 0.0f, 0.0f, 0.0f);
    buff.m_Center = m_Center;

    immediateContext->UpdateSubresource( m_pConstantBuffer, 0, NULL, &buff, 0, 0 );
    immediateContext->VSSetConstantBuffers( 0, 1, &m_pConstantBuffer );
    immediateContext->PSSetConstantBuffers( 0, 1, &m_pConstantBuffer );
}

HRESULT BlitDepth::InitConstantBuffers( ID3D11Device* pd3dDevice )
{
    HRESULT hr = S_OK;

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(ConstantBuffer);
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;
    hr = pd3dDevice->CreateBuffer( &bd, NULL, &m_pConstantBuffer );
    if( FAILED( hr ) )
        return hr;

    return hr;

}

HRESULT BlitDepth::InitState(ID3D11Device* pd3dDevice)
{
    D3D11_RASTERIZER_DESC drd = {
        D3D11_FILL_SOLID, //D3D11_FILL_MODE FillMode;
        D3D11_CULL_NONE,//D3D11_CULL_MODE CullMode;
        FALSE, //BOOL FrontCounterClockwise;
        0, //INT DepthBias;
        0.0f,//FLOAT DepthBiasClamp;
        0.0f,//FLOAT SlopeScaledDepthBias;
        FALSE,//BOOL DepthClipEnable;
        FALSE,//BOOL ScissorEnable;
        FALSE,//BOOL MultisampleEnable;
        FALSE//BOOL AntialiasedLineEnable;        
    };
    HRESULT hr = pd3dDevice->CreateRasterizerState(&drd, &m_pRasterizerState);
    if( FAILED( hr ) )
        return hr;

    // Depth stencil state
    D3D11_DEPTH_STENCIL_DESC DSDesc;
    ZeroMemory( &DSDesc, sizeof( D3D11_DEPTH_STENCIL_DESC ) );
    DSDesc.DepthEnable = TRUE;
    DSDesc.DepthWriteMask = D3D11_DEPTH_WRITE_MASK_ALL;
    DSDesc.DepthFunc = D3D11_COMPARISON_ALWAYS;
    DSDesc.StencilEnable = FALSE;
    hr = pd3dDevice->CreateDepthStencilState( &DSDesc, &m_pDepthStencilState );
    if( FAILED(hr) )
        return hr;

    hr = InitBlendState(pd3dDevice);
    if(FAILED(hr))
        return hr;

    return hr;
}
