
#ifndef BLIT_QUAD_H
#define BLIT_QUAD_H

#include <vector>

class BlitQuad
{
public:
	struct BlitRect
	{
	public:
		D3DXVECTOR2 lo;
		D3DXVECTOR2 hi;

		float& MinX() { return lo.x; }
		float& MaxX() { return hi.x; }
		float& MinY() { return lo.y; }
		float& MaxY() { return hi.y; }

		BlitRect(float loX, float loY, float hiX, float hiY)
			:	lo(loX, loY),
				hi(hiX, hiY)
		{
		}

		BlitRect(const D3DXVECTOR2& lo, const D3DXVECTOR2& hi)
			:	lo(lo), 
				hi(hi)
		{
		}

		BlitRect()
			: lo(0.0f, 0.0f),
			  hi(0.0f, 0.0f)
		{
		}
	};

	BlitQuad();

	// If _any_ of these parameters are null, it will use the default as here for _that_ parm.
	HRESULT Init(ID3D11Device* pd3dDevice,
                    const WCHAR* shaderFile = L"ShaderBlits/BlitQuad.fx", 
					const CHAR* entry = "Def", 
					const CHAR* shaderModel = "4_0");

	virtual ~BlitQuad()
	{
		Release();
	}

	// Must re-Init() after calling this.
	void Release();

	const BlitRect& GetSrcRect() const { return m_SrcRect; }
	void SetSrcRect(const BlitRect& rect) { m_SrcRect = rect; }

	const BlitRect& GetDstRect() const { return m_DstRect; }
	void SetDstRect(const BlitRect& rect) { m_DstRect = rect; }

	const POINT& GetDstResolution() const { return m_DstResolution; }
	void SetDstResolution(const POINT& res) { m_DstResolution = res; }

    const D3DXVECTOR4& GetTint() const { return m_Tint; }
    void SetTint(const D3DXVECTOR4& tint) { m_Tint = tint; }

	virtual void Render(ID3D11DeviceContext* immediateContext, ID3D11RenderTargetView* rt, ID3D11ShaderResourceView* texSRV);

    bool GetPointSample() const { return m_bPointSample; }
    void SetPointSample(ID3D11Device* pd3dDevice, bool b);

    enum Blend
    {
        kNone,
        kSrcAlpha,
        kInvSrcAlpha,
        kSrcAlphaPM,
        kAdd
    };
    const Blend& GetBlend() const { return m_Blend; }
    void SetBlend(const Blend& blend) { m_Blend = blend; }

protected:
	
    virtual HRESULT InitState(ID3D11Device* pd3dDevice);
    virtual HRESULT InitSampler(ID3D11Device* pd3dDevice);
	virtual HRESULT InitConstantBuffers(ID3D11Device* pd3dDevice);
    virtual HRESULT InitBlendState(ID3D11Device* pd3dDevice);

	virtual void SetState(ID3D11RenderTargetView* rt, ID3D11DeviceContext* immediateContext);
	virtual void SetConstants(ID3D11DeviceContext* immediateContext, ID3D11ShaderResourceView* baseTex);
	
    virtual void UnSetConstants(ID3D11DeviceContext* immediateContext);

	virtual HRESULT LoadShaders(ID3D11Device* pd3dDevice, const WCHAR* shaderFile, const CHAR* entry, const CHAR* sm);
	virtual HRESULT LoadVertexShader(ID3D11Device* pd3dDevice, const WCHAR* shaderFile, const CHAR* entry, const CHAR* sm);
	virtual HRESULT LoadPixelShader(ID3D11Device* pd3dDevice, const WCHAR* shaderFile, const CHAR* entry, const CHAR* sm);

    virtual void AppendVertexDefines(std::vector<D3D_SHADER_MACRO>& /*defines*/) { }
    virtual void AppendPixelDefines(std::vector<D3D_SHADER_MACRO>& /*defines*/) { }

    void MakeDefines(D3D_SHADER_MACRO* defines, bool vtx);

    D3DXVECTOR4                 m_Tint;

    BlitRect					m_DstRect;
    BlitRect					m_SrcRect;

    POINT						m_DstResolution;

    ID3D11VertexShader*			m_pVertexShader;
    ID3D11PixelShader*			m_pPixelShader;

    ID3D11RasterizerState*      m_pRasterizerState;
    ID3D11DepthStencilState*	m_pDepthStencilState;
    ID3D11BlendState*           m_pBlendState;

    ID3D11SamplerState*         m_pTexSampler;

    ID3D11Buffer*				m_pConstantBuffer;

    bool                        m_bPointSample;

    Blend                       m_Blend;

private:
};

#endif // BLIT_QUAD_H
