
#ifndef CASCADE_H
#define CASCADE_H

class OffscreenRT;

class Cascade
{
public:
    static void Init();

	Cascade();
	~Cascade();

	void Release();

	HRESULT Create(ID3D11Device* pd3dDevice, int width, int height, int msaa, float dup);
    HRESULT CreateBase(ID3D11Device* pd3dDevice, int width, int height, int msaa);

    // Must be called before Push to do any good.
    const D3DXVECTOR2& GetJitter() const { return m_Jitter; }
    void SetJitter(const D3DXVECTOR2& jit) { m_Jitter = jit; }

	D3DXMATRIX Push(ID3D11DeviceContext* pd3dImmediateContext, const POINT& gazePos, 
        const D3DXMATRIX& origCamToNDC, const FLOAT nearClip, const FLOAT farClip);

	void Clear(ID3D11DeviceContext* pd3dImmediateContext, const float* const color);

	void Pop(ID3D11DeviceContext* pd3dImmediateContext);

    const D3DXMATRIX& GetLastWorldToNDC() const { return m_mLastWorldToNDC; }
    void SetLastWorldToNDC(const D3DXMATRIX& w2ndc) { m_mLastWorldToNDC = w2ndc; }

    int GetWidth() const;
    int GetHeight() const;

	ID3D11ShaderResourceView* GetTexture() const;
	ID3D11ShaderResourceView* GetPreviousTex() const;
	ID3D11ShaderResourceView* GetPreviousDS() const;
    const POINT& GetGazePos() const { return m_LastGazePos; }

	float GetDuplication() const { return m_fDuplication; }

    void PingPong();

    HRESULT Save(const WCHAR* fileName, ID3D11DeviceContext* pd3dImmediateContext);

    static bool GetJitterEnabled() { return m_bJitterEnabled; }
    static void SetJitterEnabled(bool b) { m_bJitterEnabled = b; }

    float GetJitterScale() const { return m_fJitterScale; }
    void SetJitterScale(float f) { m_fJitterScale = f; }

    bool ShouldRefresh(float dt);

    void SetRefreshInterval(float dt) { m_fRefreshInterval = dt; }
    float GetRefreshInterval() const { return m_fRefreshInterval; }

private:

	OffscreenRT*	m_aOffscreenRT[2];

	float           m_fDuplication;
    int             m_iPing;

    D3DXVECTOR2     m_Jitter;
    POINT           m_LastGazePos;

    float           m_fRefreshCountDown;
    float           m_fRefreshInterval;
private:

    D3DXMATRIX m_mLastWorldToNDC;

private:
    // Jitter section
    static bool m_bJitterEnabled;

    float m_fJitterScale;
    int m_iJitterIndex;

    D3DXVECTOR2 NextJitterSample();
};

#endif // CASCADE_H
