//--------------------------------------------------------------------------------------
// File: SymbolicDiff.cpp
//
// Starting point for new Direct3D 11 samples.  For a more basic starting point, 
// use the EmptyProject11 sample instead.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------
#include "DXUT.h"
#include "DXUTgui.h"
#include "DXUTmisc.h"
#include "DXUTCamera.h"
#include "DXUTSettingsDlg.h"
#include "SDKmisc.h"
#include "SDKmesh.h"
#include <vector>
#include <map>
#include <algorithm>
#define _USE_MATH_DEFINES
#include <math.h>
#include "resource.h"

//#define DEBUG_VS   // Uncomment this line to debug D3D9 vertex shaders 
//#define DEBUG_PS   // Uncomment this line to debug D3D9 pixel shaders 
//#define MAF_NO_HUD

//--------------------------------------------------------------------------------------
/// Object stuff
#include "Objects/ParametricSurface.h"
#include "Objects/PatchSurface.h"
#include "Objects/LODObject.h"
#include "Utility/CFlyCamera.h"
#include "Utility/WaitDlg.h"

#include "FauxV8Kit/BlitComposite.h"
#include "FauxV8Kit/BlitQuad.h"
#include "FauxV8Kit/OffscreenRT.h"
#include "FauxV8Kit/Cascade.h"
#include "FauxV8Kit/FauxV8Mgr.h"
#include "FauxV8Kit/TunableFV8Mgr.h"
#include "Utility/CullVolume.h"
#include "Utility/BlitEnvMap.h"
#include "Utility/CameraModeSelect.h"
#include "Utility/CameraSave.h"
#include "Utility/CameraPan.h"

#include "Objects/Sphere.h"
#include "Objects/DrawBox.h"
#include "Objects/SimpleScape.h"
#include "Objects/ProcScape.h"

#include "SettingBlock.h"

#pragma warning(disable : 4482)

UINT _hackSkipMask = 0;

typedef std::vector<Object*>		ObjectList;
typedef ObjectList::iterator		ObjectIter;
typedef ObjectList::const_iterator	ConstObjectIter;

ObjectList objects;
ObjectList objectCache;
Scape* _scape = NULL;


float spinSpeed = 1.5f;
float spinBoost = 1.0f;
float upDownSpeed = 0.1f;

static const float kYon = 10000.0f;
static const float kHither = 0.1f;

D3DXVECTOR3 rotRate(0.0f, 0.0f, 0.0f);
D3DXVECTOR3 rotAccum(float(-M_PI_2), 0.0f, 0.0f);
float upDown = 0;

/// Array of shader files defining models. Models loaded to 
/// fill out the grid will be randomly drawn from this list.
/// Duplicated shaders have larger chance of being selected.
/// When only one model is loaded at a time (modelsX == modelsY == 1),
/// models will be loaded sequentially from this list. Add models
/// you create to this list to have them appear.
static const WCHAR* const modelShaders[] = {
    L"Shaders/rook_b.3dmodel.fx",
    L"Shaders/rook_b.3dmodel.fx",
    L"Shaders/knight_b.3dmodel.fx",
    L"Shaders/knight_b.3dmodel.fx",
    L"Shaders/bishop_b.3dmodel.fx",
    L"Shaders/bishop_b.3dmodel.fx",
    L"Shaders/king_b.3dmodel.fx",
    L"Shaders/queen_b.3dmodel.fx",
    L"Shaders/pawn_b.3dmodel.fx",
    L"Shaders/pawn_b.3dmodel.fx",
    L"Shaders/pawn_b.3dmodel.fx",
    L"Shaders/pawn_b.3dmodel.fx",
    L"Shaders/rook_w.3dmodel.fx",
    L"Shaders/rook_w.3dmodel.fx",
    L"Shaders/knight_w.3dmodel.fx",
    L"Shaders/knight_w.3dmodel.fx",
    L"Shaders/bishop_w.3dmodel.fx",
    L"Shaders/bishop_w.3dmodel.fx",
    L"Shaders/king_w.3dmodel.fx",
    L"Shaders/queen_w.3dmodel.fx",
    L"Shaders/pawn_w.3dmodel.fx",
    L"Shaders/pawn_w.3dmodel.fx",
    L"Shaders/pawn_w.3dmodel.fx",
    L"Shaders/pawn_w.3dmodel.fx",

};
static int modelIdx = 0;
//--------------------------------------------------------------------------------------


//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
CFlyCamera				g_Camera;               // A model viewing camera
CDXUTDialogResourceManager  g_DialogResourceManager; // manager for shared resources of dialogs
CD3DSettingsDlg             g_SettingsDlg;          // Device settings dialog
CDXUTTextHelper*            g_pTxtHelper = NULL;
//CDXUTDialog                 g_HUD;                  // dialog for standard controls
CDXUTDialog                 g_SampleUI;             // dialog for sample specific controls


static POINT _screenSize = { 0, 0 };
POINT _gazePos;
POINT GetGazePosition() { return _gazePos; }

//--------------------------------------------------------------------------------------
// Constant buffers
//--------------------------------------------------------------------------------------
#pragma pack(push,1)
struct CB_VS_PER_OBJECT
{
    D3DXMATRIX  m_mWorldViewProjection;
    D3DXMATRIX  m_mWorld;
    D3DXVECTOR4 m_MaterialAmbientColor;
    D3DXVECTOR4 m_MaterialDiffuseColor;
};

struct CB_VS_PER_FRAME
{
    D3DXVECTOR3 m_vLightDir;
    float       m_fTime;
    D3DXVECTOR4 m_LightDiffuse;
};
#pragma pack(pop)

ID3D11Buffer*                       g_pcbVSPerObject11 = NULL;
ID3D11Buffer*                       g_pcbVSPerFrame11 = NULL;

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------
#define IDC_TOGGLEFULLSCREEN		1
#define IDC_TOGGLEREF				2
#define IDC_CHANGEDEVICE			3

#define IDC_STATIC_
#define IDC_SLIDER_
#define IDC_CHECKBOX_

#define IDC_STATIC_INNER_RES        30 // adjust
#define IDC_SLIDER_INNER_RES        31 // adjust

#define IDC_CHECKBOX_USE_MID      100
#define IDC_STATIC_MID_RES        32 // adjust
#define IDC_SLIDER_MID_RES        33 // adjust
#define IDC_STATIC_MID_DUP        34 // adjust
#define IDC_SLIDER_MID_DUP        35 // adjust

#define IDC_STATIC_BASE_RES         36 // adjust
#define IDC_SLIDER_BASE_RES         37 // adjust

#define IDC_STATIC_COMP_FALLOFF     38 // adjust
#define IDC_SLIDER_COMP_FALLOFF     39 // adjust

#define IDC_SLIDER_HIGH_TRIS_IN        48  // adjust
#define IDC_SLIDER_HIGH_DIST_IN        49  // adjust
#define IDC_SLIDER_LOW_TRIS_IN         50  // adjust
#define IDC_SLIDER_LOW_DIST_IN         51  // adjust

#define IDC_SLIDER_HIGH_TRIS_UP        52  // adjust
#define IDC_SLIDER_HIGH_DIST_UP        53  // adjust
#define IDC_SLIDER_LOW_TRIS_UP         54  // adjust
#define IDC_SLIDER_LOW_DIST_UP         55  // adjust

#define IDC_SLIDER_HIGH_TRIS_BA        56  // adjust
#define IDC_SLIDER_HIGH_DIST_BA        57  // adjust
#define IDC_SLIDER_LOW_TRIS_BA         58  // adjust
#define IDC_SLIDER_LOW_DIST_BA         59  // adjust

#define IDC_STATIC_HIGH_TRIS_IN        348  // adjust
#define IDC_STATIC_HIGH_DIST_IN        349  // adjust
#define IDC_STATIC_LOW_TRIS_IN         350  // adjust
#define IDC_STATIC_LOW_DIST_IN         351  // adjust

#define IDC_STATIC_HIGH_TRIS_UP        352  // adjust
#define IDC_STATIC_HIGH_DIST_UP        353  // adjust
#define IDC_STATIC_LOW_TRIS_UP         354  // adjust
#define IDC_STATIC_LOW_DIST_UP         355  // adjust

#define IDC_STATIC_HIGH_TRIS_BA        356  // adjust
#define IDC_STATIC_HIGH_DIST_BA        357  // adjust
#define IDC_STATIC_LOW_TRIS_BA         358  // adjust
#define IDC_STATIC_LOW_DIST_BA         359  // adjust


#define IDC_STATIC_MSAA             60  // adjust
#define IDC_SLIDER_MSAA             61  // adjust

#define IDC_STATIC_MODELSX          62  // adjust
#define IDC_SLIDER_MODELSX          63  // adjust
#define IDC_STATIC_MODEL_SPACING          66  // adjust
#define IDC_SLIDER_MODEL_SPACING          67  // adjust
#define IDC_STATIC_DEPTH_TOL        68
#define IDC_SLIDER_DEPTH_TOL        69
#define IDC_STATIC_DRAW_FRAMES      70
#define IDC_SLIDER_DISCARD_FRAMES      71
#define IDC_STATIC_FRAME_DECAY_CEN         72
#define IDC_SLIDER_FRAME_DECAY_CEN         73
#define IDC_STATIC_JITTER_SCALE_CEN     74
#define IDC_SLIDER_JITTER_SCALE_CEN     75
#define IDC_STATIC_FRAME_DECAY_MID         76
#define IDC_SLIDER_FRAME_DECAY_MID         77
#define IDC_STATIC_JITTER_SCALE_MID     78
#define IDC_SLIDER_JITTER_SCALE_MID     79
#define IDC_STATIC_FRAME_DECAY_OUT         80
#define IDC_SLIDER_FRAME_DECAY_OUT         81
#define IDC_STATIC_JITTER_SCALE_OUT     82
#define IDC_SLIDER_JITTER_SCALE_OUT     83
#define IDC_STATIC_SKIP_MID             84
#define IDC_SLIDER_SKIP_MID             85
#define IDC_STATIC_SKIP_OUT             86
#define IDC_SLIDER_SKIP_OUT             87
#define IDC_STATIC_SKIP_SECS_MID             88
#define IDC_SLIDER_SKIP_SECS_MID             89
#define IDC_STATIC_SKIP_SECS_OUT             90
#define IDC_SLIDER_SKIP_SECS_OUT             91
#define IDC_STATIC_LATENCY					92
#define IDC_SLIDER_LATENCY					93

#define IDC_CHECKBOX_USE_LODS       101 // adjust
#define IDC_CHECKBOX_SHOW_LEVELS    102 // adjust
#define IDC_CHECKBOX_LIGHTS         104 // adjust
#define IDC_CHECKBOX_SPECULAR       105 // adjust
#define IDC_CHECKBOX_ENVIRONMENT    106 // adjust
#define IDC_CHECKBOX_BACKGROUND     107 // adjust
#define IDC_CHECKBOX_MSBASE         109 // adjust
#define IDC_CHECKBOX_MSMID          110 // adjust
#define IDC_CHECKBOX_MSINNER        111 // adjust
#define IDC_CHECKBOX_MSAAINNER      112 // adjust
#define IDC_CHECKBOX_WIREFRAME      113 // adjust
#define IDC_CHECKBOX_SHOWBOUNDS     114 // adjust

#define IDC_STATIC_COMPOSITE        200 //
#define IDC_COMBOBOX_COMPOSITE      201 //

#define IDC_STATIC_CAMERA_MODE      202
#define IDC_COMBOBOX_CAMERA_MODE    203

#define MAX_TESSELLATION_FACTOR	(64.0f)

static bool pauseTexture = true;
static bool useLOD = false;
static bool textureBackground = true;
static bool zeroLatency = true;
POINT numModels = { 20, 20 };
float modelSpacing = 3.1f;
static bool _modelRenderEnabled = true;
static bool displayStats = true;
static bool procTerrain = true;
static int latencyMS = 0;
static bool smoothTime = false;

static bool guiActive = false;
static bool experimentActive = false;

static bool _logServer = false;
static bool _maximizeWindow = false;

bool DisableCameraKeyboard()
{
    return guiActive || experimentActive;
}

enum GazeMode
{
    kGaze,
    kMouse,
    kLocked,

    kGazeCycle,

};
static GazeMode gazeMode = kMouse;
static const WCHAR* GazeModeString()
{
    switch(gazeMode)
    {
    case kGaze:
        return L"Gaze";
    case kMouse:
        return L"Mouse";
    case kLocked:
        return L"Locked";
    default:
        break;
    }
    assert(false);
    return NULL;
}
GazeMode operator++(GazeMode& rhs)
{
    int gz = int(rhs) + 1;
    if(gz >= kGazeCycle)
        gz = 0;
    return rhs = GazeMode(gz);
}

enum CompositeMode
{
    kCompositeBoxed,
    kCompositeAllHigh,

    kNumCompositeModes
};
static CompositeMode compositeMode = kCompositeBoxed;
static const WCHAR* compositeStrings[kNumCompositeModes] =
{ 
    L"Boxed",
    L"All High",
};

bool GetCompositeOn()
{
    return compositeMode == kCompositeBoxed;
}
void SetCompositeOn(bool on)
{
    compositeMode = on ? kCompositeBoxed : kCompositeAllHigh;
}

struct RenderCounter
{
    int NumTris;
    int NumObjDrawn;
    int NumObjTested;
    int NumPix;
    double LastDraw;
    double DrawInterval;
    bool Active;

    RenderCounter()
        :   NumTris(0),
            NumObjDrawn(0),
            NumObjTested(0),
            NumPix(0),
            LastDraw(0),
            DrawInterval(0),
            Active(false)
    {
    }

    RenderCounter(const RenderCounter& rhs)
        :   NumTris(rhs.NumTris),
            NumObjDrawn(rhs.NumObjDrawn),
            NumObjTested(rhs.NumObjTested),
            NumPix(rhs.NumPix),
            Active(rhs.Active)
    {
    }

    RenderCounter& operator=(const RenderCounter& rhs)
    {
        NumTris = rhs.NumTris;
        NumObjDrawn = rhs.NumObjDrawn;
        NumObjTested = rhs.NumObjTested;
        NumPix = rhs.NumPix;
        Active = rhs.Active;
        return *this;
    }

    RenderCounter& operator+=(const RenderCounter& rhs)
    {
        NumTris += rhs.NumTris;
        NumObjTested += rhs.NumObjTested;
        NumObjDrawn += rhs.NumObjDrawn;
        NumPix += rhs.NumPix;
        return *this;
    }

    void Clear()
    {
        NumTris = 0;
        NumObjDrawn = 0;
        NumObjTested = 0;
        NumPix = 0;
        Active = false;
    }

    void SetDrawTime()
    {
        double drawTime = DXUTGetTime();
        DrawInterval = drawTime - LastDraw;
        LastDraw = drawTime;
        Active = true;
    }
};
static TunableFV8Mgr tuneFV8Mgr;
static FauxV8Mgr& fv8Mgr = tuneFV8Mgr;
static TunableFV8Mgr tuneNonFoveated;
static FauxV8Mgr& nonFv8Mgr = tuneNonFoveated;
static const int _maxCascades = BlitComposite::kMaxLayers;
static const int _baseCascade = BlitComposite::kBaseLayer;
static const int _maxUpper = 1;
static float _cascadeHighTris[_maxCascades] = { 20000, 20000, 9000 };
static float _cascadeLowTris[_maxCascades] = { 500, 500, 500 };
static float _cascadeHighDist[_maxCascades] = { 5.0f, 5.0f, 5.0f };
static float _cascadeLowDist[_maxCascades] = { 100.0f, 100.0f, 100.0f };
static RenderCounter _cascadeCounts[_maxCascades];
//--------------------------------------------------------------------------------------
// Forward declarations 
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing,
                          void* pUserContext );
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );


bool CALLBACK IsD3D11DeviceAcceptable( const CD3D11EnumAdapterInfo *AdapterInfo, UINT Output, const CD3D11EnumDeviceInfo *DeviceInfo,
                                       DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
HRESULT CALLBACK OnD3D11CreateDevice( ID3D11Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc,
                                     void* pUserContext );
HRESULT CALLBACK OnD3D11ResizedSwapChain( ID3D11Device* pd3dDevice, IDXGISwapChain* pSwapChain,
                                         const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void CALLBACK OnD3D11ReleasingSwapChain( void* pUserContext );
void CALLBACK OnD3D11DestroyDevice( void* pUserContext );
void CALLBACK OnD3D11FrameRender( ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext, double fTime,
                                 float fElapsedTime, void* pUserContext );

void MakeModel(bool advance);
void InitApp();
void RenderText();
static void ClearEnvironment(ID3D11DeviceContext* pd3dImmediateContext);

//--------------------------------------------------------------------------------------
int RandomModel()
{
	const double rndIdx = double(rand()) / double(RAND_MAX);

    const int numModels = ARRAYSIZE(modelShaders);
    const int baseModel = 0;
	
	return int(rndIdx * numModels) % numModels + baseModel;
}

class StrCompare
{
public:
    bool operator() (const WCHAR* lhs, const WCHAR* rhs) const
    {
        return wcscmp(lhs, rhs) < 0;
    }
};

typedef std::map<const WCHAR*, bool, StrCompare> CheckedMap;
typedef CheckedMap::iterator CheckedIter;
typedef std::pair<const WCHAR*, bool> CheckedEntry;

static CheckedMap checked;
static void UnloadCheckedMap()
{
    for( CheckedIter iter = checked.begin(); iter != checked.end(); ++iter)
    {
        delete[] iter->first;
    }
    checked.clear();
}
const WCHAR* GetLODName(const WCHAR* const shaderName, const int lod)
{
    WCHAR tmp[MAX_PATH];
    wcscpy_s(tmp, MAX_PATH, shaderName);
    WCHAR* p = wcsstr(tmp, L".3dmodel");
    assert(p);

    *p = NULL;

    WCHAR buff[MAX_PATH];
    wsprintf(buff, L"%s_%d.3dmodel.fx",
        tmp,
        lod);

    CheckedIter iter = checked.find(buff);
    if(iter != checked.end())
    {
        if(iter->second)
            return iter->first;
        return NULL;
    }

    const WCHAR* name = _wcsdup(buff);
    FILE* fp = NULL;
    if( _wfopen_s(&fp, name, L"r") )
    {
        // failed to open, return false;
        checked.insert(CheckedEntry(name, false));
        return NULL;
    }
    fclose(fp);
    checked.insert(CheckedEntry(name, true));
    return name;
}
// Discard any models we've loaded.
void CleanupModels()
{
	for(ObjectIter iter = objectCache.begin(); iter != objectCache.end(); ++iter)
	{
		Object* obj = *iter;
		obj->Cleanup();
		delete obj;
	}
    objectCache.clear();
	objects.clear();
    
    _scape->Cleanup();
    SAFE_DELETE(_scape);

    UnloadCheckedMap();
}

typedef std::map<const WCHAR*, AABB, StrCompare> BoundsMap;
typedef BoundsMap::iterator BoundsIter;
typedef std::pair<const WCHAR*, AABB> BoundsEntry;
BoundsMap boundsMap;
void InitBounds()
{

    boundsMap.insert(
        BoundsEntry(L"Shaders/rook_b.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.3f, -0.7f), D3DXVECTOR3( 0.3f, 0.3f, 0.3f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/knight_b.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.6f, -0.7f), D3DXVECTOR3( 0.3f, 0.3f, 0.3f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/bishop_b.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.3f, -0.8f), D3DXVECTOR3( 0.3f, 0.3f, 0.5f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/king_b.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.2f, -0.2f, -0.7f), D3DXVECTOR3( 0.2f, 0.2f, 0.5f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/queen_b.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.4f, -0.4f, -0.8f), D3DXVECTOR3( 0.4f, 0.4f, 0.7f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/pawn_b.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.3f, -0.7f), D3DXVECTOR3( 0.3f, 0.3f, 0.1f)))
                    );

    boundsMap.insert(
        BoundsEntry(L"Shaders/rook_w.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.3f, -0.7f), D3DXVECTOR3( 0.3f, 0.3f, 0.3f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/knight_w.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.6f, -0.7f), D3DXVECTOR3( 0.3f, 0.3f, 0.3f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/bishop_w.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.3f, -0.8f), D3DXVECTOR3( 0.3f, 0.3f, 0.5f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/king_w.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.2f, -0.2f, -0.7f), D3DXVECTOR3( 0.2f, 0.2f, 0.5f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/queen_w.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.4f, -0.4f, -0.8f), D3DXVECTOR3( 0.4f, 0.4f, 0.7f)))
                    );
    boundsMap.insert(
        BoundsEntry(L"Shaders/pawn_w.3dmodel.fx",
                    AABB( D3DXVECTOR3( -0.3f, -0.3f, -0.7f), D3DXVECTOR3( 0.3f, 0.3f, 0.1f)))
                    );

}
AABB GetBounds(const WCHAR* model)
{
    BoundsIter iter = boundsMap.find(model);
    if(iter != boundsMap.end())
    {
        return iter->second;
    }
    
    return AABB(D3DXVECTOR3(-1.46f, -1.46f, -1.46f),
            D3DXVECTOR3(1.46f, 1.46f, 1.46f));
}
Object* LoadModel(const WCHAR* const shaderName)
{
    LODObject* group = new LODObject();

    AABB bound = GetBounds(shaderName);

    Object* lod = NULL;
    lod = new ParametricSurface(shaderName);
    group->SetLODObject(lod, 0);

	const int MAX_LOD = 2;
	for(int i = 1; i <= MAX_LOD; ++i)
	{
		if( const WCHAR* buff = GetLODName(shaderName, i) )
		{
			lod = new ParametricSurface(buff);
			group->SetLODObject(lod, i);
		}
	}
    group->SetLocalBounds(bound.Min(), bound.Max());
    return group;
}

void CheckScape()
{
    if(!_scape)
    {
        if(!procTerrain)
        {
            _scape = new SimpleScape();
        }
        else
        {
            _scape = new ProcScape();
        }
        if(_scape->Init() != S_OK)
        {
            SAFE_DELETE(_scape);
        }
    }
    if(_scape)
    {
        _scape->SetChessBoard(modelSpacing);
        objects.push_back(_scape);
    }
}
void TearDownScape()
{
    if(_scape)
    {
        objects.erase(find(objects.begin(), objects.end(), _scape));

        _scape->Cleanup();
    }
    SAFE_DELETE(_scape);
}
void RebuildScape()
{
    TearDownScape();
    CheckScape();
}
//--------------------------------------------------------------------------------------
// Load the next model.
void MakeModel(bool advance)
{
    CWaitDlg CompilingShadersDlg;

    if(advance)
	    CleanupModels();

    objects.clear();

    CheckScape();

	/// This define creates a grid of objects randomly selected from modelShaders[]. Undefine
	/// it to create a simple test blob instead.
#define MAF_GENERATIVE_MODELS
#ifdef MAF_GENERATIVE_MODELS
    srand(1);

	const float spacing = modelSpacing;
	/// To enable a fly through of many objects, set the below modelX & modelY to values higher than 1.
	/// Models will be randomly selected from the pool modelShaders[] to fill out a modelsX x modelsY grid.
	const int modelsX = numModels.x;
	const int modelsY = numModels.y;
	const bool randomModels = modelsX * modelsY > 1;

	static bool firstGo = true;

    const bool showDialog = firstGo; // (modelsX * modelsY > int(objectCache.size()));
	firstGo = false;

    if(showDialog)
    {
        CompilingShadersDlg.ShowDialog( L"Compiling Shaders: Please wait, this could take several minutes" );
    }

	for(int y = 0; y < modelsY; ++y)
	{
		for(int x = 0; x < modelsX; ++x)
		{
            const int objectIdx = x + y * modelsX;
            Object* surf = objectIdx < int(objectCache.size()) ? objectCache[objectIdx] : NULL;

            if(!surf)
            {
                int thisIdx = 0;

			    if(randomModels)
			    {
				    thisIdx = RandomModel();
			    }
			    else
			    {
				    thisIdx = modelIdx;
				    if(advance)
				    {
					    do {
						    if(++modelIdx >= ARRAYSIZE(modelShaders))
						    {
							    modelIdx = 0;
						    }
					    } while((modelIdx != thisIdx) && !wcscmp(modelShaders[thisIdx], modelShaders[modelIdx]));
				    }
				    thisIdx = modelIdx;
			    }

			    const WCHAR* modelShader = modelShaders[thisIdx];

			    surf = LoadModel(modelShader);
                if(surf->Init() == S_OK)
                {
                    objectCache.push_back(surf);
                }
                else
                {
                    SAFE_DELETE(surf);
                }
            }
            if(surf)
			{
				D3DXMATRIX l2w;
                ZeroMemory(&l2w, sizeof(l2w));
                l2w.m[0][0] = 1.0f;
                l2w.m[1][2] = -1.0f;
                l2w.m[2][1] = 1.0f;
                l2w.m[3][3] = 1.0f;
                const D3DXVECTOR2 pos(float(x - (modelsX - 1.0) / 2.0) * spacing, float(y - (modelsY - 1.0) / 2.0) * spacing);
				l2w.m[3][0] = pos.x;
				l2w.m[3][2] = pos.y;
                if(_scape)
                {
                    D3DXVECTOR3 norm;
                    float height;
                    _scape->QueryHeight(pos, height, norm);
                    l2w.m[3][1] = height - surf->GetLocalBounds().Min().z;
                }
				surf->SetLocalToWorld(l2w);
				objects.push_back(surf);
			}
		}
	}
    if(showDialog)
    {
        CompilingShadersDlg.DestroyDialog();
    }

#else // MAF_GENERATIVE_MODELS
	/// Make a simple sphere.
    numModels.x = numModels.y = 1;
	Sphere* sphere = new Sphere();
    sphere->SetRadius(1.0f);
	if(sphere->Init() == S_OK)
	{
		objects.push_back(sphere);
	}
	else
	{
		delete sphere;
	}
#endif // MAF_GENERATIVE_MODELS
}

void ResizeModels()
{
	for(ObjectIter iter = objects.begin(); iter != objects.end(); ++iter)
	{
		Object* obj = *iter;
		obj->OnResize();
	}
}

void MoveCamera(float fElapsedTime)
{
    g_Camera.FrameMove( fElapsedTime );
}

void MoveModels(float fElapsedTime)
{
	D3DXMATRIX rotTotal;
	D3DXMatrixIdentity(&rotTotal);
	D3DXMATRIX rot;

	rotAccum += rotRate * fElapsedTime;

	D3DXMatrixRotationZ(&rot, rotAccum.z);
	D3DXMatrixMultiply(&rotTotal, &rotTotal, &rot);

	D3DXMatrixRotationY(&rot, rotAccum.y);
	D3DXMatrixMultiply(&rotTotal, &rotTotal, &rot);

	D3DXMatrixRotationX(&rot, rotAccum.x);
	D3DXMatrixMultiply(&rotTotal, &rotTotal, &rot);

	for(ObjectIter iter = objects.begin(); iter != objects.end(); ++iter)
	{
		Object* obj = *iter;

		const D3DXMATRIX& l2wCurr = obj->GetLocalToWorld();

		rotTotal.m[3][0] = l2wCurr.m[3][0];
		rotTotal.m[3][1] = l2wCurr.m[3][1];
		rotTotal.m[3][2] = l2wCurr.m[3][2];

		obj->SetLocalToWorld(rotTotal);

		if(!pauseTexture)
			obj->IncrementTime(fElapsedTime);
	}
}


RenderCounter RenderModels()
{
    GridObject::SetEveryFrame();

    static CullVolume culler;
    const D3DXMATRIX worldToNDC = *g_Camera.GetViewMatrix() * *g_Camera.GetProjMatrix();
    culler.Setup(worldToNDC);

	RenderCounter cnt;

    if(_modelRenderEnabled)
    {
	    for(ObjectIter iter = objects.begin(); iter != objects.end(); ++iter)
	    {
		    Object* obj = *iter;

            ++cnt.NumObjTested;

            if(obj->GetWorldBounds().CullTest(culler) != Bounds::kAllOut)
            {
		        cnt.NumTris += obj->Render();
                ++cnt.NumObjDrawn;
            }
	    }
    }
    else if(Scape::GetEnableDraw())
    {
        cnt.NumTris += _scape->Render();
        ++cnt.NumObjDrawn;
    }
	return cnt;
}

struct ReprojectionBlock
{
    D3DXVECTOR4     m_Line0; // blend, jitter scale, jitter table offset, unused
    D3DXVECTOR4     m_Line1; // width, height, 0.5 / width, 0.5 / height
};
static ID3D11Buffer*       pReprojectionCB = NULL;
void DisableReprojection()
{
    if(!pReprojectionCB)
    {
        D3D11_BUFFER_DESC bd;
        ZeroMemory( &bd, sizeof(bd) );
        bd.Usage = D3D11_USAGE_DEFAULT;
        bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
        bd.CPUAccessFlags = 0;

        bd.ByteWidth = sizeof(ReprojectionBlock);
        DXUTGetD3D11Device()->CreateBuffer( &bd, NULL, &pReprojectionCB);
    }
    ID3D11DeviceContext* pd3dImmediateContext = DXUTGetD3D11DeviceContext();
    ReprojectionBlock cb;
    cb.m_Line0 = D3DXVECTOR4(
        0.0f,
        0.0f,
        0.0f,
        0.0f);
    cb.m_Line1 = D3DXVECTOR4(
        0.0f,
        0.0f,
        0.0f,
        0.0f);
    pd3dImmediateContext->UpdateSubresource( pReprojectionCB, 0, NULL, &cb, 0, 0 );

    const UINT kConstantBufferSlot = 10;
    pd3dImmediateContext->VSSetConstantBuffers( kConstantBufferSlot, 1, &pReprojectionCB );
    pd3dImmediateContext->PSSetConstantBuffers( kConstantBufferSlot, 1, &pReprojectionCB );
}
void ClearCounts()
{
    for(int layer = 0; layer < _maxCascades; ++layer)
    {
        _cascadeCounts[layer].Clear();
    }
}
void PreRenderCounts()
{
    for(int layer = 0; layer < _maxCascades; ++layer)
    {
        _cascadeCounts[layer].Active = false;
    }
}

void RenderHighLoop(ID3D11DeviceContext* pd3dImmediateContext, const D3DXMATRIX& camToNDC)
{
    const int iLayer = 0;

    RenderCounter& cnt = _cascadeCounts[iLayer];
    cnt.Clear();

    CameraSave cameraCache;
    cameraCache.Save(g_Camera);
    g_Camera.SetProjection(&camToNDC);
    // Temp bit
    g_Camera.SetLastWorldToNDC(nonFv8Mgr.GetLastWorldToNDC());

    if(textureBackground)
    {
        ClearEnvironment(pd3dImmediateContext);
    }
    GridObject::SetHighNumTris(_cascadeHighTris[iLayer]);
    GridObject::SetLowNumTris(_cascadeLowTris[iLayer]);
    GridObject::SetHighDistance(_cascadeHighDist[iLayer]);
    GridObject::SetLowDistance(_cascadeLowDist[iLayer]);

    GridObject::SetFrameDecay(tuneNonFoveated.GetFrameDecay(iLayer));

    const POINT size = nonFv8Mgr.GetResolution(iLayer);
    GridObject::SetWindowSize(D3DXVECTOR2(
        FLOAT(size.x), 
        FLOAT(size.y)));

    cnt += RenderModels();
    cnt.NumPix += size.x * size.y;
    cnt.SetDrawTime();

    cameraCache.Restore(g_Camera);
}
void RenderModelsHigh()
{
    if(nonFv8Mgr.NeedInit())
    {
        tuneNonFoveated.SetNumLayers(1);
        tuneNonFoveated.SetRenderInterval(0, FauxV8Mgr::eRenderInterval::kRenderAlways);
        tuneNonFoveated.SetRefreshSecs(0, 0.0f);
        const RECT rc = DXUTGetWindowClientRect();
        tuneNonFoveated.SetSize(0, rc.right - rc.left);
        tuneNonFoveated.SetMSRead(0, false);
        tuneNonFoveated.SetMSAA(tuneFV8Mgr.GetMSAA());

        nonFv8Mgr.Init(DXUTGetD3D11Device());
    }
    tuneNonFoveated.SetFrameDecay(0, tuneFV8Mgr.GetFrameDecay(0));
    tuneNonFoveated.SetJitterScale(0, tuneFV8Mgr.GetJitterScale(0));
    
    ID3D11DeviceContext* pd3dImmediateContext = DXUTGetD3D11DeviceContext();

    // Send this frame's camera information.
    nonFv8Mgr.SetCamera(*g_Camera.GetViewMatrix(), *g_Camera.GetProjMatrix(), g_Camera.GetNearClip(), g_Camera.GetFarClip());
    // Send the gaze position within the window.
    nonFv8Mgr.SetFoveaCenter(GetGazePosition());

    // Only the one layer for the non-foveated.
    const int iLayer = 0;

        // Different layers refresh at different rates. Check if this layer is due.
    assert(nonFv8Mgr.ShouldRefresh(iLayer, DXUTGetElapsedTime()));

    // Pushing this layer as a render target returns a modified projection matrix.
    D3DXMATRIX cam2NDC = nonFv8Mgr.PushLayer(iLayer, pd3dImmediateContext);

    // Clear the screen.
    const float clear[] = { 1.0f, 0.0f, 0.0f, 1.0f };
    nonFv8Mgr.Clear(pd3dImmediateContext, clear);

    // Render the scene as normal, using the modified projection matrix. This
    // presumes the cull volume is generated from the camera transforms. Also,
    // LOD and effects quality can be reduced for increasing layer index.
    RenderHighLoop(pd3dImmediateContext, cam2NDC);

    // Finish this layer, resetting state to before the PushLayer() above.
    nonFv8Mgr.PopLayer(pd3dImmediateContext);

    // Now composite the layers into the destination render target (the back buffer here).
    ID3D11RenderTargetView* pRTV = DXUTGetD3D11RenderTargetView();
    nonFv8Mgr.FinishFrame(pd3dImmediateContext, pRTV);
}

float AdvanceRot(float rot)
{
	if(rot > 0)
	{
		rot = -spinSpeed * spinBoost;
	}
	else if(rot < 0)
	{
		rot = 0;
	}
	else
	{
		rot = spinSpeed * spinBoost;
	}
	return rot;
}

static const int kWindowSize = 10;
static float _smoothDelta[kWindowSize] = { 0.0f };
static int _smoothCurrent = -1;
static void AddSmoothSecs(float dt)
{
    // It's generally a good idea to clamp dt here, but that's pretty app specific.
    if(_smoothCurrent < 0)
    {
        for(int i = 0; i < kWindowSize; ++i)
            _smoothDelta[i] = dt;
        _smoothCurrent = 0;
    }
    else
    {
        _smoothDelta[_smoothCurrent] = dt;
        if(++_smoothCurrent >= kWindowSize)
            _smoothCurrent = 0;
    }
}
static float GetSmoothSecs()
{
    float dt = 0;
    for(int i = 0; i < kWindowSize; ++i)
        dt += _smoothDelta[i];
    return dt / kWindowSize;
}
static int GetAddedLatencyUSecs()
{
    int addedMS = latencyMS;
    if(addedMS)
    {
        addedMS -= int(GetSmoothSecs() * 1000.0f * 1.5f + 0.5f);
    }
    return addedMS * 1000;
}

static bool IsGazeDataValid()
{
    // GAZE DATA STUB.
    // Returning false here when unreliable gaze data appears will enable the 
    // status bar to fade in from green to red as the gaze tracking deteriorates.
    // This is helpful in separating failure of the foveation technique from
    // faulty gaze data.
    return false;
}
static void ReadGazePosFromCursor(POINT& pos)
{
    if(gazeMode == kMouse)
    {
        GetCursorPos(&_gazePos);
        ScreenToClient(DXUTGetHWND(), &_gazePos);
    }

}
static void ReadGazePos(POINT& pos)
{
    if(gazeMode == kGaze)
    {
        // GAZE DATA STUB
        // Fill in the pos struct here with gaze data (if available).
        // pos coordinates are in window pixel coordinates.
        // For reference, see ReadGazePosFromCursor(), which emulates gaze data from
        // the cursor position.
    }
    else if(gazeMode == kMouse)
    {
        ReadGazePosFromCursor(pos);
    }
}

void SetupProcessor(int nth)
{

    SYSTEM_INFO systemInfo;
    ZeroMemory(&systemInfo, sizeof(systemInfo));
    GetSystemInfo(&systemInfo);

#if defined(DEBUG) | defined(_DEBUG)
    const int kProcessor = systemInfo.dwNumberOfProcessors - nth;

    const int kProcessorMask = 1 << kProcessor;
    DWORD procAffinityMask = 0;
    DWORD sysAffinityMask = 0;
    GetProcessAffinityMask(GetCurrentProcess(), &procAffinityMask, &sysAffinityMask);
    assert(procAffinityMask & sysAffinityMask & kProcessorMask);
#endif // DEBUG

    const DWORD kPriority = REALTIME_PRIORITY_CLASS;
    if(!SetPriorityClass(GetCurrentProcess(), kPriority))
    {
        MessageBox( 0, L"Failure setting PriorityClass", L"Error", MB_ICONEXCLAMATION );
        assert(false);
    }
}


static FILE* _settingsFP = NULL;
static float _settingsInterval = -1.0f;
static float _settingsBlank = -1.0f;
static int _settingsLoops = -1;
static bool _settingsReset = false;

typedef std::vector<SettingBlock> SettingVec;
typedef SettingVec::iterator SettingIter;
typedef SettingVec::const_iterator SettingConstIter;
static SettingVec _settings;
static size_t _settingsCurrent = 0;
static WCHAR _settingsLabel[MAX_PATH] = L"";
size_t GetCurrentSettingIndex() { return _settingsCurrent; }
const SettingBlock& GetCurrentSetting() { return _settings[_settingsCurrent]; }

static int ScaleToScreen(int val)
{
//#define MAF_SCALE_ECCENTRICITY_LAYERS
#ifdef MAF_SCALE_ECCENTRICITY_LAYERS
    val = int(val * _screenSize.x / 1920.0);
#endif // MAF_SCALE_ECCENTRICITY_LAYERS
    return val;
}
void SetInnerRadius(const SettingBlock::ValueChunk& val)
{
    const int size = ScaleToScreen(val.iValue);
    tuneFV8Mgr.SetSize(0, size);
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Inner resolution %d", tuneFV8Mgr.GetSize(0) );
    g_SampleUI.GetStatic( IDC_STATIC_INNER_RES )->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_INNER_RES )->SetValue( tuneFV8Mgr.GetSize(0) );
}
void SetMidRadius(const SettingBlock::ValueChunk& val)
{
    const int size = ScaleToScreen(val.iValue);
    tuneFV8Mgr.SetSize(1, size);
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Mid resolution %d", tuneFV8Mgr.GetSize(1) );
    g_SampleUI.GetStatic( IDC_STATIC_MID_RES)->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_MID_RES )->SetValue( tuneFV8Mgr.GetSize(1) );
}
void SetBaseRadius(const SettingBlock::ValueChunk& val)
{
    const int size = ScaleToScreen(val.iValue);
    tuneFV8Mgr.SetSize(2, size);
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Base resolution %d", tuneFV8Mgr.GetSize(2) );
    g_SampleUI.GetStatic( IDC_STATIC_BASE_RES)->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_BASE_RES )->SetValue( tuneFV8Mgr.GetSize(2) );
}
void SetLatency(const SettingBlock::ValueChunk& val)
{
    latencyMS = val.iValue;
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Latency ms %d", latencyMS);
    g_SampleUI.GetStatic( IDC_STATIC_LATENCY )->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_LATENCY )->SetValue( latencyMS );
}
void SetNumModels(int num)
{
    numModels.x = numModels.y = num;
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Num Models %d", numModels.x * numModels.y);
    g_SampleUI.GetStatic( IDC_STATIC_MODELSX )->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_MODELSX )->SetValue( numModels.x );

    MakeModel(false);
}
void SetNumModels(const SettingBlock::ValueChunk& val)
{
    SetNumModels(val.iValue);
}
void SetCameraMode(int idx)
{
    const CameraModeSelect::eCameraModes mode = CameraModeSelect::eCameraModes(idx);
    g_Camera.SetFlyMode(mode);
    g_SampleUI.GetComboBox( IDC_COMBOBOX_CAMERA_MODE )->SetSelectedByIndex(idx);
}
void SetCameraMode(const SettingBlock::ValueChunk& val)
{
    SetCameraMode(val.iValue);
}
void SetCameraSpeed(float speed)
{
    g_Camera.SetSpeedGoal(speed);
}
void SetCameraSpeed(const SettingBlock::ValueChunk& val)
{
    SetCameraSpeed(val.fValue);
}
void SetCameraFrequency(const SettingBlock::ValueChunk& val)
{
    CameraPan::SetFrequency(val.fValue);
}
void SetCameraAmplitude(const SettingBlock::ValueChunk& val)
{
    CameraPan::SetAmplitude(val.fValue);
}
void SetCompositeMode(int iValue)
{
    int mode = iValue < 0 ? 0 : (iValue >= kNumCompositeModes ? kNumCompositeModes - 1 : iValue);
    compositeMode = CompositeMode(mode);
    g_SampleUI.GetComboBox(IDC_COMBOBOX_COMPOSITE)->SetSelectedByIndex(compositeMode);
}
void SetCompositeMode(const SettingBlock::ValueChunk& val)
{
    SetCompositeMode(val.iValue);
}
void SetInnerDecay(const SettingBlock::ValueChunk& val)
{
    float f = val.fValue;
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Blend Cen %.2f", f);
    g_SampleUI.GetStatic( IDC_STATIC_FRAME_DECAY_CEN)->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_FRAME_DECAY_CEN)->SetValue( int(f * 100.0f) );
    tuneFV8Mgr.SetFrameDecay(0, f);
}
void SetMidDecay(const SettingBlock::ValueChunk& val)
{
    float f = val.fValue;
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Blend Mid %.2f", f);
    g_SampleUI.GetStatic( IDC_STATIC_FRAME_DECAY_MID)->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_FRAME_DECAY_MID)->SetValue( int(f * 100.0f) );
    tuneFV8Mgr.SetFrameDecay(1, f);
}
void SetOuterDecay(const SettingBlock::ValueChunk& val)
{
    float f = val.fValue;
    WCHAR szTemp[MAX_PATH];
    swprintf_s( szTemp, L"Blend Out %.2f", f);
    g_SampleUI.GetStatic( IDC_STATIC_FRAME_DECAY_OUT)->SetText( szTemp );
    g_SampleUI.GetSlider( IDC_SLIDER_FRAME_DECAY_OUT)->SetValue( int(f * 100.0f) );
    tuneFV8Mgr.SetFrameDecay(2, f);
}
void SetUseMiddle(const SettingBlock::ValueChunk& val)
{
    bool useMid = val.bValue;
    g_SampleUI.GetCheckBox(IDC_CHECKBOX_USE_MID)->SetChecked(useMid);
    tuneFV8Mgr.SetUseMiddle(useMid);
}
void SetSettingsLabel(const SettingBlock::ValueChunk& val)
{
    wcscpy_s(_settingsLabel, val.cValue);
}

SettingBlock CreateSettingsPattern()
{
    SettingBlock pattern;

    pattern.AddChunk(SettingBlock::ValueChunk(L"InnerRadius", &SetInnerRadius, &SettingBlock::ValueChunk::GetInt));

    pattern.AddChunk(SettingBlock::ValueChunk(L"MidRadius", &SetMidRadius, &SettingBlock::ValueChunk::GetInt));

    pattern.AddChunk(SettingBlock::ValueChunk(L"BaseRadius", &SetBaseRadius, &SettingBlock::ValueChunk::GetInt));

    pattern.AddChunk(SettingBlock::ValueChunk(L"NumModels", &SetNumModels, &SettingBlock::ValueChunk::GetInt));

    pattern.AddChunk(SettingBlock::ValueChunk(L"Latency", &SetLatency, &SettingBlock::ValueChunk::GetInt));

    pattern.AddChunk(SettingBlock::ValueChunk(L"CameraMode", &SetCameraMode, &SettingBlock::ValueChunk::GetInt));

    pattern.AddChunk(SettingBlock::ValueChunk(L"CameraSpeed", &SetCameraSpeed, &SettingBlock::ValueChunk::GetFloat));

    pattern.AddChunk(SettingBlock::ValueChunk(L"CameraFrequency", &SetCameraFrequency, &SettingBlock::ValueChunk::GetFloat));

    pattern.AddChunk(SettingBlock::ValueChunk(L"CameraAmplitude", &SetCameraAmplitude, &SettingBlock::ValueChunk::GetFloat));

    pattern.AddChunk(SettingBlock::ValueChunk(L"CompositeMode", &SetCompositeMode, &SettingBlock::ValueChunk::GetInt));

    pattern.AddChunk(SettingBlock::ValueChunk(L"ReprojInner", &SetInnerDecay, &SettingBlock::ValueChunk::GetFloat));

    pattern.AddChunk(SettingBlock::ValueChunk(L"ReprojMid", &SetMidDecay, &SettingBlock::ValueChunk::GetFloat));

    pattern.AddChunk(SettingBlock::ValueChunk(L"ReprojOuter", &SetOuterDecay, &SettingBlock::ValueChunk::GetFloat));

    pattern.AddChunk(SettingBlock::ValueChunk(L"UseMiddle", &SetUseMiddle, &SettingBlock::ValueChunk::GetBool));

    pattern.AddChunk(SettingBlock::ValueChunk(L"Label", &SetSettingsLabel, &SettingBlock::ValueChunk::GetString));

    return pattern;
}

bool DXUTIsNextArg( WCHAR*& strCmdLine, const WCHAR* strArg );
bool DXUTGetCmdParam( WCHAR*& strCmdLine, WCHAR* strFlag );
void ParseCommandLine(LPWSTR lpCmdLine)
{
    WCHAR buff[MAX_PATH];
    int nNumArgs;
    LPWSTR* pstrArgList = CommandLineToArgvW( lpCmdLine, &nNumArgs );
    for(int iArg = 0; iArg < nNumArgs; ++iArg)
    {
        LPWSTR strCmd = pstrArgList[iArg];
        if( *strCmd == L'/' || *strCmd == L'-' )
        {
            strCmd++;

            // Delete any trailing special characters. The DX code isn't expecting them.
            if(strCmd[wcslen(strCmd) - 1] <= ' ')
                strCmd[wcslen(strCmd) - 1] = 0;

            if( DXUTIsNextArg( strCmd, L"settings" ) )
            {
                if( DXUTGetCmdParam( strCmd, buff ) )
                {
                    _wfopen_s(&_settingsFP, buff, L"rt");
                }
            }

            if(DXUTIsNextArg( strCmd, L"autoset" ))
            {
                if(DXUTGetCmdParam( strCmd, buff))
                {
                    _settingsInterval = float(_wtof(buff));
                }
            }

            if(DXUTIsNextArg( strCmd, L"blank" ) )
            {
                if(DXUTGetCmdParam( strCmd, buff))
                {
                    _settingsBlank = float(_wtof(buff));
                }
            }

            if(DXUTIsNextArg( strCmd, L"loops" ) )
            {
                if(DXUTGetCmdParam( strCmd, buff))
                {
                    _settingsLoops = _wtoi(buff);
                }
            }

            if(DXUTIsNextArg( strCmd, L"silent" ) )
            {
                bool silent = true;
                if(DXUTGetCmdParam( strCmd, buff))
                {
                    silent = _wtoi(buff) != 0;
                }
                if(silent)
                {
                    displayStats = false;
                    g_SampleUI.SetVisible(false);
                }
            }

            if( DXUTIsNextArg( strCmd, L"reset" ) )
            {
                _settingsReset = true;
                if( DXUTGetCmdParam( strCmd, buff ) )
                {
                    _settingsReset = _wtoi(buff) != 0;
                }
            }

            if( DXUTIsNextArg( strCmd, L"maximize" ) )
            {
                _maximizeWindow = true;
                if(DXUTGetCmdParam( strCmd, buff ) )
                {
                    _maximizeWindow = _wtoi(buff) != 0;
                }
            }
            if( DXUTIsNextArg( strCmd, L"noTer" ) )
            {
                Scape::SetEnableDraw(false);
            }
            if( DXUTIsNextArg( strCmd, L"noMod" ) )
            {
                _modelRenderEnabled = false;
            }
            if( DXUTIsNextArg( strCmd, L"noProc" ) )
            {
                procTerrain = false;
            }
            if( DXUTIsNextArg( strCmd, L"noEnv" ) )
            {
                textureBackground = false;
            }
            if( DXUTIsNextArg( strCmd, L"nobo" ) )
            {
                DXUTSetBorderless(true);
            }
        }
    }

    static LPWSTR forceSettings = L"settings.txt";
    if(!_settingsFP && forceSettings && *forceSettings)
    {
        _wfopen_s(&_settingsFP, forceSettings, L"rt");
    }
}

void LoadSettings()
{
    SettingBlock pattern = CreateSettingsPattern();
    _settings.clear();
    FILE* fp = _settingsFP;
    if(fp)
    {
        WCHAR cmdLine[MAX_PATH];
        bool go = fgetws(cmdLine, MAX_PATH, fp) != NULL;
        if(go)
        {
            ParseCommandLine(cmdLine);
        }
        while(go)
        {
            SettingBlock block;
            if(block.Load(fp, pattern))
            {
                _settings.push_back(block);
            }
            else
            {
                go = false;
            }
        }
        if(_settings.size())
        {
            _settingsCurrent = 0;
        }
        fclose(fp);

    }
}
void ReloadSettings()
{
    if(_settingsCurrent < _settings.size())
    {
        _settings[_settingsCurrent].Set();
    }
}

//--------------------------------------------------------------------------------------
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    SetupProcessor(2);

    // DXUT will create and use the best device (either D3D9 or D3D11) 
    // that is available on the system depending on which D3D callbacks are set below

    // Set DXUT callbacks
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackKeyboard( OnKeyboard );
    DXUTSetCallbackFrameMove( OnFrameMove );
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );

    DXUTSetCallbackD3D11DeviceAcceptable( IsD3D11DeviceAcceptable );
    DXUTSetCallbackD3D11DeviceCreated( OnD3D11CreateDevice );
    DXUTSetCallbackD3D11SwapChainResized( OnD3D11ResizedSwapChain );
    DXUTSetCallbackD3D11SwapChainReleasing( OnD3D11ReleasingSwapChain );
    DXUTSetCallbackD3D11DeviceDestroyed( OnD3D11DestroyDevice );
    DXUTSetCallbackD3D11FrameRender( OnD3D11FrameRender );

    ParseCommandLine(lpCmdLine);
    LoadSettings();
    InitApp();
    DXUTInit( true, false, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
    DXUTSetCursorSettings( true, true );
    DXUTCreateWindow( L"Faux V-8 (3D)" );

    // Only require 10-level hardware, change to D3D_FEATURE_LEVEL_11_0 to require 11-class hardware
    // Switch to D3D_FEATURE_LEVEL_9_x for 10level9 hardware
    DXUTCreateDevice( D3D_FEATURE_LEVEL_10_0, true, 1024, 768 );
	//DXUTCreateDevice( D3D_FEATURE_LEVEL_10_0, true, 640, 480 );
	//DXUTCreateDevice( D3D_FEATURE_LEVEL_10_0, true, 1280, 960);

    if(_maximizeWindow)
    {
        ShowWindow( DXUTGetHWND(), SW_SHOWMAXIMIZED );
    }
    SetForegroundWindow( DXUTGetHWND() );

    DXUTMainLoop(); // Enter into the DXUT render loop

    return DXUTGetExitCode();
}

static float ToLogScale(float v, float s)
{
    return log10(v / s);
}

static float FromLogScale(float v, float s)
{
    return pow(10.f, v) * s;
}

static float kNumTrisLogScale = 2.0f;
static float kDistanceLogScale = 2.0f;
static const int IDC_STRING_OFFSET = 300;


void SetLODSliderLabel(int nControlID, float value)
{
    const int baseID = nControlID - IDC_SLIDER_HIGH_TRIS_IN;
    const int iCasc = baseID / 4;
    const int iParm = baseID % 4;

    WCHAR szTemp[256];

    const WCHAR* hilo = iParm < 2 ? L"Hi" : L"Lo";
    const WCHAR* tridist = iParm & 1 ? L"Dist" : L"Tris";
    const WCHAR* labels[3] = { L"Cen", L"Mid", L"Bkg" };

    swprintf_s(szTemp, L"%s %s %s %d", hilo, tridist, labels[iCasc], int(value));

    g_SampleUI.GetStatic( nControlID + IDC_STRING_OFFSET )->SetText( szTemp );
}

static void AddNumTrisSlider(UINT baseIDC, int& iX_0, int& iY_0, int& iX_1, int& iY_1,
    float highNumTris, float highDistance, float lowNumTris, float lowDistance)
{
    {
        g_SampleUI.AddStatic( baseIDC + 0 + IDC_STRING_OFFSET, L"", iX_0+5, iY_0 += 25, 108, 24 );

        float logHighTris = ToLogScale(highNumTris, kNumTrisLogScale) * 100.0f;
        g_SampleUI.AddSlider( baseIDC + 0, iX_0, iY_0 += 25, 140, 24, 200, 500,
                              int(logHighTris + 0.5f), false );
        
        SetLODSliderLabel(baseIDC + 0, highNumTris);
    }

    {
        g_SampleUI.AddStatic( baseIDC + 1 + IDC_STRING_OFFSET, L"", iX_1+5, iY_1 += 25, 108, 24 );

	    float logHighDistance = ToLogScale(highDistance, kDistanceLogScale) * 100.0f;
        g_SampleUI.AddSlider( baseIDC + 1, iX_1, iY_1 += 25, 140, 24, 50, 300,
                              int(logHighDistance+ 0.5f), false );
        SetLODSliderLabel(baseIDC + 1, highDistance);
    }

    {
        g_SampleUI.AddStatic( baseIDC + 2 + IDC_STRING_OFFSET, L"", iX_0+5, iY_0 += 25, 108, 24 );

	    float logLowTris = ToLogScale(lowNumTris, kNumTrisLogScale) * 100.0f;
        g_SampleUI.AddSlider( baseIDC + 2, iX_0, iY_0 += 25, 140, 24, 200, 500,
                              int(logLowTris + 0.5f), false );
        SetLODSliderLabel(baseIDC + 2, lowNumTris);
    }

    {
        g_SampleUI.AddStatic( baseIDC + 3 + IDC_STRING_OFFSET, L"", iX_1+5, iY_1 += 25, 108, 24 );

	    float logLowDistance = ToLogScale(lowDistance, kDistanceLogScale) * 100.0f;
        g_SampleUI.AddSlider( baseIDC + 3, iX_1, iY_1 += 25, 140, 24, 50, 300,
                              int(logLowDistance + 0.5f), false );
        SetLODSliderLabel(baseIDC + 3, lowDistance);
    }
}


int log2Hack(UINT v)
{
    assert(v);
    int logVal = 0;
    while(v >>= 1)
    {
        ++logVal;
    }
    return logVal;
}
static void InitCameraView()
{
    // Setup the camera's view parameters
    D3DXVECTOR3 vecEye( 0.0f, 2.0f, 5.0f );
    D3DXVECTOR3 vecAt ( 0.0f, 2.0f, 0.0f );
    g_Camera.SetViewParams( &vecEye, &vecAt );

}
//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{
    g_SettingsDlg.Init( &g_DialogResourceManager );
    //g_HUD.Init( &g_DialogResourceManager );
    g_SampleUI.Init( &g_DialogResourceManager );
    g_SampleUI.EnableCaption( true );
    g_SampleUI.SetCaptionText( L"Parameters" );
    g_SampleUI.SetBackgroundColors(0x9f4f4f4f);
    g_SampleUI.SetMinimized(true);
    g_SampleUI.EnableKeyboardInput(false);

    WCHAR szTemp[256];

    //g_HUD.SetCallback( OnGUIEvent );
    //int iY = 30;

#ifndef MAF_NO_HUD
    //g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 0, iY, 170, 22 );
    //g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 0, iY += iYo, 170, 22, VK_F3 );
    //g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 0, iY += iYo, 170, 22, VK_F2 );
    //g_HUD.SetVisible(false);

	int iX_Col0 = 20;
    int iX_Col1 = 200;
	int iY_Col0 = 0;
    int iY_Col1 = 0;

    g_SampleUI.AddCheckBox( IDC_CHECKBOX_MSBASE, L"MultSamp Back", iX_Col0, iY_Col0 += 25, 140, 24, 
                            tuneFV8Mgr.GetMSRead(2), 0);
    g_SampleUI.AddCheckBox( IDC_CHECKBOX_MSMID, L"MultSamp Mid", iX_Col0, iY_Col0 += 25, 140, 24, 
                            tuneFV8Mgr.GetMSRead(1), 0);
    g_SampleUI.AddCheckBox( IDC_CHECKBOX_MSINNER, L"MultSamp Inner", iX_Col0, iY_Col0 += 25, 140, 24, 
                            tuneFV8Mgr.GetMSRead(0), 0);
    g_SampleUI.AddCheckBox( IDC_CHECKBOX_MSAAINNER, L"MSAA Inner", iX_Col0, iY_Col0 += 25, 140, 24, 
                            tuneFV8Mgr.GetMSAAInner() != 0, 0);

    //g_SampleUI.AddCheckBox( IDC_CHECKBOX_USE_LODS, L"Use Model LODs", iX_Col0, iY_Col0 += 25, 140, 24, 
    //                        useLOD, 0);

    g_SampleUI.AddCheckBox( IDC_CHECKBOX_USE_MID, L"Use Mid", iX_Col1, iY_Col1 += 25, 140, 24, 
                            tuneFV8Mgr.GetUseMiddle(), 0);

    g_SampleUI.AddCheckBox( IDC_CHECKBOX_SHOW_LEVELS, L"Show Levels", iX_Col0, iY_Col0 += 25, 140, 24, 
                            BlitComposite::GetShowLevelMode() != BlitComposite::kShowNone, 0);

    g_SampleUI.AddCheckBox( IDC_CHECKBOX_LIGHTS, L"Use Lights", iX_Col1, iY_Col1 += 25, 140, 24, 
                            GridObject::GetLights(), 0);
    g_SampleUI.AddCheckBox( IDC_CHECKBOX_SPECULAR, L"Use Specular", iX_Col1, iY_Col1 += 25, 140, 24, 
                            GridObject::GetSpecular(), 0);
    g_SampleUI.AddCheckBox( IDC_CHECKBOX_ENVIRONMENT, L"Use Env Map", iX_Col1, iY_Col1 += 25, 140, 24, 
                            GridObject::GetEnvMap(), 0);
    g_SampleUI.AddCheckBox( IDC_CHECKBOX_BACKGROUND, L"Texture Background", iX_Col1, iY_Col1 += 25, 140, 24, 
                            textureBackground, 0);

    //g_SampleUI.AddCheckBox( IDC_CHECKBOX_WIREFRAME, L"Wireframe", iX_Col1, iY_Col1 += 25, 140, 24, 
    //                        GridObject::GetWireFrame(), 0);

    //g_SampleUI.AddCheckBox( IDC_CHECKBOX_SHOWBOUNDS, L"Show Bounds", iX_Col0, iY_Col0 += 25, 140, 24, 
    //                        GridObject::GetShowBounds(), 0);

    //g_SampleUI.AddCheckBox( IDC_CHECKBOX_SHOW_HUD, L"Show HUD", iX_Col1, iY_Col1 += 25, 140, 24, 
    //                        false, 0);

    /// Resync our columns.
    iY_Col0 = max(iY_Col0, iY_Col1) + 10;
    iY_Col1 = iY_Col0;

    g_SampleUI.AddComboBox( IDC_COMBOBOX_CAMERA_MODE, iX_Col0, iY_Col0 += 25, 108, 24 );
    for(int i = 0; i < CameraModeSelect::kNumModes; ++i)
    {
        g_SampleUI.GetComboBox( IDC_COMBOBOX_CAMERA_MODE)->AddItem( CameraModeSelect::Name(CameraModeSelect::eCameraModes(i)), NULL );
    }
    g_SampleUI.GetComboBox( IDC_COMBOBOX_CAMERA_MODE )->SetSelectedByIndex(g_Camera.GetFlyMode());

    g_SampleUI.AddComboBox( IDC_COMBOBOX_COMPOSITE, iX_Col1, iY_Col1 += 25, 108, 24 );
    for(int i = 0; i < kNumCompositeModes; ++i)
    {
        g_SampleUI.GetComboBox( IDC_COMBOBOX_COMPOSITE )->AddItem( compositeStrings[i], NULL );
    }
    g_SampleUI.GetComboBox( IDC_COMBOBOX_COMPOSITE )->SetSelectedByIndex(compositeMode);

    /// Resync our columns.
    iY_Col0 = max(iY_Col0, iY_Col1) + 10;
    iY_Col1 = iY_Col0;

 
    // Col0 begin
    swprintf_s( szTemp, L"Latency ms %d", latencyMS );
    g_SampleUI.AddStatic(IDC_STATIC_LATENCY, szTemp, iX_Col0 + 5, iY_Col0 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_LATENCY, iX_Col0, iY_Col0 += 25, 140, 24, 0, 100, latencyMS, false );

    int msaa = log2Hack(GridObject::GetMSAA());
    swprintf_s( szTemp, L"MSAA %d", GridObject::GetMSAA() );
    g_SampleUI.AddStatic( IDC_STATIC_MSAA, szTemp, iX_Col0+5, iY_Col0 += 25, 108, 24 );
    g_SampleUI.AddSlider( IDC_SLIDER_MSAA, iX_Col0, iY_Col0 += 25, 140, 24, 0, 3,
                          msaa, false );


    swprintf_s( szTemp, L"Composite decay %d", int(tuneFV8Mgr.GetCompositeBlendStart() * 100.0f) );
    g_SampleUI.AddStatic( IDC_STATIC_COMP_FALLOFF, szTemp, iX_Col0+5, iY_Col0 += 25, 108, 24 );
    g_SampleUI.AddSlider( IDC_SLIDER_COMP_FALLOFF, iX_Col0, iY_Col0 += 25, 140, 24, 0, 100,
                          int(tuneFV8Mgr.GetCompositeBlendStart() * 100.0f), false );

    swprintf_s( szTemp, L"Num Model %d", numModels.x * numModels.y );
    g_SampleUI.AddStatic(IDC_STATIC_MODELSX, szTemp, iX_Col0 + 5, iY_Col0 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_MODELSX, iX_Col0, iY_Col0 += 25, 140, 24, 0, 100, numModels.x, false );


    swprintf_s( szTemp, L"Inner resolution %d", tuneFV8Mgr.GetSize(0) );
    g_SampleUI.AddStatic( IDC_STATIC_INNER_RES, szTemp, iX_Col1+5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider( IDC_SLIDER_INNER_RES, iX_Col1, iY_Col1 += 25, 140, 24, 32, 1024,
                          tuneFV8Mgr.GetSize(0), false );

    swprintf_s( szTemp, L"Mid resolution %d", tuneFV8Mgr.GetSize(1) );
    g_SampleUI.AddStatic( IDC_STATIC_MID_RES, szTemp, iX_Col1+5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider( IDC_SLIDER_MID_RES, iX_Col1, iY_Col1 += 25, 140, 24, 16, 1024,
                          tuneFV8Mgr.GetSize(1), false );

    //swprintf_s( szTemp, L"Mid pixel cover %.1f", _cascadeDups[1]);
    //g_SampleUI.AddStatic( IDC_STATIC_MID_DUP, szTemp, iX_Col1+5, iY_Col1 += 25, 108, 24 );
    //g_SampleUI.AddSlider( IDC_SLIDER_MID_DUP, iX_Col1, iY_Col1 += 25, 140, 24, 10, 80,
    //                      int(_cascadeDups[1] * 10.0f), false );

    swprintf_s( szTemp, L"Back resolution %d", tuneFV8Mgr.GetSize(2) );
    g_SampleUI.AddStatic( IDC_STATIC_BASE_RES, szTemp, iX_Col1+5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider( IDC_SLIDER_BASE_RES, iX_Col1, iY_Col1 += 25, 140, 24, 16, 1024,
                          tuneFV8Mgr.GetSize(2), false );

    swprintf_s( szTemp, L"Model Spacing %.1f", modelSpacing );
    g_SampleUI.AddStatic(IDC_STATIC_MODEL_SPACING, szTemp, iX_Col1 + 5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_MODEL_SPACING, iX_Col1, iY_Col1 += 25, 140, 24, 1, 200, int(modelSpacing * 10.0f), false );

    /// Resync our columns.
    iY_Col0 = max(iY_Col0, iY_Col1);
    iY_Col1 = iY_Col0;

    AddNumTrisSlider(IDC_SLIDER_HIGH_TRIS_IN, iX_Col0, iY_Col0, iX_Col1, iY_Col1,
        _cascadeHighTris[0], _cascadeHighDist[0], _cascadeLowTris[0], _cascadeLowDist[0]);

    AddNumTrisSlider(IDC_SLIDER_HIGH_TRIS_UP, iX_Col0, iY_Col0, iX_Col1, iY_Col1,
        _cascadeHighTris[1], _cascadeHighDist[1], _cascadeLowTris[1], _cascadeLowDist[1]);

    const int ba = _baseCascade;
    AddNumTrisSlider(IDC_SLIDER_HIGH_TRIS_BA, iX_Col0, iY_Col0, iX_Col1, iY_Col1,
        _cascadeHighTris[ba], _cascadeHighDist[ba], _cascadeLowTris[ba], _cascadeLowDist[ba]);

    swprintf_s( szTemp, L"Jitter Cen %0.2f", tuneFV8Mgr.GetJitterScale(0) );
    g_SampleUI.AddStatic(IDC_STATIC_JITTER_SCALE_CEN, szTemp, iX_Col0 + 5, iY_Col0 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_JITTER_SCALE_CEN, iX_Col0, iY_Col0 += 25, 140, 24, 0, 100, 
        int(tuneFV8Mgr.GetJitterScale(0) * 100.0f), false );

    swprintf_s( szTemp, L"Jitter Mid %0.2f", tuneFV8Mgr.GetJitterScale(1) );
    g_SampleUI.AddStatic(IDC_STATIC_JITTER_SCALE_MID, szTemp, iX_Col0 + 5, iY_Col0 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_JITTER_SCALE_MID, iX_Col0, iY_Col0 += 25, 140, 24, 0, 100, 
        int(tuneFV8Mgr.GetJitterScale(1) * 100.0f), false );

    swprintf_s( szTemp, L"Jitter Out %0.2f", tuneFV8Mgr.GetJitterScale(2) );
    g_SampleUI.AddStatic(IDC_STATIC_JITTER_SCALE_OUT, szTemp, iX_Col0 + 5, iY_Col0 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_JITTER_SCALE_OUT, iX_Col0, iY_Col0 += 25, 140, 24, 0, 100, 
        int(tuneFV8Mgr.GetJitterScale(2) * 100.0f), false );

    swprintf_s( szTemp, L"Blend Cen %0.2f", tuneFV8Mgr.GetFrameDecay(0) );
    g_SampleUI.AddStatic(IDC_STATIC_FRAME_DECAY_CEN, szTemp, iX_Col1 + 5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_FRAME_DECAY_CEN, iX_Col1, iY_Col1 += 25, 140, 24, 0, 100, 
        int(tuneFV8Mgr.GetFrameDecay(0) * 100.0f), false );

    swprintf_s( szTemp, L"Blend Mid %0.2f", tuneFV8Mgr.GetFrameDecay(1) );
    g_SampleUI.AddStatic(IDC_STATIC_FRAME_DECAY_MID, szTemp, iX_Col1 + 5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_FRAME_DECAY_MID, iX_Col1, iY_Col1 += 25, 140, 24, 0, 100, 
        int(tuneFV8Mgr.GetFrameDecay(1) * 100.0f), false );

    swprintf_s( szTemp, L"Blend Out %0.2f", tuneFV8Mgr.GetFrameDecay(2) );
    g_SampleUI.AddStatic(IDC_STATIC_FRAME_DECAY_OUT, szTemp, iX_Col1 + 5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_FRAME_DECAY_OUT, iX_Col1, iY_Col1 += 25, 140, 24, 0, 100, 
        int(tuneFV8Mgr.GetFrameDecay(2) * 100.0f), false );

    swprintf_s( szTemp, L"Discard Frames %d", 0 );
    g_SampleUI.AddStatic(IDC_STATIC_DRAW_FRAMES, szTemp, iX_Col1 + 5, iY_Col1 += 25, 108, 24 );
    g_SampleUI.AddSlider(IDC_SLIDER_DISCARD_FRAMES, iX_Col1, iY_Col1 += 25, 140, 24, 0, 7, 0, false );

#endif // MAF_NO_HUD

    Cascade::Init();

	g_SampleUI.SetCallback( OnGUIEvent ); 

    InitCameraView();

    InitBounds();
}

int CountDrawnObjects()
{
    int numObj = 0;
    for(int i = 0; i < _maxCascades; ++i)
    {
        RenderCounter& cnt = _cascadeCounts[i];
        if(cnt.Active)
        {
            numObj += cnt.NumObjDrawn;
        }
    }
    return numObj;
}
int CountDrawnTris()
{
    int numTris = 0;
    for(int i = 0; i < _maxCascades; ++i)
    {
        RenderCounter& cnt = _cascadeCounts[i];
        if(cnt.Active)
        {
            numTris += cnt.NumTris;
        }
    }
    return numTris;
}
int CurrentPix()
{
    int numPix = 0;
    if(compositeMode == kCompositeAllHigh)
    {
        const POINT res = nonFv8Mgr.GetResolution(0);
        numPix += res.x * res.y;
    }
    else
    {
        for(int i = 0; i < _maxCascades; ++i)
        {
            if(_cascadeCounts[i].Active)
                numPix += _cascadeCounts[i].NumPix;
        }
    }
    return numPix;
}
void RenderShadowedText(const WCHAR* text, const D3DXCOLOR& fore, const D3DXCOLOR& back)
{
    const POINT basePt = g_pTxtHelper->GetInsertionPoint();
    g_pTxtHelper->SetForegroundColor( back );
    g_pTxtHelper->DrawTextLine(text);

    const POINT nextPt = g_pTxtHelper->GetInsertionPoint();
    g_pTxtHelper->SetInsertionPos(basePt.x + 1, basePt.y - 1);
    g_pTxtHelper->SetForegroundColor( fore );
    g_pTxtHelper->DrawTextLine(text);
    g_pTxtHelper->SetInsertionPos(nextPt.x, nextPt.y);
}
//--------------------------------------------------------------------------------------
// Render the help and statistics text. This function uses the ID3DXFont interface for 
// efficient text rendering.
//--------------------------------------------------------------------------------------
void RenderText()
{
    if(displayStats)
    {
#ifndef MAF_NO_HUD
        g_pTxtHelper->Begin();
        g_pTxtHelper->SetInsertionPos( 5, 5 );
        g_pTxtHelper->SetForegroundColor( D3DXCOLOR( 0.2f, 0.2f, 0.6f, 1.0f ) );
        g_pTxtHelper->DrawTextLine( DXUTGetFrameStats( true /*DXUTIsVsyncEnabled()*/ ) );
        g_pTxtHelper->DrawTextLine( DXUTGetDeviceStats() );
	    WCHAR buff[256];
        for(int i = 0; i < _maxCascades; ++i)
        {
            RenderCounter& cnt = _cascadeCounts[i];
            if(cnt.NumPix)
            {
	            swprintf_s(buff, L"%d: %d tris %d(%d) obj(tot) %d pix %5.1f", 
                    i,
                    cnt.NumTris,
                    cnt.NumObjDrawn, cnt.NumObjTested,
                    cnt.NumPix,
                    cnt.DrawInterval * 1000.0);
	            g_pTxtHelper->DrawTextLine(buff);
            }
        }

	    {
		    swprintf_s(buff, L"MousePos uninitialized, %s cs=%.1f sm=%.1f %s", 
                GazeModeString(),
                g_Camera.GetSpeedGoal(),
                GetSmoothSecs() * 1000.0f,
                smoothTime ? L"SMO" : L"RAW");
	    }
	    g_pTxtHelper->DrawTextLine(buff);

        swprintf_s(buff, L"%.1f %.1f %.1f, %.1f %.1f %.1f %.1f",
            g_Camera.GetEyePt()->x, g_Camera.GetEyePt()->y, g_Camera.GetEyePt()->z,
            g_Camera.GetLookAtPt()->x, g_Camera.GetLookAtPt()->y, g_Camera.GetLookAtPt()->z,
            tuneFV8Mgr.GetPerPixelJitter());
        g_pTxtHelper->DrawTextLine(buff);

        swprintf_s(buff, L"Settings: %s", _settingsLabel);
        RenderShadowedText(buff, D3DXCOLOR( 0.7f, 0.9f, 0.15f, 1.0f ), D3DXCOLOR( 0.1f, 0.1f, 0.1f, 1.0f ));

        int addedLatencyMS = GetAddedLatencyUSecs() / 1000;
        if(addedLatencyMS < 0)
        {
            swprintf_s(buff, L"WARN: Desired latency %d < frame timeX1.5= %.1f",
                latencyMS, GetSmoothSecs() * 1.5f * 1000.0f);
            RenderShadowedText(buff, D3DXCOLOR( 0.9f, 0.3f, 0.15f, 1.0f ), D3DXCOLOR( 0.1f, 0.1f, 0.1f, 1.0f ));
        }
        g_pTxtHelper->End();
#endif // MAF_NO_HUD
    }

}


//--------------------------------------------------------------------------------------
// Reject any D3D11 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D11DeviceAcceptable( const CD3D11EnumAdapterInfo *AdapterInfo, UINT Output, const CD3D11EnumDeviceInfo *DeviceInfo,
                                       DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    return true;
}

void SetZeroLatency()
{
	IDXGIDevice1* pDXGIDev = NULL;
	HRESULT hr = DXUTGetD3D11Device()->QueryInterface( __uuidof( IDXGIDevice1 ), ( LPVOID* )&pDXGIDev );
	if( SUCCEEDED( hr ) && pDXGIDev )
	{
		pDXGIDev->SetMaximumFrameLatency(zeroLatency ? 1 : 0);
		SAFE_RELEASE(pDXGIDev);
	}
}

static BlitEnvMap* _blitEnv = NULL;
static BlitEnvMap* CheckBlitEnv(ID3D11Device* pd3dDevice)
{
    if(!_blitEnv)
    {
        _blitEnv = new BlitEnvMap();
        //_blitEnv->SetPointSample(pd3dDevice, true);
        _blitEnv->Init(pd3dDevice, L"ShaderBlits/BlitEnvMap.fx");

		const BlitQuad::BlitRect dstRect( D3DXVECTOR2(0.0f, 0.0f), D3DXVECTOR2(1.0f, 1.0f) );
		_blitEnv->SetDstRect(dstRect);
    }
    return _blitEnv;
}
static BlitQuad* _blitQuad = NULL;
BlitQuad* CheckBlitQuad(ID3D11Device* pd3dDevice)
{
	if(!_blitQuad)
	{
		_blitQuad = new BlitQuad();
		_blitQuad->Init(pd3dDevice);
	}
	return _blitQuad;
}
static BlitQuad* _cursorBlit = NULL;
static ID3D11ShaderResourceView* _cursorSRV = NULL;
static bool _cursorShow = false;
BlitQuad* CheckCursorBlit(ID3D11Device* pd3dDevice)
{
    if(!_cursorBlit)
    {
        _cursorBlit = new BlitQuad();
        _cursorBlit->SetBlend(BlitQuad::kSrcAlphaPM);
        _cursorBlit->SetTint(D3DXVECTOR4(1.0f, 0.0f, 0.0f, 1.0f));
        if (FAILED(_cursorBlit->Init(pd3dDevice)) )
        {
            assert(false);
            return NULL;
        }
    }
    if(!_cursorSRV)
    {
        if( FAILED(D3DX11CreateShaderResourceViewFromFile( pd3dDevice, L"Media/CursorShow.dds", NULL, NULL, &_cursorSRV, NULL )) )
        {
            assert(false);
            return NULL;
        }
    }
    return _cursorBlit;
}

//--------------------------------------------------------------------------------------
// Create any D3D11 resources that aren't dependent on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D11CreateDevice( ID3D11Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc,
                                     void* pUserContext )
{
    HRESULT hr;

    ID3D11DeviceContext* pd3dImmediateContext = DXUTGetD3D11DeviceContext();
    V_RETURN( g_DialogResourceManager.OnD3D11CreateDevice( pd3dDevice, pd3dImmediateContext ) );
    V_RETURN( g_SettingsDlg.OnD3D11CreateDevice( pd3dDevice ) );
    g_pTxtHelper = new CDXUTTextHelper( pd3dDevice, pd3dImmediateContext, &g_DialogResourceManager, 15 );

    // Create constant buffers
    D3D11_BUFFER_DESC cbDesc;
    ZeroMemory( &cbDesc, sizeof(cbDesc) );
    cbDesc.Usage = D3D11_USAGE_DYNAMIC;
    cbDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    cbDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;

    cbDesc.ByteWidth = sizeof( CB_VS_PER_OBJECT );
    V_RETURN( pd3dDevice->CreateBuffer( &cbDesc, NULL, &g_pcbVSPerObject11 ) );
    DXUT_SetDebugName( g_pcbVSPerObject11, "CB_VS_PER_OBJECT" );

    cbDesc.ByteWidth = sizeof( CB_VS_PER_FRAME );
    V_RETURN( pd3dDevice->CreateBuffer( &cbDesc, NULL, &g_pcbVSPerFrame11 ) );
    DXUT_SetDebugName( g_pcbVSPerFrame11, "CB_VS_PER_FRAME" );

    // Create other render resources here

    tuneFV8Mgr.SetFBWidthAndHeight(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);
    tuneNonFoveated.SetFBWidthAndHeight(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);

    _screenSize.x = pBackBufferSurfaceDesc->Width;
    _screenSize.y = pBackBufferSurfaceDesc->Height;

	MakeModel(false);

	SetZeroLatency();

    DrawBox::Init();

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Create any D3D11 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D11ResizedSwapChain( ID3D11Device* pd3dDevice, IDXGISwapChain* pSwapChain,
                                         const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    V_RETURN( g_DialogResourceManager.OnD3D11ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );
    V_RETURN( g_SettingsDlg.OnD3D11ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );

    // Setup the camera's projection parameters
    const float fAspectRatio = pBackBufferSurfaceDesc->Width / ( FLOAT )pBackBufferSurfaceDesc->Height;
    g_Camera.SetProjParams( D3DX_PI / 3 / fAspectRatio, fAspectRatio, kHither, kYon );
    g_Camera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );
    g_Camera.SetButtonMasks( MOUSE_LEFT_BUTTON, MOUSE_WHEEL, MOUSE_LEFT_BUTTON | MOUSE_MIDDLE_BUTTON | MOUSE_RIGHT_BUTTON);

    const int uiStart = 5;
	const int uiHeight = pBackBufferSurfaceDesc->Height - uiStart;
    const int uiWidth = 400;
    g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width - uiWidth, uiStart );
    g_SampleUI.SetSize( uiWidth, uiHeight );

	ResizeModels();

    ReloadSettings();

    CheckBlitEnv(pd3dDevice);
    CheckBlitQuad(pd3dDevice);
    CheckCursorBlit(pd3dDevice);

    tuneFV8Mgr.SetFBWidthAndHeight(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);
    tuneNonFoveated.SetFBWidthAndHeight(pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);

    _screenSize.x = pBackBufferSurfaceDesc->Width;
    _screenSize.y = pBackBufferSurfaceDesc->Height;

    return S_OK;
}

static void ClearEnvironment(ID3D11DeviceContext* pd3dImmediateContext)
{
    BlitEnvMap* blitEnv = _blitEnv;

    if(blitEnv)
    {
	    UINT numViewports = 10;
	    D3D11_VIEWPORT viewports[10];
	    pd3dImmediateContext->RSGetViewports(&numViewports, viewports);
	    const LONG width = (LONG)(viewports[0].Width + 0.5f);
	    const LONG height = (LONG)(viewports[0].Height + 0.5f);
	    const POINT dims = { width, height };

	    blitEnv->SetDstResolution(dims);

	    ID3D11RenderTargetView* pRTV = NULL; // DXUTGetD3D11RenderTargetView();

        blitEnv->Render(pd3dImmediateContext, pRTV, GridObject::GetEnvironmentMap());
    }
}

static const float ClearColor[4] = { 0.8f, 0.9f, 0.6f, 1.0f };

void RenderInnerLoop(ID3D11DeviceContext* pd3dImmediateContext, const D3DXMATRIX& camToNDC, int iLayer)
{
    RenderCounter& cnt = _cascadeCounts[iLayer];
    cnt.Clear();

    if(!(_hackSkipMask & (1 << iLayer)))
    {
        CameraSave cameraCache;
        cameraCache.Save(g_Camera);
        g_Camera.SetProjection(&camToNDC);
        // Temp bit
        g_Camera.SetLastWorldToNDC(fv8Mgr.GetLastWorldToNDC());

        if(textureBackground)
        {
            ClearEnvironment(pd3dImmediateContext);
        }
        GridObject::SetHighNumTris(_cascadeHighTris[iLayer]);
        GridObject::SetLowNumTris(_cascadeLowTris[iLayer]);
        GridObject::SetHighDistance(_cascadeHighDist[iLayer]);
        GridObject::SetLowDistance(_cascadeLowDist[iLayer]);
    
        GridObject::SetFrameDecay(tuneFV8Mgr.GetFrameDecay(iLayer));

        const POINT size = fv8Mgr.GetResolution(iLayer);
        GridObject::SetWindowSize(D3DXVECTOR2(
            FLOAT(size.x), 
            FLOAT(size.y)));

        cnt += RenderModels();
        cnt.NumPix += size.x * size.y;
        cnt.SetDrawTime();

        cameraCache.Restore(g_Camera);
    }
}
void TestBoxed(ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext)
{
    if(fv8Mgr.NeedInit())
    {
        fv8Mgr.Init(pd3dDevice);
    }

    // Send this frame's camera information.
    fv8Mgr.SetCamera(*g_Camera.GetViewMatrix(), *g_Camera.GetProjMatrix(), g_Camera.GetNearClip(), g_Camera.GetFarClip());
    // Send the gaze position within the window.
    fv8Mgr.SetFoveaCenter(GetGazePosition());

    // Loop over the eccentricity layers.
    for(int iLayer = 0; iLayer < fv8Mgr.GetNumLayers(); ++iLayer)
    {
        // Different layers refresh at different rates. Check if this layer is due.
        if(fv8Mgr.ShouldRefresh(iLayer, DXUTGetElapsedTime()))
        {
            // Pushing this layer as a render target returns a modified projection matrix.
            D3DXMATRIX cam2NDC = fv8Mgr.PushLayer(iLayer, pd3dImmediateContext);

            // Clear the screen.
            const float* const clearColor = textureBackground ? NULL : ClearColor;
            fv8Mgr.Clear(pd3dImmediateContext, clearColor);

            // Render the scene as normal, using the modified projection matrix. This
            // presumes the cull volume is generated from the camera transforms. Also,
            // LOD and effects quality can be reduced for increasing layer index.
            RenderInnerLoop(pd3dImmediateContext, cam2NDC, iLayer);

            // Finish this layer, resetting state to before the PushLayer() above.
            fv8Mgr.PopLayer(pd3dImmediateContext);
        }
    }
    // Now composite the layers into the destination render target (the back buffer here).
	ID3D11RenderTargetView* pRTV = DXUTGetD3D11RenderTargetView();
    fv8Mgr.FinishFrame(pd3dImmediateContext, pRTV);
}

void NewSet()
{
    ReloadSettings();

    if(_settingsReset)
    {
        g_Camera.Reset();
        InitCameraView();
    }
}
void NextSet()
{
    if(++_settingsCurrent >= _settings.size())
    {
        --_settingsLoops;
        _settingsCurrent = 0;
    }
    NewSet();
}
void PrevSet()
{
    if(--_settingsCurrent >= _settings.size())
    {
        _settingsCurrent = _settings.size() - 1;
    }
    NewSet();
}
void SetSetting(size_t idx)
{
    assert(idx >= 0 && idx < int(_settings.size()));
    if(idx != _settingsCurrent)
    {
        _settingsCurrent = idx;
        NewSet();
        ClearCounts();
    }
}
void IShowCursor(ID3D11DeviceContext* pd3dImmediateContext)
{
    BlitQuad* bq = _cursorBlit;
    if(bq && _cursorShow)
    {
        UINT numViewports = 10;
        D3D11_VIEWPORT viewports[10];
        pd3dImmediateContext->RSGetViewports(&numViewports, viewports);
        const float width = viewports[0].Width;
        const float height = viewports[0].Height;

        const float blitWidth = 60 / width;
        const float blitHeight = 60 / height;
        const POINT gazePos = GetGazePosition();
        const D3DXVECTOR2 center(gazePos.x / (float)width, 1.0f - gazePos.y / (float)height);
        const D3DXVECTOR2 dstMin(center.x - blitWidth * 0.5f, center.y - blitHeight * 0.5f);
        const D3DXVECTOR2 dstMax(center.x + blitWidth * 0.5f, center.y + blitHeight * 0.5f);

        const POINT dstRes = { int(width), int(height) };
        bq->SetDstResolution(dstRes);
        bq->SetSrcRect(BlitQuad::BlitRect(0.0f, 0.0f, 1.0f, 1.0f));
        bq->SetDstRect(BlitQuad::BlitRect(dstMin, dstMax));

        bq->Render(pd3dImmediateContext, NULL, _cursorSRV);
    }
}

static float GetBadness()
{
    const int kBadBufferLen = 120;
    static int badBuffer[kBadBufferLen] = { 0 };
    static int badIndex = 0;
    static int badCount = 0;

    if(badBuffer[badIndex])
    {
        assert(badCount > 0);
        --badCount;
    }
    badBuffer[badIndex] = !IsGazeDataValid();
    if(badBuffer[badIndex])
    {
        ++badCount;
    }
    if(++badIndex >= kBadBufferLen)
        badIndex = 0;

    return float(badCount) / float(kBadBufferLen);
}
static void ShowGazeStatus(ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext)
{
    static D3DXVECTOR4 tint = D3DXVECTOR4(1.0f, 1.0f, 0.0f, 0.0f);
    const float kMinBadness = 0.25f;
    const float kFadeRate = 1.0f / 0.5f / 120.0f; // 1/seconds to fade out   * seconds/frame  (== 1/fps)
    const float badNess = GetBadness();
    if(badNess > kMinBadness)
    {
        tint.w += kFadeRate;
        if(tint.w > 1.0f)
            tint.w = 1.0f;

        const float yellow = 1.0f - (badNess - kMinBadness) / (1.0f - kMinBadness);
        tint.y = yellow;
    }
    else if(tint.w > 0)
    {
        tint.y = 1;
        tint.w -= kFadeRate;
        if(tint.w < 0)
            tint.w = 0;
    }
    if(tint.w > 0)
    {
        static BlitQuad* quad = NULL;
        if(!quad)
        {
            quad = new BlitQuad();
            quad->SetBlend(BlitQuad::Blend::kSrcAlpha);
            quad->SetDstRect(BlitQuad::BlitRect(0.25f, 0.15f, 0.75f, 0.175f));
            const POINT res = { int(_screenSize.x), int(_screenSize.y) };
            quad->SetDstResolution(res);
            quad->Init(pd3dDevice, L"ShaderBlits\\BlitQuad.fx", "Flat");
        }
        quad->SetTint(tint);
        quad->Render(pd3dImmediateContext, DXUTGetD3D11RenderTargetView(), NULL);
    }
}


//--------------------------------------------------------------------------------------
// Render the scene using the D3D11 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D11FrameRender( ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext, double fTime,
                                 float fElapsedTime, void* pUserContext )
{
    // If the settings dialog is being shown, then render it instead of rendering the app's scene
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }       

    static const float ClearColor[4] = { 0.00f, 0.00f, 0.00f, 1.0f };
    ID3D11RenderTargetView* pRTV = DXUTGetD3D11RenderTargetView();
    pd3dImmediateContext->ClearRenderTargetView( pRTV, ClearColor );

    // Clear the depth stencil
    ID3D11DepthStencilView* pDSV = DXUTGetD3D11DepthStencilView();
    pd3dImmediateContext->ClearDepthStencilView( pDSV, D3D11_CLEAR_DEPTH, 1.0, 0 );

    // Get the projection & view matrix from the camera class
    D3DXMATRIX mView = *g_Camera.GetViewMatrix();
    D3DXMATRIX mProj = *g_Camera.GetProjMatrix();
    D3DXMATRIX mWorldViewProjection = mView * mProj;

    // Set the constant buffers
    HRESULT hr;
    D3D11_MAPPED_SUBRESOURCE MappedResource;
    V( pd3dImmediateContext->Map( g_pcbVSPerFrame11, 0, D3D11_MAP_WRITE_DISCARD, 0, &MappedResource ) );
    CB_VS_PER_FRAME* pVSPerFrame = ( CB_VS_PER_FRAME* )MappedResource.pData;
    pVSPerFrame->m_vLightDir = D3DXVECTOR3( 0,0.707f,-0.707f );
    pVSPerFrame->m_fTime = (float)fTime;
    pVSPerFrame->m_LightDiffuse = D3DXVECTOR4( 1.f, 1.f, 1.f, 1.f );
    pd3dImmediateContext->Unmap( g_pcbVSPerFrame11, 0 );
    pd3dImmediateContext->VSSetConstantBuffers( 1, 1, &g_pcbVSPerFrame11 );

    V( pd3dImmediateContext->Map( g_pcbVSPerObject11, 0, D3D11_MAP_WRITE_DISCARD, 0, &MappedResource ) );
    CB_VS_PER_OBJECT* pVSPerObject = ( CB_VS_PER_OBJECT* )MappedResource.pData;
    D3DXMatrixTranspose( &pVSPerObject->m_mWorldViewProjection, &mWorldViewProjection );
	D3DXMatrixIdentity(&pVSPerObject->m_mWorld);
    pVSPerObject->m_MaterialAmbientColor = D3DXVECTOR4( 0.3f, 0.3f, 0.3f, 1.0f );
    pVSPerObject->m_MaterialDiffuseColor = D3DXVECTOR4( 0.7f, 0.7f, 0.7f, 1.0f );
    pd3dImmediateContext->Unmap( g_pcbVSPerObject11, 0 );
    pd3dImmediateContext->VSSetConstantBuffers( 0, 1, &g_pcbVSPerObject11 );

	{
		ReadGazePos(_gazePos);
	}

    PreRenderCounts();

    {
        switch(compositeMode)
        {
        case kCompositeBoxed:
            TestBoxed(pd3dDevice, pd3dImmediateContext);
            break;
        case kCompositeAllHigh:
            RenderModelsHigh();
            break;


        default:
            break;
        }

    }

    DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"HUD / Stats" );
    //g_HUD.OnRender( fElapsedTime );
    g_SampleUI.OnRender( fElapsedTime );
    RenderText();
    DXUT_EndPerfEvent();

    IShowCursor(pd3dImmediateContext);

    ShowGazeStatus(pd3dDevice, pd3dImmediateContext);
}

//--------------------------------------------------------------------------------------
// Release D3D11 resources created in OnD3D11ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D11ReleasingSwapChain( void* pUserContext )
{
    g_DialogResourceManager.OnD3D11ReleasingSwapChain();
}


//--------------------------------------------------------------------------------------
// Release D3D11 resources created in OnD3D11CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D11DestroyDevice( void* pUserContext )
{
    SAFE_DELETE(_cursorBlit);
    SAFE_RELEASE(_cursorSRV);
	SAFE_DELETE(_blitQuad);
    SAFE_DELETE(_blitEnv);
    SAFE_RELEASE(pReprojectionCB);

	CleanupModels();
    DrawBox::Cleanup();

    g_DialogResourceManager.OnD3D11DestroyDevice();
    g_SettingsDlg.OnD3D11DestroyDevice();
    DXUTGetGlobalResourceCache().OnDestroyDevice();
    SAFE_DELETE( g_pTxtHelper );

    fv8Mgr.Release();

    // Delete additional render resources here...

    SAFE_RELEASE( g_pcbVSPerObject11 );
    SAFE_RELEASE( g_pcbVSPerFrame11 );

}


//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D11 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
    if( pDeviceSettings->ver == DXUT_D3D9_DEVICE )
    {
        IDirect3D9* pD3D = DXUTGetD3D9Object();
        D3DCAPS9 Caps;
        pD3D->GetDeviceCaps( pDeviceSettings->d3d9.AdapterOrdinal, pDeviceSettings->d3d9.DeviceType, &Caps );

        // If device doesn't support HW T&L or doesn't support 1.1 vertex shaders in HW 
        // then switch to SWVP.
        if( ( Caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ) == 0 ||
            Caps.VertexShaderVersion < D3DVS_VERSION( 1, 1 ) )
        {
            pDeviceSettings->d3d9.BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
        }

        // Debugging vertex shaders requires either REF or software vertex processing 
        // and debugging pixel shaders requires REF.  
#ifdef DEBUG_VS
        if( pDeviceSettings->d3d9.DeviceType != D3DDEVTYPE_REF )
        {
            pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_HARDWARE_VERTEXPROCESSING;
            pDeviceSettings->d3d9.BehaviorFlags &= ~D3DCREATE_PUREDEVICE;
            pDeviceSettings->d3d9.BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
        }
#endif
#ifdef DEBUG_PS
        pDeviceSettings->d3d9.DeviceType = D3DDEVTYPE_REF;
#endif
    }

    // For the first device created if its a REF device, optionally display a warning dialog box
    static bool s_bFirstTime = true;
    if( s_bFirstTime )
    {
        s_bFirstTime = false;
        if( ( DXUT_D3D9_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF ) ||
            ( DXUT_D3D11_DEVICE == pDeviceSettings->ver &&
            pDeviceSettings->d3d11.DriverType == D3D_DRIVER_TYPE_REFERENCE ) )
        {
            DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
        }

    }

    return true;
}

//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
    AddSmoothSecs(fElapsedTime);

    if(smoothTime)
    {
        fElapsedTime = GetSmoothSecs();
    }

    // Update the camera's position based on user input 
    MoveCamera( fElapsedTime );

	MoveModels(fElapsedTime);
}

//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing,
                          void* pUserContext )
{

    // Pass messages to dialog resource manager calls so GUI state is updated correctly
    *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass messages to settings dialog if its active
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    //*pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    //if( *pbNoFurtherProcessing )
    //    return 0;
    *pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
    {
        guiActive = true;
        return 0;
    }

    if(uMsg == WM_LBUTTONUP)
        guiActive = false;

    // Pass all remaining windows messages to camera so it can respond to user input
    g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}

class TrickState
{
public:
    TrickState()
        :   disabled(false)
    {
    }
    void Toggle(TunableFV8Mgr& tune)
    {
        if(disabled)
        {
            Restore(tune);
        }
        else
        {
            CaptureAndDisable(tune);
        }
    }
    void CaptureAndDisable(TunableFV8Mgr& tune)
    {
        for(int i = 0; i < 3; ++i)
        {
            m_Interval[i] = tune.GetRenderInterval(i);
            m_RefreshSecs[i] = tune.GetRefreshSecs(i);

            tune.SetRenderInterval(i, FauxV8Mgr::kRenderAlways);
            tune.SetRefreshSecs(i, 0.0f);
        }
        m_MSAA = tune.GetMSAA();            
        tune.SetMSAA(0);
        m_JitterEnable = Cascade::GetJitterEnabled();
        Cascade::SetJitterEnabled(false);
        m_ReprojEnable = tune.GetDebugReprojEnable();
        tune.SetDebugReprojEnable(false);
        disabled = true;
    }
    void Restore(TunableFV8Mgr& tune)
    {
        for(int i = 0; i < 3; ++i)
        {
            tune.SetRenderInterval(i, m_Interval[i]);
            tune.SetRefreshSecs(i, m_RefreshSecs[i]);
        }
        tune.SetMSAA(m_MSAA);
        Cascade::SetJitterEnabled(m_JitterEnable);
        tune.SetDebugReprojEnable(m_ReprojEnable);
        disabled = false;
    }

private:

    FauxV8Mgr::eRenderInterval  m_Interval[3];
    float                       m_RefreshSecs[3];
    int     m_MSAA;
    bool    m_JitterEnable;
    bool    m_ReprojEnable;

    bool disabled;
};


float boundRate = 0.01f;
//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
	switch(nChar)
	{
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
        if(bKeyDown)
        {
            const int digit = nChar - '0';
            if(DXUTIsKeyDown(VK_SHIFT))
            { 
                if(digit)
                    tuneFV8Mgr.SetDebugShowLevels(BlitComposite::kShowDarken, digit);
                else
                    tuneFV8Mgr.SetDebugShowLevels(BlitComposite::kShowNone, digit);
            }
            else if(DXUTIsKeyDown(VK_CONTROL))
            {
                static CompositeMode saveCompMode = compositeMode;
                if(digit)
                    saveCompMode = compositeMode;
                tuneFV8Mgr.SetDebugDiscardMask(digit);
                compositeMode = digit ? kCompositeBoxed : saveCompMode;
                g_SampleUI.GetComboBox(IDC_COMBOBOX_COMPOSITE)->SetSelectedByIndex(compositeMode);
            }
        }
        break;
	//case 'X':
	//	if(bKeyDown)
	//		rotRate.x = AdvanceRot(rotRate.x);
	//	break;
	case 'X':
		if(bKeyDown)
        {
            float blend = tuneFV8Mgr.GetCompositeBlendStart();
            if(blend < 1)
                blend = 1;
            else
                blend = 0.6f;
            tuneFV8Mgr.SetCompositeBlendStart(blend);
        }
		break;
	case 'Y':
		if(bKeyDown)
			rotRate.y = AdvanceRot(rotRate.y);
		break;
	case 'Z':
		if(bKeyDown)
			rotRate.z = AdvanceRot(rotRate.z);
		break;
	case 'U':
		if(bKeyDown)
		{
			if(upDown != 0)
				upDown = 0;
			else
				upDown = upDownSpeed;
		}
		break;

    case 'P':
        if(bKeyDown)
        {
            procTerrain = !procTerrain;
            RebuildScape();
        }
        break;

    case 'J':
        if(bKeyDown)
        {
            if(!DXUTIsKeyDown(VK_SHIFT))
            {
                Cascade::SetJitterEnabled(!Cascade::GetJitterEnabled());
            }
            else
            {
                tuneFV8Mgr.SetPerPixelJitter(tuneFV8Mgr.GetPerPixelJitter() > 0 ? 0.0f : 1.0f);
            }
        }
        break;
	case 'W':
		if(bKeyDown)
		{
			GridObject::SetWireFrame(!GridObject::GetWireFrame());
            CDXUTCheckBox* checkBox = g_SampleUI.GetCheckBox(IDC_CHECKBOX_WIREFRAME);
            if(checkBox)
                checkBox->SetChecked(GridObject::GetWireFrame());
		}
		break;
	case 'B':
		if(bKeyDown)
		{
			GridObject::SetShowBounds(!GridObject::GetShowBounds());
            CDXUTCheckBox* checkBox = g_SampleUI.GetCheckBox(IDC_CHECKBOX_SHOWBOUNDS);
            if(checkBox)
                checkBox->SetChecked(GridObject::GetShowBounds());
		}
		break;
    case 'C':
        if(bKeyDown)
        {
            compositeMode = CompositeMode(compositeMode + 1);
            if(compositeMode >= 2) // kNumCompositeModes)
                compositeMode = CompositeMode(0);
            g_SampleUI.GetComboBox(IDC_COMBOBOX_COMPOSITE)->SetSelectedByIndex(compositeMode);
            ClearCounts();
        }
        break;
    case 'M':
        if(bKeyDown)
        {
            if(tuneFV8Mgr.GetMSAA())
                tuneFV8Mgr.SetMSAA(0);
            else
                tuneFV8Mgr.SetMSAA(4);
        }
        break;
    case 'F':
        if(bKeyDown)
        {
            ++gazeMode;
        }
        break;
    case 'D':
        if(bKeyDown)
        {
            const BlitComposite::ShowLevelMode mode = BlitComposite::GetShowLevelMode() == BlitComposite::kShowNone
                            ? BlitComposite::kShowDarken
                            : BlitComposite::kShowNone;
            tuneFV8Mgr.SetDebugShowLevels(mode);
        }
        break;
	case 'E':
		if(bKeyDown)
		{
            const BlitComposite::ShowLevelMode mode = BlitComposite::GetShowLevelMode() == BlitComposite::kShowNone
                            ? BlitComposite::kShowColor
                            : BlitComposite::kShowNone;
            tuneFV8Mgr.SetDebugShowLevels(mode);
		}
		break;

    case VK_OEM_PERIOD:
        if(bKeyDown)
        {
            GridObject::SetUseLast(!GridObject::GetUseLast());
        }
        break;

	//case 'L':
	//	if(bKeyDown)
	//	{
	//		GridObject::SetLights(!GridObject::GetLights());
	//	}
	//	break;
	//case 'S':
	//	if(bKeyDown)
	//	{
	//		GridObject::SetSpecular(!GridObject::GetSpecular());
	//	}
	//	break;
    //case 'S':
    //    if(bKeyDown)
    //    {
    //        smoothTime = !smoothTime;
    //    }
    //    break;
    case 'S':
        if(bKeyDown)
        {
            _cursorShow = !_cursorShow;
        }
        break;

	case 'N':
		if(bKeyDown)
		{
			MakeModel(true);
		}
		break;
    case 'T':
        if(bKeyDown)
        {
            static TrickState trickState;
            trickState.Toggle(tuneFV8Mgr);
        }
        break;

    case 'R':
        if(bKeyDown)
        {
            bool reprojEnabled = !tuneFV8Mgr.GetDebugReprojEnable();
            g_SampleUI.GetSlider(IDC_SLIDER_FRAME_DECAY_CEN)->SetEnabled(reprojEnabled);
            g_SampleUI.GetSlider(IDC_SLIDER_FRAME_DECAY_MID)->SetEnabled(reprojEnabled);
            g_SampleUI.GetSlider(IDC_SLIDER_FRAME_DECAY_OUT)->SetEnabled(reprojEnabled);
            tuneFV8Mgr.SetDebugReprojEnable(reprojEnabled);
        }
        break;

    case VK_F1:
        if(bKeyDown)
        {
            displayStats = !displayStats;
        }
        break;
    case VK_F2:
        if(bKeyDown)
        {
            g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() );
        }
        break;

	case VK_SHIFT:
		if(bKeyDown)
		{
			const float kFastSpin = 5.0f;
			if(spinBoost != kFastSpin)
			{
				spinBoost = kFastSpin;
				rotRate *= spinBoost;
			}
		}
		else
		{
			if(spinBoost != 1.0f)
			{
				rotRate /= spinBoost;
				spinBoost = 1.0f;
			}
		}
		break;

    case VK_RETURN:
		if(bKeyDown)
		{
			rotRate = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
			rotAccum = D3DXVECTOR3(float(-M_PI_2), 0.0f, 0.0f);
			g_Camera.Reset();
            InitCameraView();
			//CleanupModels();
			//MakeModel(false);
		}
		break;

    case VK_NEXT:
        if(bKeyDown)
        {
            NextSet();

        }
        break;
    case VK_PRIOR:
        if(bKeyDown)
        {
            PrevSet();
        }
        break;
	}
}

static void HandleLODSlider(int nControlID, CDXUTControl* pControl)
{
    const int baseID = nControlID - IDC_SLIDER_HIGH_TRIS_IN;
    const int iCasc = baseID / 4 > _maxUpper ? _baseCascade : baseID / 4;
    const int iParm = baseID % 4;

    if(iParm & 1) // it's a distance
    {
		float logDist = ((CDXUTSlider*)pControl)->GetValue() / 100.0f;
        float dist = FromLogScale(logDist, kDistanceLogScale);
        if(iParm < 2)
		    _cascadeHighDist[iCasc] = dist;
        else
            _cascadeLowDist[iCasc] = dist;
        SetLODSliderLabel(nControlID, dist);
    }
    else // it's a number of tris
    {
		float logBaseTris = ((CDXUTSlider*)pControl)->GetValue() / 100.0f;
        float baseTris = FromLogScale(logBaseTris, kNumTrisLogScale);
        if(iParm < 2)
		    _cascadeHighTris[iCasc] = baseTris;
        else
            _cascadeLowTris[iCasc] = baseTris;
        SetLODSliderLabel(nControlID, baseTris);
    }
}

//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    WCHAR szTemp[256];

	switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN:
            DXUTToggleFullScreen();
            break;
        case IDC_TOGGLEREF:
            DXUTToggleREF();
            break;
        case IDC_CHANGEDEVICE:
            g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() );
            break;

        case IDC_SLIDER_COMP_FALLOFF:
        {
			int percent = ((CDXUTSlider*)pControl)->GetValue();
            tuneFV8Mgr.SetCompositeBlendStart(percent / 100.0f);

            swprintf_s( szTemp, L"Composite decay %d", percent );
			g_SampleUI.GetStatic( IDC_STATIC_COMP_FALLOFF)->SetText( szTemp );
        }
        break;

        case IDC_SLIDER_MSAA:
        {
            int logMsaa = ((CDXUTSlider*)pControl)->GetValue();
            int msaa = 1 << logMsaa;
            GridObject::SetMSAA(msaa);
            tuneFV8Mgr.SetMSAA(msaa);
            tuneNonFoveated.SetMSAA(msaa);

			swprintf_s( szTemp, L"msaa %d", msaa);
			g_SampleUI.GetStatic( IDC_STATIC_MSAA)->SetText( szTemp );
        }
        break;

        case IDC_SLIDER_MODELSX:
        {
            numModels.x = numModels.y = ((CDXUTSlider*)pControl)->GetValue();

			swprintf_s( szTemp, L"Num Models %d", numModels.x * numModels.y);
			g_SampleUI.GetStatic( IDC_STATIC_MODELSX)->SetText( szTemp );
            MakeModel(false);
        }
        break;

        case IDC_SLIDER_MODEL_SPACING:
        {
            modelSpacing = float(((CDXUTSlider*)pControl)->GetValue()) / 10.0f;

            swprintf_s( szTemp, L"Model spacing %.2f", modelSpacing);
			g_SampleUI.GetStatic( IDC_STATIC_MODEL_SPACING)->SetText( szTemp );
            MakeModel(false);
        }
        break;

        case IDC_SLIDER_DISCARD_FRAMES:
        {
            int discardFrames = ((CDXUTSlider*)pControl)->GetValue();
            tuneFV8Mgr.SetDebugDiscardMask(discardFrames);
            swprintf_s( szTemp, L"Discard Frames %d", discardFrames& 0x7);
            g_SampleUI.GetStatic( IDC_STATIC_DRAW_FRAMES)->SetText( szTemp );
        }
        break;

        case IDC_SLIDER_JITTER_SCALE_CEN:
        {
            float f = float(((CDXUTSlider*)pControl)->GetValue()) / 100.0f;
			swprintf_s( szTemp, L"Jitter Cen %.2f", f);
			g_SampleUI.GetStatic( IDC_STATIC_JITTER_SCALE_CEN)->SetText( szTemp );
            tuneFV8Mgr.SetJitterScale(0, f);
        }
        break;

        case IDC_SLIDER_JITTER_SCALE_MID:
        {
            float f = float(((CDXUTSlider*)pControl)->GetValue()) / 100.0f;
			swprintf_s( szTemp, L"Jitter Mid %.2f", f);
			g_SampleUI.GetStatic( IDC_STATIC_JITTER_SCALE_MID)->SetText( szTemp );
            tuneFV8Mgr.SetJitterScale(1, f);
        }
        break;

        case IDC_SLIDER_JITTER_SCALE_OUT:
        {
            float f = float(((CDXUTSlider*)pControl)->GetValue()) / 100.0f;
			swprintf_s( szTemp, L"Jitter Out %.2f", f);
			g_SampleUI.GetStatic( IDC_STATIC_JITTER_SCALE_OUT)->SetText( szTemp );
            tuneFV8Mgr.SetJitterScale(2, f);
        }
        break;

        case IDC_SLIDER_FRAME_DECAY_CEN:
        {
            float f = float(((CDXUTSlider*)pControl)->GetValue()) / 100.0f;
			swprintf_s( szTemp, L"Blend Cen %.2f", f);
			g_SampleUI.GetStatic( IDC_STATIC_FRAME_DECAY_CEN)->SetText( szTemp );
            tuneFV8Mgr.SetFrameDecay(0, f);
        }
        break;

        case IDC_SLIDER_FRAME_DECAY_MID:
        {
            float f = float(((CDXUTSlider*)pControl)->GetValue()) / 100.0f;
			swprintf_s( szTemp, L"Blend Mid %.2f", f);
			g_SampleUI.GetStatic( IDC_STATIC_FRAME_DECAY_MID)->SetText( szTemp );
            tuneFV8Mgr.SetFrameDecay(1, f);
        }
        break;

        case IDC_SLIDER_FRAME_DECAY_OUT:
        {
            float f = float(((CDXUTSlider*)pControl)->GetValue()) / 100.0f;
			swprintf_s( szTemp, L"Blend Out %.2f", f);
			g_SampleUI.GetStatic( IDC_STATIC_FRAME_DECAY_OUT)->SetText( szTemp );
            tuneFV8Mgr.SetFrameDecay(2, f);
        }
        break;

        case IDC_SLIDER_INNER_RES:
        {
            int res = ((CDXUTSlider*)pControl)->GetValue();
            tuneFV8Mgr.SetSize(0, res);

            swprintf_s( szTemp, L"Inner resolution %d", tuneFV8Mgr.GetSize(0) );
			g_SampleUI.GetStatic( IDC_STATIC_INNER_RES)->SetText( szTemp );
        }
        break;
        case IDC_SLIDER_MID_RES:
        {
            int res = ((CDXUTSlider*)pControl)->GetValue();
            tuneFV8Mgr.SetSize(1, res);

            swprintf_s( szTemp, L"Mid resolution %d", tuneFV8Mgr.GetSize(1) );
			g_SampleUI.GetStatic( IDC_STATIC_MID_RES)->SetText( szTemp );
        }
        break;
        case IDC_SLIDER_BASE_RES:
        {
            int res = ((CDXUTSlider*)pControl)->GetValue();
            tuneFV8Mgr.SetSize(2, res);

            swprintf_s( szTemp, L"Back resolution %d", tuneFV8Mgr.GetSize(2) );
			g_SampleUI.GetStatic( IDC_STATIC_BASE_RES)->SetText( szTemp );
        }
        break;
        case IDC_SLIDER_LATENCY:
        {
            int ms = ((CDXUTSlider*)pControl)->GetValue();
            latencyMS = ms;
            swprintf_s( szTemp, L"Latency ms %d", ms);
			g_SampleUI.GetStatic( IDC_STATIC_LATENCY)->SetText( szTemp );
        }
        break;
        case IDC_SLIDER_HIGH_TRIS_IN:
        case IDC_SLIDER_HIGH_DIST_IN:
        case IDC_SLIDER_LOW_TRIS_IN:
        case IDC_SLIDER_LOW_DIST_IN:
        case IDC_SLIDER_HIGH_TRIS_UP:
        case IDC_SLIDER_HIGH_DIST_UP:
        case IDC_SLIDER_LOW_TRIS_UP:
        case IDC_SLIDER_LOW_DIST_UP:
        case IDC_SLIDER_HIGH_TRIS_BA:
        case IDC_SLIDER_HIGH_DIST_BA:
        case IDC_SLIDER_LOW_TRIS_BA:
        case IDC_SLIDER_LOW_DIST_BA:
        {
            HandleLODSlider(nControlID, pControl);
            CDXUTDialog::ClearFocus();
		}
        break;

        case IDC_CHECKBOX_USE_LODS:
            useLOD = ((CDXUTCheckBox*)pControl)->GetChecked();
            break;

        case IDC_CHECKBOX_SHOW_LEVELS:
            {
                BlitComposite::ShowLevelMode mode = ((CDXUTCheckBox*)pControl)->GetChecked()
                        ? BlitComposite::kShowDarken
                        : BlitComposite::kShowNone;
                tuneFV8Mgr.SetDebugShowLevels(mode);
            }
            break;

        case IDC_CHECKBOX_LIGHTS:
            GridObject::SetLights(((CDXUTCheckBox*)pControl)->GetChecked());
            break;

        case IDC_CHECKBOX_SPECULAR:
            GridObject::SetSpecular(((CDXUTCheckBox*)pControl)->GetChecked());
            break;

        case IDC_CHECKBOX_ENVIRONMENT:
            GridObject::SetEnvMap(((CDXUTCheckBox*)pControl)->GetChecked());
            break;

        case IDC_CHECKBOX_BACKGROUND:
            textureBackground = ((CDXUTCheckBox*)pControl)->GetChecked();
            break;

        case IDC_CHECKBOX_MSBASE:
            tuneFV8Mgr.SetMSRead(2, ((CDXUTCheckBox*)pControl)->GetChecked());
            break;
        case IDC_CHECKBOX_MSMID:
            tuneFV8Mgr.SetMSRead(1, ((CDXUTCheckBox*)pControl)->GetChecked());
            break;
        case IDC_CHECKBOX_MSINNER:
            tuneFV8Mgr.SetMSRead(0, ((CDXUTCheckBox*)pControl)->GetChecked());
            break;
        case IDC_CHECKBOX_MSAAINNER:
            {
                const bool msaaInner = ((CDXUTCheckBox*)pControl)->GetChecked();;
                tuneFV8Mgr.SetMSAAInner(msaaInner ? tuneFV8Mgr.GetMSAA() : 0);
            }
            break;
        case IDC_CHECKBOX_USE_MID:
            {
                bool useMid = ((CDXUTCheckBox*)pControl)->GetChecked();;
                tuneFV8Mgr.SetUseMiddle(useMid);
            }
            break;

        case IDC_CHECKBOX_WIREFRAME:
            GridObject::SetWireFrame(((CDXUTCheckBox*)pControl)->GetChecked());
            break;

        case IDC_CHECKBOX_SHOWBOUNDS:
            GridObject::SetShowBounds(((CDXUTCheckBox*)pControl)->GetChecked());
            break;

        case IDC_COMBOBOX_COMPOSITE:
            compositeMode = CompositeMode(g_SampleUI.GetComboBox( IDC_COMBOBOX_COMPOSITE )->GetSelectedIndex());
            break;

        case IDC_COMBOBOX_CAMERA_MODE:
            {
                const int idx = g_SampleUI.GetComboBox( IDC_COMBOBOX_CAMERA_MODE )->GetSelectedIndex();
                const CameraModeSelect::eCameraModes mode = CameraModeSelect::eCameraModes(idx);
                g_Camera.SetFlyMode(mode);
            }
            break;
    }
}

