
#ifndef DRAW_BOX_H
#define DRAW_BOX_H

class DrawBox
{
public:
    static void Render(const D3DXVECTOR3* const pts, const D3DXMATRIX& l2w) { Instance().InternalRender(pts, l2w); }

    static HRESULT Init() { return Instance().InternalInit(); }
    static void Cleanup() { Instance().InternalCleanup(); }
private:
    static DrawBox& Instance()
    {
        static DrawBox inst;
        return inst;
    }

    void InternalRender(const D3DXVECTOR3* const pts, const D3DXMATRIX& l2w);

    HRESULT InternalInit();
    void InternalCleanup();

private:
    void SetState(ID3D11DeviceContext* pImmediateContext);
    void SetConstants(ID3D11DeviceContext* pImmediateContext, const D3DXMATRIX& l2w);
    void SetVerts(const D3DXVECTOR3* const pts);

    HRESULT MakeShaders();
    HRESULT MakeGeometry();
    HRESULT MakeState();
    HRESULT MakeConstants();

    DrawBox();
    virtual ~DrawBox();

private:
    ID3D11RasterizerState*   s_pRasterizerState;

    ID3D11VertexShader*      s_pVertexShader;
    ID3D11PixelShader*       s_pPixelShader;
    ID3D11InputLayout*       s_pVertexLayout;
    ID3D11Buffer*            s_pVertexBuffer;
    ID3D11Buffer*            s_pIndexBuffer;
    ID3D11Buffer*            s_pConstantBuffer;

};

#endif //  DRAW_BOX_H
