
#include <windows.h>
#include <d3dx11.h>
#include <D3DX10math.h>
#include <assert.h>

#include "LODObject.h"

int LODObject::currentLOD = 0;

static inline void SafeDelete(Object*& obj)
{
    if(obj)
    {
        delete obj;
        obj = NULL;
    }
}
LODObject::LODObject()
{
}

LODObject::~LODObject()
{
    for(LODIter iter = lods.begin(); iter != lods.end(); ++iter)
    {
        Object* lod = *iter;
        SafeDelete(lod);
    }
}

HRESULT LODObject::Init()
{
    HRESULT ret = S_OK;
    for(LODIter iter = lods.begin(); iter != lods.end(); ++iter)
    {
        Object* lod = *iter;
        if(lod)
        {
            if(FAILED(ret = lod->Init()))
                return ret;
        }
    }
    return ret;
}

void LODObject::OnResize()
{
    for(LODIter iter = lods.begin(); iter != lods.end(); ++iter)
    {
        Object* lod = *iter;
        if(lod)
        {
            lod->OnResize();
        }
    }
}

void LODObject::Cleanup()
{
    for(LODIter iter = lods.begin(); iter != lods.end(); ++iter)
    {
        Object* lod = *iter;
        if(lod)
        {
            lod->Cleanup();
        }
    }
}

int LODObject::Render()
{
    return GetCurrentObject()->Render();
}

const D3DXMATRIX& LODObject::GetLocalToWorld() const
{
    return GetCurrentObject()->GetLocalToWorld();
}

void LODObject::SetLocalToWorld(const D3DXMATRIX& l2w)
{
    for(LODIter iter = lods.begin(); iter != lods.end(); ++iter)
    {
        Object* lod = *iter;
        if(lod)
        {
            lod->SetLocalToWorld(l2w);
        }
    }
}

void LODObject::SetLocalBounds(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi)
{
    for(LODIter iter = lods.begin(); iter != lods.end(); ++iter)
    {
        Object* lod = *iter;
        if(lod)
        {
            lod->SetLocalBounds(lo, hi);
        }
    }
}

const Bounds& LODObject::GetLocalBounds() const
{
    return GetCurrentObject()->GetLocalBounds();
}

const Bounds& LODObject::GetWorldBounds() const
{
    return GetCurrentObject()->GetWorldBounds();
}

void LODObject::IncrementTime(float fAdvanceTime)
{
    for(LODIter iter = lods.begin(); iter != lods.end(); ++iter)
    {
        Object* lod = *iter;
        if(lod)
        {
            lod->IncrementTime(fAdvanceTime);
        }
    }
}

void LODObject::SetLODObject(Object* obj, int lod)
{
    if(lod >= int(lods.size()))
    {
        for(int i = lods.size(); i < lod; ++i)
            lods.push_back(NULL);
    
        lods.push_back(obj);
    }
    else
    {
        if(lods[lod])
        {
            SafeDelete(lods[lod]);
        }
        lods[lod] = obj;
    }
}

Object* LODObject::GetCurrentObject() const
{
    int myLOD = min(currentLOD, int(lods.size()-1));

    while(!lods[myLOD])
    {
        --myLOD;
    }

    assert(myLOD >= 0);
    assert(lods[myLOD]);

    return lods[myLOD];
}