
#include <windows.h>
#include <d3d11.h>
#include <d3dx11.h>
#include <d3dcompiler.h>
#include <D3DX11.h>

#define _USE_MATH_DEFINES
#include <math.h>

#include <DXUT.h>

#include "ParametricSurface.h"

ParametricSurface::ParametricSurface(const WCHAR* shaderFile = L"Shaders/toy_fir_tree3.3dmodel.fx")
	: GridObject(),
		shaderFile(shaderFile)
{
}

HRESULT ParametricSurface::Init()
{
	return GridObject::Init();
}

void ParametricSurface::Cleanup()
{
	GridObject::Cleanup();

	SAFE_RELEASE(materialCB);
}

HRESULT ParametricSurface::GetConstantBuffers()
{
	HRESULT hr = S_OK;

	hr = GridObject::GetConstantBuffers();
    if( FAILED( hr ) )
        return hr;

	ID3D11Device* pd3dDevice = DXUTGetD3D11Device();

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;

    ZeroMemory( &bd, sizeof(bd) );
    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(MaterialCB);
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;

    hr = pd3dDevice->CreateBuffer( &bd, NULL, &materialCB);
    if( FAILED( hr ) )
        return hr;

	return hr;
}

void ParametricSurface::UpdateConstants()
{
	GridObject::UpdateConstants();

	ID3D11DeviceContext* pImmediateContext = DXUTGetD3D11DeviceContext();

	// Moving the materials setup to be consts embedded in the .fx file,
	// so that the .fx file completely determines the appearance of the model.
	MaterialCB mCb;
	D3DXVECTOR4 vec(0.8, 0.7, -1.0, 0.0);
	D3DXVec4Normalize(&mCb.LightVector, &vec);
	mCb.TessFactor = D3DXVECTOR4(tessFactor, tessFactor, tessFactor, tessFactor);


    pImmediateContext->UpdateSubresource( materialCB, 0, NULL, &mCb, 0, 0 );

    pImmediateContext->VSSetConstantBuffers( 3, 1, &materialCB);
    pImmediateContext->PSSetConstantBuffers( 3, 1, &materialCB);
}

D3DXVECTOR4 ParametricSurface::ModDivScale() const
{
	D3DXVECTOR4 modDivScale(float(NumU() - 1), float(NumU() - 1), 1.0f / float(NumU() - 1), 1.0f / float(NumV() - 1));
	return modDivScale;
}

static float overTess = -1.0f;
int ParametricSurface::NumU() const 
{ 
	float minTris = float(GetMinTris());
	float minQuads = minTris * 0.5f;
	float minEdges = sqrt(minQuads);
	minEdges = ceil(minEdges);

	if(powerOfTwo)
	{
		tessFactor = minEdges;
		float logTess = float(log(tessFactor) / M_LN2);
		logTess += 0.5f;
		int intLogTess = int(logTess);
		float fracLogTess = logTess - intLogTess;
		logTess = float(intLogTess);
		tessFactor = pow(2.0f, logTess) - 1.0f + fracLogTess;

		if(overTess > 0)
			tessFactor = overTess;

		minEdges = ceil(tessFactor);
	}


	float numVerts = minEdges + 1;

	return int(numVerts); 
}

int ParametricSurface::NumV() const 
{ 
	return NumU();
}

