
#ifndef PATCH_SURFACE_H
#define PATCH_SURFACE_H

#include "GridObject.h"

class PatchSurface : public GridObject
{
private:
	const WCHAR* shaderFile;

	ID3D11HullShader*		hullShader;
	ID3D11DomainShader*		domainShader;

	ID3D11Buffer*			materialCB;

	struct MaterialCB
	{
		D3DXVECTOR4 LightVector;
		D3DXVECTOR4 TessFactor;
	};

public:

	PatchSurface();
	PatchSurface(const WCHAR* shaderFile);

	virtual HRESULT Init();

	virtual void Cleanup();

protected:
	virtual const WCHAR* ShaderFile() const { return shaderFile; }
	virtual int ShaderModel() const { return 5; }
	virtual const CHAR* ShaderExtention() const { return "Patch"; }
	virtual const WCHAR* WideShaderExtention() const { return L"Patch"; }

	virtual void UpdateConstants();
	virtual HRESULT GetConstantBuffers();
	virtual HRESULT MakeGeometry();
	virtual void SetBuffers();
	virtual int Render();

	virtual int NumU() const;
	virtual int NumV() const;

	virtual D3DXVECTOR4 ModDivScale() const;

private:
	virtual void SetResources();

	HRESULT GetDomainShader(const WCHAR* fileName, ID3D11DomainShader*& pShader);
	HRESULT GetHullShader(const WCHAR* fileName, ID3D11HullShader*& pShader);
};




#endif // PATCH_SURFACE_H
