
#ifndef PROC_SCAPE_H
#define PROC_SCAPE_H

#include "Scape.h"

#include "../Utility/Bounds.h"

struct ScapeVertex;

class ProcScape : public Scape
{
public:
    ProcScape();

	virtual HRESULT Init();
    virtual void OnResize() {}
	virtual void Cleanup();

	virtual int Render();

    virtual const D3DXMATRIX& GetLocalToWorld() const { return localToWorld; }
	virtual void SetLocalToWorld(const D3DXMATRIX& l2w);

    virtual void IncrementTime(float fAdvanceTime) {}

    virtual void SetLocalBounds(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi) { this->lo = lo; this->hi = hi; RefreshTransform(); }
    virtual const Bounds& GetLocalBounds() const { return localBounds; }
    virtual const Bounds& GetWorldBounds() const { return worldBounds; }

public:

    void SetResolution(int numU, int numV);

    virtual bool QueryHeight(const D3DXVECTOR2& xyLoc, float& height, D3DXVECTOR3& normal);

    virtual void SetChessBoard(float squareSize);

private:
    
    void RefreshTransform();
    ScapeVertex* GenVerts();
    void CalcNormals(ScapeVertex* verts);
    WORD* GenIndices();

    HRESULT MakeTextures();
    HRESULT MakeShaders();
    HRESULT MakeGeometry();
    HRESULT MakeState();
    HRESULT MakeConstants();

    void SetState(ID3D11DeviceContext* pImmediateContext);
    void CheckState();
    void SetConstants(ID3D11DeviceContext* pImmediateContext, const D3DXMATRIX& l2w);
    void SetVerts(const ScapeVertex* const pts, int count);
    void SetTextures(ID3D11DeviceContext* pImmediateContext);

    int NumU() const { return numU; }
    int NumV() const { return numV; }
    int NumVerts() const { return NumU() * NumV(); }
    int NumQuads() const { return (NumU() - 1) * (NumV() - 1); }
    int NumTris() const { return NumQuads() * 2; }
    int NumIndices() const { return NumTris() * 3; }

    D3DXVECTOR3 lo;
    D3DXVECTOR3 hi;
    OBB localBounds;
    OBB worldBounds;
    D3DXMATRIX localToWorld;
    D3DXMATRIX worldToLocal;
    D3DXMATRIX lastLocalToWorld;
    
    int numU;
    int numV;

    ID3D11RasterizerState*   m_pRasterizerState;

    ID3D11VertexShader*      m_pVertexShader;
    ID3D11PixelShader*       m_pPixelShader;
    ID3D11InputLayout*       m_pVertexLayout;
    ID3D11Buffer*            m_pVertexBuffer;
    ID3D11Buffer*            m_pIndexBuffer;
    ID3D11Buffer*            m_pConstantBuffer;

	ID3D11ShaderResourceView*           m_pDetailTop;
	ID3D11ShaderResourceView*           m_pDetailBot;
	ID3D11ShaderResourceView*           m_pMask;
    ID3D11SamplerState*                 m_pSampler;
	ID3D11SamplerState*                 m_pSamplePrevious;

    ScapeVertex*            m_pVerts;

    bool                    m_bWireframe;

    float                   m_SquareSize;
};

#endif // PROC_SCAPE_H
