
#ifndef SPHERE_H
#define SPHERE_H

#include "Object.h"

class Sphere : public Object
{

public:

	Sphere();
	virtual ~Sphere();

	virtual HRESULT Init();
	virtual void OnResize();
	virtual void Cleanup();

	virtual int Render();

	virtual const D3DXMATRIX& GetLocalToWorld() const;
	virtual void SetLocalToWorld(const D3DXMATRIX& l2w);

	virtual void IncrementTime(FLOAT fAdvanceTime);

    virtual void SetLocalBounds(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi);
    virtual const Bounds& GetLocalBounds() const { return localBounds; }
    virtual const Bounds& GetWorldBounds() const { return worldBounds; }

    void SetRadius(FLOAT f);
    FLOAT GetRadius() const { return m_Radius; }

protected:

	void SetOnResize();
	HRESULT MakeVertexShader();
	HRESULT MakePixelShader();
	HRESULT MakeConstantBuffers();
	HRESULT MakeRasterizerStates();

	void UpdateConstants();
	void SetResources();
	void SetBuffers();
	void SetState();
	int NumU() const;
	int NumV() const;
	int NumTris() const;

	ID3D11VertexShader*		m_pVertexShader;
	ID3D11PixelShader*		m_pPixelShader;

	ID3D11Buffer*			m_pEveryFrameCB;

	ID3D11RasterizerState*  m_pRasterizerStateSolid;
	ID3D11RasterizerState*  m_pRasterizerStateWireframe;

	D3DXMATRIX m_LocalToWorld;
	FLOAT m_fTime;

    FLOAT m_Radius;
    SphereBounds localBounds;
    SphereBounds worldBounds;

	struct CBChangesEveryFrame
	{
		D3DXMATRIX mCameraToNDC;

        D3DXMATRIX m_LocalToWorld;

		D3DXVECTOR4 m_ModDivScale;

		D3DXMATRIX m_WorldToCamera;
		D3DXVECTOR4 m_WorldCamPos;
	};

};


#endif // SPHERE_H
