
#include <windows.h>
#include <d3d11.h>
#include <d3dx11.h>
#include <D3DX10math.h>
#include <assert.h>

#include "FauxV8Kit/Utility.h"

#include "SettingBlock.h"

static WCHAR _currentLine[MAX_PATH] = { '0' };
static const WCHAR _delim[] = L" ,:\t\n";

static WCHAR* _token = NULL;
static WCHAR* _tokenNext = NULL;


SettingBlock::SettingBlock()
{
}

SettingBlock::~SettingBlock()
{

}

bool SettingBlock::Load( FILE* fp, const SettingBlock& pattern )
{
    const WCHAR* tok = NextToken(fp);
    while(tok)
    {
        if(!wcscmp(tok, L"EndBlock"))
            break;

        const int cnt = int(pattern.m_Values.size());
        for(int i = 0; i < cnt; ++i)
        {
            if(!wcscmp(tok, pattern.m_Values[i].m_Label))
            {
                ValueChunk val(pattern.m_Values[i]);
                val.m_LoadFunc(val, fp);
                m_Values.push_back(val);
                break;
            }
        }
        tok = NextToken(fp);
    }
    return m_Values.size() > 0;
}

void SettingBlock::Set()
{
    const int cnt = int(m_Values.size());
    for(int i = 0; i < cnt; ++i)
    {
        const ValueChunk& val = m_Values[i];
        val.m_SetFunc(val);
    }
}

POINT SettingBlock::LoadPoint( FILE* fp )
{
    POINT ret;
    ret.x = LoadInt(fp);
    ret.y = LoadInt(fp);
    return ret;
}

D3DXVECTOR2 SettingBlock::LoadVector2( FILE* fp )
{
    return D3DXVECTOR2(LoadFloat(fp), LoadFloat(fp));
}

D3DXVECTOR3 SettingBlock::LoadVector3( FILE* fp )
{
    return D3DXVECTOR3(LoadFloat(fp), LoadFloat(fp), LoadFloat(fp));
}

D3DXVECTOR4 SettingBlock::LoadVector4( FILE* fp )
{
    return D3DXVECTOR4(LoadFloat(fp), LoadFloat(fp), LoadFloat(fp), LoadFloat(fp));
}

float SettingBlock::LoadFloat( FILE* fp )
{
    WCHAR* tok = NextToken(fp);
    assert(tok);
    return float(_wtof(tok));
}

int SettingBlock::LoadInt( FILE* fp )
{
    WCHAR* tok = NextToken(fp);
    assert(tok);
    return _wtoi(tok);
}

bool SettingBlock::LoadBool( FILE* fp )
{
    WCHAR* tok = NextToken(fp);
    if(!wcscmp(L"true", tok))
        return true;
    if(!wcscmp(L"false", tok))
        return false;
    return LoadInt(fp) != 0;
}

TCHAR* SettingBlock::LoadString( FILE* fp, TCHAR* dst )
{
    int i = 0;
    while( (i < MAX_PATH - 1) && _tokenNext[i] && (_tokenNext[i] != ';'))
    {
        dst[i] = _tokenNext[i];
        ++i;
    }
    dst[i] = 0;

    // We've just consumed the rest of the line (except possible comment), now advance beyond.
    // We don't need to check for success, if we're at EOF we'll catch the failure when we go
    // for the next token.
    //NextLine(fp);
    _tokenNext += wcslen(_tokenNext);

    return dst;
}

bool SettingBlock::NextLine( FILE* fp )
{
    _token = _tokenNext = NULL;
    if(fgetws(_currentLine, MAX_PATH, fp))
    {
        _token = wcstok_s(_currentLine, _delim, &_tokenNext);
        return true;
    }
    return false;
}

WCHAR* SettingBlock::NextToken( FILE* fp )
{
#pragma warning(push)
#pragma warning(disable: 4127)
    while(true)
    {
        if(!_token)
        {
            if(!NextLine(fp))
                return NULL;
        }
        else
        {
            _token = wcstok_s(NULL, _delim, &_tokenNext);
        }
        if(_token)
        {
            if(*_token == ';')
                _token = NULL;
            else
                break;
        }
    }
#pragma warning(pop)
    return _token;
}

void SettingBlock::AddChunk( const ValueChunk& val )
{
    m_Values.push_back(val);
}

