
#ifndef SETTING_BLOCK_H
#define SETTING_BLOCK_H

#include <stdio.h>
#include <vector>

class SettingBlock
{
public:
    SettingBlock();
    ~SettingBlock();

    bool Load( FILE* fp, const SettingBlock& pattern );

    void Set();

    struct ValueChunk
    {
        typedef void SetValueFunc(const ValueChunk& val);
        typedef void LoadValueFunc(ValueChunk& val, FILE* fp);

        const WCHAR*    m_Label;
        SetValueFunc*    m_SetFunc;
        LoadValueFunc*   m_LoadFunc;
        union
        {
            int     iValue;
            float   fValue;
            bool    bValue;
            TCHAR   cValue[MAX_PATH];
        };
        static void GetFloat( ValueChunk& val, FILE* fp ) { val.fValue = LoadFloat(fp); }
        static void GetInt( ValueChunk& val, FILE* fp ) { val.iValue = LoadInt(fp); }
        static void GetBool( ValueChunk& val, FILE* fp ) { val.bValue = LoadBool(fp); }
        static void GetString( ValueChunk& val, FILE* fp) { LoadString(fp, val.cValue); }

        ValueChunk()
            :   m_Label(NULL),
                m_SetFunc(NULL),
                m_LoadFunc(NULL),
                iValue(0)
        {
        }
        ValueChunk(const WCHAR* label, SetValueFunc* setFunc, LoadValueFunc* loadFunc)
            :   m_Label(label),
                m_SetFunc(setFunc),
                m_LoadFunc(loadFunc),
                iValue(0)
        {}
        ValueChunk(const ValueChunk& other)
            :   m_Label(other.m_Label),
                m_SetFunc(other.m_SetFunc),
                m_LoadFunc(other.m_LoadFunc)
        {
            // Need to be sure we're copying in the largest member of the union. Maybe should just memcpy?
            memcpy_s(cValue, sizeof(cValue), other.cValue, sizeof(other.cValue));
        }
    };

    void AddChunk(const ValueChunk& val);

private:

    static float LoadFloat( FILE* fp );
    static int LoadInt( FILE* fp );
    static bool LoadBool( FILE* fp );
    static TCHAR* LoadString( FILE* fp, TCHAR* dst ); // dst must be big enough w/ max MAX_PATH

    static POINT LoadPoint( FILE* fp );
    static D3DXVECTOR2 LoadVector2( FILE* fp );
    static D3DXVECTOR3 LoadVector3( FILE* fp );
    static D3DXVECTOR4 LoadVector4( FILE* fp );

    static WCHAR* NextToken( FILE* fp );
    static bool NextLine( FILE* fp );
    typedef std::vector<ValueChunk> ValueVector;
    typedef ValueVector::iterator ValueIter;
    typedef ValueVector::const_iterator ValueConstIter;
    ValueVector m_Values;
};

#endif // SETTING_BLOCK_H