
#ifndef BLIT_QUAD_FX
#define BLIT_QUAD_FX

Texture2D<float4> tex0 : register( t0 );
Texture2D<float4> tex1 : register( t1 ); // inner
Texture2D<float4> tex2 : register( t2 ); // intermediate
Texture2D<float4> tex3 : register( t3 ); // this index must match BlitComposite::kMaxLayers

int NumTextures;

sampler LinearSampler
{
	Filter = MIN_MAG_MIP_LINEAR;
};


cbuffer cbBlitQuad : register( b0 )
{
	float4 scaleOffDst;

	float4 scaleOffDomainToTex1;
	float4 scaleOffDomainToTex2;
	float4 scaleOffDomainToTex3;

	float4 pixOffset1;
	float4 pixOffset2;
	float4 pixOffset3;

	float4 Decay;

};

struct VS_INPUT
{
	uint Index	: SV_VertexID;
};

struct PS_INPUT
{
	float4 Position : SV_POSITION;
	float2 TexCoord : TEXCOORD0;
};

PS_INPUT MakeVertex(uint vertId)
{
	PS_INPUT output;
	const float2 domainPos = float2(
		(vertId % 2),
		(vertId / 2)
		);

	output.Position = float4(domainPos * scaleOffDst.xy + scaleOffDst.zw, 0.5f, 1.0f);

	output.TexCoord.xy = domainPos;

	return output;
}

PS_INPUT VS_Def(VS_INPUT input)
{
	PS_INPUT output = MakeVertex(input.Index);

	return output;
}

//#define MAF_SHOW_LEVELS
#ifdef MAF_SHOW_LEVELS
float4 ShowLevels(const float4 inColor, const float4 bkgColor, const float level, const float inStrength)
{
	float4 color = inColor;
    if(MAF_SHOW_LEVELS & (1 << int(level)))
    {
	    if(inStrength < 1.0f)
	    {
		    float strength = 1.0f - inStrength;
		    strength /= MAF_SHOW_LEVELS == 2.0f ? 1.0f : (1.0f + level);
		    color *= strength;
	    }
    }
	return color;
}
#elif defined(MAF_COLOR_LEVELS)
float4 ShowLevels(const float4 inColor, const float4 bkgColor, const float level, const float inStrength)
{
	float4 color = inColor;
	if(inStrength < 1.0f)
	{
		color = level < 1 ? float4(1.0f, 1.0f, 1.0f, 1.0f)
							 : level < 2 ? float4(1.0f, 1.0f, 0.3f, 1.0f)
										 : float4(1.0f, 0.3f, 0.3f, 1.0f);
		color = lerp(color, bkgColor, inStrength);
	}
	return color;
}
#else // MAF_SHOW_LEVELS
float4 ShowLevels(const float4 inColor, const float4 bkgColor, const float level, const float strength)
{
	return inColor;
}
#endif // MAF_SHOW_LEVELS

float4 BlendLayer5(const float4 current, Texture2D<float4> tex, const float level, 
	const float4 scaleOff, const float4 pixOff, const float2 domainPos)
{
	const float2 uv = domainPos * scaleOff.xy + scaleOff.zw;

	const float len = length(uv * 2.0f.xx - 1.0f.xx);

	const float strength = smoothstep(Decay.x, Decay.y, len);

//	float4 layer = tex.Sample( LinearSampler, uv );

	float4 layer = tex.Sample( LinearSampler, uv + pixOff.xy ); 
	layer += tex.Sample( LinearSampler, uv + pixOff.zy ); 
	layer += tex.Sample( LinearSampler, uv + pixOff.xw ); 
	layer += tex.Sample( LinearSampler, uv + pixOff.zw ); 

	float4 color = lerp( layer * 0.25f.xxxx, current, strength );

	color = ShowLevels(color, current, level, strength);

	return color;
}

float4 BlendLayer(const float4 current, Texture2D<float4> tex, const float level, 
	const float4 scaleOff, const float2 domainPos)
{
	const float2 uv = domainPos * scaleOff.xy + scaleOff.zw;

	const float len = length(uv * 2.0f.xx - 1.0f.xx);

    float strength = smoothstep(Decay.x, Decay.y, len);
//strength = strength * 0.5f + 0.5f;
//strength = uv.x >= 0.0f && uv.x <= 1.0f && uv.y >= 0.0f && uv.y <= 1.0f ? 0.5f : 1.0f;
//if(level == 0)
    //strength = 1;

	const float4 layer = tex.Sample( LinearSampler, uv );

	float4 color = lerp( layer, current, strength );

	color = ShowLevels(color, current, level, strength);
	return color;
}

float4 MSBackground(const float2 domainPos, const float4 scaleOff, const float4 pixOff, Texture2D tex)
{
	const float2 uv = domainPos * scaleOff.xy + scaleOff.zw;
	float4 color = tex.Sample( LinearSampler, uv + pixOff.xy );
	color += tex.Sample( LinearSampler, uv + pixOff.zy );
	color += tex.Sample( LinearSampler, uv + pixOff.xw );
	color += tex.Sample( LinearSampler, uv + pixOff.zw );

	return color * 0.25f.xxxx;
}

//#define MAF_NUM_LAYERS (2)
//#define MAF_MS_BASE
//#define MAF_MS_MID
//#define MAF_MS_INNER
//
float4 PS_Def(PS_INPUT input) : SV_Target
{
//	return float4(1.0f, 0.5f, 0.5f, 1.0f);

	const float2 uv = input.TexCoord;

#ifdef MAF_MS_BASE
	float4 color = MSBackground(uv, scaleOffDomainToTex3, pixOffset3, tex3);
#else // MAF_MS_BASE
	float4 color = tex3.Sample( LinearSampler, uv * scaleOffDomainToTex3.xy + scaleOffDomainToTex3.zw );
#endif // MAF_MS_BASE

#ifdef MAF_COLOR_LEVELS
	color = float4(1.0f, 0.1f, 0.1f, 1.0f);
#endif // MAF_COLOR_LEVELS

#if (MAF_NUM_LAYERS > 1)
	#if defined(MAF_MS_MID)
		color = BlendLayer5( color, tex2, 1, scaleOffDomainToTex2, pixOffset2, uv );
	#else // defined(MAF_MS_MID)
		color = BlendLayer( color, tex2, 1, scaleOffDomainToTex2, uv );
	#endif // defined(MAF_MS_MID)
#endif // (MAF_NUM_LAYERS > 1)

#if (MAF_NUM_LAYERS > 0)
	#ifdef MAF_MS_INNER
		color = BlendLayer5( color, tex1, 0, scaleOffDomainToTex1, pixOffset1, uv );
	#else // MAF_MS_INNER
		color = BlendLayer( color, tex1, 0, scaleOffDomainToTex1, uv );
	#endif // MAF_MS_INNER
#endif

	return color;
}


#endif // BLIT_QUAD_FX
