
#ifndef BLIT_DEPTH_FX
#define BLIT_DEPTH_FX


cbuffer cbBlitDepth : register( b0 )
{
	float4 scaleOffSrc;
	float4 scaleOffDst;
    
    /// radiusSq means different things to PS_Middle vs. PS_Base.
    /// inset refers to the layer on top of this layer being blitted to.
    /// falloff refers to the point at which the inset STARTS blending out (e.g. 0.6).
    /// PS_Middle: rSq = inset res/this res * inset dup / this dup * inset falloff
    /// PS_Base: rSq = inset res * inset dup * inset falloff
    /// radiusSq = PS_Middle rSq
    /// radiusSq = PS_Base rSq
    /// Note this might take revision if we support more than 3 layers.
    float4 radiusSq;
    float4 center; // screen coord center.xy, screen size.zw
};

struct VS_INPUT
{
	uint Index	: SV_VertexID;
};

struct PS_INPUT
{
	float4 Position : SV_POSITION;
	float2 TexCoord : TEXCOORD0;
};

PS_INPUT MakeVertex(uint vertId)
{
	PS_INPUT output;
	float2 domainPos = float2(
		(vertId % 2),
		(vertId / 2)
		);

	output.Position = float4(domainPos * scaleOffDst.xy + scaleOffDst.zw, 0.1f, 1.0f);

	output.TexCoord.xy = domainPos * scaleOffSrc.xy + scaleOffSrc.zw;

	return output;
}
PS_INPUT VS_Def(VS_INPUT input)
{
	return MakeVertex(input.Index);
}

PS_INPUT VS_Inner(VS_INPUT input)
{
    return VS_Def(input);
}

PS_INPUT VS_Middle(VS_INPUT input)
{
    return VS_Def(input);
}

PS_INPUT VS_Base(VS_INPUT input)
{
    return VS_Def(input);
}

float4 PS_Inner(PS_INPUT input) : SV_Target
{
    const float2 off = input.TexCoord.xy * 2.0f.xx - 1.0f.xx;
    const float distSq = dot(off, off);

    if(distSq < 1.0f)
        discard;

    return float4(0.0f, 1.0f, 1.0f, 1.0f);
}

float4 PS_Middle(PS_INPUT input) : SV_Target
{
    const float2 off = input.TexCoord.xy * 2.0f.xx - 1.0f.xx;
    const float distSq = dot(off, off);

    if((distSq < 1.0f) && (distSq > radiusSq.x))
        discard;

    return float4(0.0f, 1.0f, 0.0f, 1.0f);
}

float4 PS_Base(PS_INPUT input) : SV_Target
{
    const float2 uvScreen = input.TexCoord.xy * center.zw;
    const float2 off = center.xy - uvScreen;
    const float distSq = dot(off, off);

    if( distSq > radiusSq.x )
        discard;

    return float4(1.0f, 1.0f, 0.0f, 1.0f);
}

#endif // BLIT_DEPTH_FX
