
#ifndef BLIT_QUAD_FX
#define BLIT_QUAD_FX

TextureCube <float4> cubeMap : register( t0 );
SamplerState cubeSampler : register( s0 );

SamplerState envSampleClamp
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};



cbuffer cbBlitQuad : register( b0 )
{
	float4 scaleOffDst;
	float4x4 ndcToWorld;
	float4 eyePos;
};

struct VS_INPUT
{
	uint Index	: SV_VertexID;
};

struct PS_INPUT
{
	float4 Position : SV_POSITION;
	float2 TexCoord : TEXCOORD0;
};

PS_INPUT MakeVertex(uint vertId)
{
	PS_INPUT output;
	float2 domainPos = float2(
		(vertId % 2),
		(vertId / 2)
		);

	output.Position = float4(domainPos * scaleOffDst.xy + scaleOffDst.zw, 0.5f, 1.0f);

//	output.TexCoord.xy = domainPos * scaleOffSrc.xy + scaleOffSrc.zw;
	output.TexCoord.xy = domainPos * scaleOffDst.xy + scaleOffDst.zw;

	return output;
}
PS_INPUT VS_Def(VS_INPUT input)
{
	PS_INPUT output = MakeVertex(input.Index);

	return output;
}

float4 SampleEnv(const float2 uv)
{
	float4 worldDir = mul(float4(uv.x, uv.y, 0.0f, 1.0f), ndcToWorld);
	worldDir /= worldDir.w;
	worldDir.xyz -= eyePos.xyz;
//    worldDir.y -= 0.005f;
	worldDir.xyz = normalize(worldDir.xyz);

	return cubeMap.Sample( envSampleClamp, worldDir.xyz );
}

float4 PS_Def(PS_INPUT input) : SV_Target
{
	return SampleEnv(input.TexCoord);
}

float4 PS_Def_Multi(PS_INPUT input) : SV_Target
{
	//return float4(0.0f, 1.0f, 0.5f, 1.0f);

	float4 uvOff = float4(1.5f / 256.0f, 1.5f / 256.0f, -1.5f / 256.0f, -1.5f / 256.0f);

	float4 color = SampleEnv(input.TexCoord + uvOff.xy);
	color += SampleEnv(input.TexCoord + uvOff.xw);
	color += SampleEnv(input.TexCoord + uvOff.zy);
	color += SampleEnv(input.TexCoord + uvOff.zw);

	return color * 0.25f;
}

#endif // BLIT_QUAD_FX
