
#ifndef BSPLINE_FX
#define BSPLINE_FX

//------------------------------------------------------------------------------
//
// BSpline 2D curve definitions
// These only get defined once, so are bracketted within #ifndef BSPLINE_FX
//
//------------------------------------------------------------------------------

struct PARAMETRICCURVE2D_OUTPUT
{
    float2	Position     : POSITION;        // 2D position on the curve
};


static const float4x4 BSpline_BaseMatrix = float4x4(-1,  3, -3,  1,
													  3, -6,  3,  0,
													 -3,  0,  3,  0,
													  1,  4,  1,  0) / 6;


#endif // BSPLINE_FX

//------------------------------------------------------------------------------
//
// BSpline 2D curve methods
// This gets included multiple times with different data to generate a 
// different function for each set of control points. Would be nice to
// pass in an array, but all function parameters are by value, so an
// indefinite array (e.g. "float2 controlPoints[]" or "float2* controlPoints")
// is not supported.
//
//------------------------------------------------------------------------------

// [t] ---> [x,y, x',y', x",y"]
PARAMETRICCURVE2D_OUTPUT FUNC_LABEL ( const float input )
{
	const float BSpline_SingularityTolerance = .0001f;

    float scaledT = float( CONTROL_POINT_COUNT - 3 ) * input;
    uint currentSegment = uint(floor(scaledT));
    
    // NOTE skipping if(t==1) {currentSegment--;}
    // Prevent bad values by assuming an extra control point at the end
    
    float2 c0 = CONTROL_POINTS[currentSegment + 0];
    float2 c1 = CONTROL_POINTS[currentSegment + 1];
    float2 c2 = CONTROL_POINTS[currentSegment + 2];
    float2 c3 = CONTROL_POINTS[currentSegment + 3];
    
	scaledT = frac(scaledT);  // Param is in [0..1] range
	scaledT`(input) = 1.0f;
        
    // GPU-optimized singularity treatment: to avoid dynamic branching and testing
    // control points for the occurence of a singularity, simply clamp relative T
    // to the [0+delta, 1-delta] range.
//    float relativeT = clamp(scaledT, BSpline_SingularityTolerance, 1.f-BSpline_SingularityTolerance);
	float relativeT = clamp(scaledT, 0.f, 1.f);

    float relativeT2 = relativeT * relativeT;
    float relativeT3 = relativeT2 * relativeT;
    
    float4x4 controlMatrix = float4x4(c0.xy, 0, 0,
                                  c1.xy, 0, 0,
                                  c2.xy, 0, 0,
                                  c3.xy, 0, 0);
                           
    float4x4 combinedMatrix = mul(BSpline_BaseMatrix, controlMatrix);
    
    float4 tVector0 = float4(relativeT3, relativeT2, relativeT, 1.0f);

    PARAMETRICCURVE2D_OUTPUT output;
    output.Position = mul(tVector0, combinedMatrix).xy;

    return output;
}

