
#ifndef BASESHADER_FX
#define BASESHADER_FX

#include "BaseBuffers.fx"

#include "PointLight.fx"

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float2 Tex : TEXCOORD0;
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
};

//--------------------------------------------------------------------------------------
// Vertex Shader
//--------------------------------------------------------------------------------------
PS_INPUT VS( VS_INPUT input )
{

    PS_INPUT output = (PS_INPUT)0;
	float3 surface = SurfaceFunc(input.Tex);
    output.Pos = mul( float4(surface, 1.0f), LocalToWorld );
    output.Pos = mul( output.Pos, WorldToCamera );
    output.Pos = mul( output.Pos, CameraToNDC );
    output.Tex = input.Tex;
    
    return output;
}

//--------------------------------------------------------------------------------------
// Pixel Shader
//--------------------------------------------------------------------------------------
float4 PS( PS_INPUT input) : SV_Target
{

	float2 uv = input.Tex;
	float3 surface = SurfaceFunc(uv);

	float3 norm = cross(surface`(uv.x), surface`(uv.y));

	norm = NORMALIZE_CHECK(norm);

	norm = mul(norm, (float3x3)LocalToWorld);
//return float4(norm * 0.5 + 0.5.xxx, 1.0);

	float4 pos = mul(float4(surface, 1.0f), LocalToWorld);
	float3 eyeVec = pos.xyz - WorldCamPos.xyz;
	eyeVec.xyz = reflect(eyeVec.xyz, norm);

	float4 env = cubeMap.Sample( cubeSampler, eyeVec.xyz );


	float3 color = PointLight(pos, norm, WorldCamPos.xyz);
	env.xyz = color;
	env.xyz *= color;
	env.xyz = env * 0.2 + color;
//	env.xyz = lerp(env.xyz, color, 0.75);

	return float4(env.xyz, 1.0f);
}

#endif // BASESHADER_FX
