
//--------------------------------------------------------------------------------------
// File: DrawTerrain.fx
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

Texture2D <float4> detailTop : register( t0 );
Texture2D <float4> detailBot : register( t1 );
Texture2D <float4> mask : register( t2 );
SamplerState sampleWrap : register( s0 );

static float4 LightVector = float4(0.8, 0.7, -1.0, 0.0);

//SamplerState sampleWrap
//{
    //AddressU = Clamp;
    //AddressV = Clamp;
//
//#if 0
	//Filter = MIN_MAG_MIP_LINEAR;
//#else
	//Filter = ANISOTROPIC;
	//MaxAnisotropy = 16;
//#endif
//};
//
//--------------------------------------------------------------------------------------
// Constant Buffer Variables
//--------------------------------------------------------------------------------------
cbuffer ConstantBuffer : register( b0 )
{
	float4x4 LocalToNDC;
	float4x4 LocalToWorld;
	float4 Color;
	float4x4 LastLocalToNDC;
	float4 TexScale;
	float4 Light;
	float4 WindowSize;
    float4 ChessBoard;
}

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float4 Pos		: POSITION;
	float3 Norm		: NORMAL;
	float2 TexCoord	: TEXCOORD0;
};

struct PS_INPUT
{
    float4 Pos		: SV_POSITION;
    float3 WorldPos : POSITION;
	float3 Norm		: NORMAL;
    float4 Color	: COLOR0;
	float2 TexCoord	: TEXCOORD0;
	float4 LastPos	: TEXCOORD1;
};

#include "Reproj.fx"

//--------------------------------------------------------------------------------------
// Vertex Shader
//--------------------------------------------------------------------------------------
PS_INPUT VS_Def( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    output.Pos = mul( input.Pos, LocalToNDC );
    output.WorldPos = mul( input.Pos, LocalToWorld );
	output.Norm = mul( input.Norm, (float3x3)LocalToWorld); 
	output.Color = float4(input.TexCoord.x, input.TexCoord.y, 0.5f, 1.0f);
	output.TexCoord = input.TexCoord;
//output.Color = float4(1.0f, 0.5f, 0.5f, 1.0f);
//output.Color = float4(input.Norm, 1.0f);

	output.LastPos = VS_Reproject(input.Pos.xyz, LastLocalToNDC);
    
    return output;
}


//--------------------------------------------------------------------------------------
// Pixel Shader
//--------------------------------------------------------------------------------------
float4 PS_Def( PS_INPUT input) : SV_Target
{
//return float4(normalize(input.Norm) * 0.5f + 0.5f, 1.0f);
	const float2 topCoord = input.TexCoord * TexScale.xy;
	const float2 botCoord = input.TexCoord * TexScale.zw;
	const float4 top = detailTop.Sample( sampleWrap, topCoord );
	const float4 bot = detailBot.Sample( sampleWrap, botCoord );

	const float trans = 1.0f - mask.Sample(sampleWrap, input.TexCoord ).x;

	const float3 norm = normalize(input.Norm);
    float fDiffuse = saturate(dot(norm, LightVector.xyz ) ) * Light.x;

	const float3 tex = lerp(top.xyz, bot.xyz, trans.xxx).xyz;

    if(ChessBoard.x > 0)
    {
        float2 gridPos = input.WorldPos.xz * ChessBoard.y;
        gridPos = gridPos > 0 ? gridPos : -gridPos + 0.5f.xx;
        gridPos = frac(gridPos);
        gridPos = gridPos > 0.5f.xx ? 1.0f : 0.0f;
        fDiffuse *= gridPos.x + gridPos.y == 1.0f ? 1.0f : ChessBoard.z;
    }

	return PS_Reproject( input.Pos, input.LastPos, float4(tex * fDiffuse, 1.0f) );
}
