
#ifndef PROC_TERRAIN_FX
#define PROC_TERRAIN_FX


//--------------------------------------------------------------------------------------
// File: DrawTerrain.fx
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

Texture2D <float4> detailTop : register( t0 );
Texture2D <float4> detailBot : register( t3 );
Texture2D <float4> mask : register( t5 );
SamplerState sampleWrap : register( s0 );

static float4 LightVector = float4(0.8, 0.7, -1.0, 0.0);

//SamplerState sampleWrap
//{
    //AddressU = Clamp;
    //AddressV = Clamp;
//
//#if 0
	//Filter = MIN_MAG_MIP_LINEAR;
//#else
	//Filter = ANISOTROPIC;
	//MaxAnisotropy = 16;
//#endif
//};
//
//--------------------------------------------------------------------------------------
// Constant Buffer Variables
//--------------------------------------------------------------------------------------
cbuffer ConstantBuffer : register( b0 )
{
	float4x4 LocalToNDC;
	float4x4 LocalToWorld;
	float4 Color;
	float4x4 LastLocalToNDC;
	float4 TexScale;
	float4 Light;
	float4 WindowSize;
    float4 ChessBoard;
	float4x4 WorldToCamera; // newfish
	float4x4 CameraToNDC;	// newfish
	float4 DetailParams;	// newfish
	float4 WorldCamPos;		// newfish
}

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float4 Pos		: POSITION;
	float3 Norm		: NORMAL;
	float2 TexCoord	: TEXCOORD0;
};

struct PS_INPUT
{
    float4 Pos			: SV_POSITION;
	float3 LocalPos		: POSITION;
	float3 LocalNorm	: NORMAL;
    float4 Color		: COLOR0;
	float2 TexCoord		: TEXCOORD0;
	float4 LastPos		: TEXCOORD1;
};

#include "Reproj.fx"

#include "ProcSurface.fx"

//--------------------------------------------------------------------------------------
// Vertex Shader
//--------------------------------------------------------------------------------------
PS_INPUT VS_Def( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
    output.Pos = mul( input.Pos, LocalToNDC );
//    output.Pos = VS_Perturb(output.Pos);
	output.LocalPos = input.Pos.xyz;
	output.LocalNorm = input.Norm;
	output.Color = float4(input.TexCoord.x, input.TexCoord.y, 0.5f, 1.0f);
	output.TexCoord = input.TexCoord;
//output.Color = float4(1.0f, 0.5f, 0.5f, 1.0f);
//output.Color = float4(input.Norm, 1.0f);

	output.LastPos = VS_Reproject(input.Pos.xyz, LastLocalToNDC);
    
    return output;
}

float RollingPS0(  
	const float3 localPos, 
	const float3 localNorm, 
	const float3 localBin, 
	const float3 localTan, 
	const float4 scales, // x localPos, x noiseOut, x height
	out float3 bumpNorm )
{
	const float4 NoiseIn = float4( localPos * scales.x, 0.0f );
	const float NoiseOut = fBm( NoiseIn, 8, 2.0f, 0.5f, NT_SIMPLEX3D, localPos );

	const float height = NoiseOut * scales.y;

	const float3 deriv = EXTRACT_DERIV3(height, localPos).xyz;

	bumpNorm = normalize(localNorm 
		- localTan * dot(localTan, deriv)
		- localBin * dot(localBin, deriv));

	return height;
}

float RollingPS1(  
	const float3 localPos, 
	const float3 localNorm, 
	const float3 localBin, 
	const float3 localTan, 
	const float4 scales, // x localPos, x noiseOut, x height
	out float3 bumpNorm )
{
	const float4 NoiseIn0 = float4( localPos * scales.x, 0.0f );
	const float NoiseOut0 = fBm( NoiseIn0, 8, 2.0f, 0.5f, NT_SIMPLEX3D, localPos );

	const float4 NoiseIn1 = float4( localPos * scales.z, 0.0f );
	const float NoiseOut1 = fBm( NoiseIn1, 8, 1.3f, 0.77f, NT_SIMPLEX3D_TURB, localPos );

	const float height = NoiseOut0 * scales.y + NoiseOut1 * scales.w;

	const float3 deriv = EXTRACT_DERIV3(height, localPos).xyz;

	bumpNorm = normalize(localNorm 
		- localTan * dot(localTan, deriv)
		- localBin * dot(localBin, deriv));

	return height;
}

float4 Flatten(const float4 color)
{
	const float kRange = 0.4f;
	const float kEmpty = 1.0f - kRange;
	return color * kRange + kEmpty;
}

//--------------------------------------------------------------------------------------
// Pixel Shader
//--------------------------------------------------------------------------------------
float4 PS_Def( PS_INPUT input) : SV_Target
{
//return float4(input.LocalPos, 1.0f);
//return float4(normalize(input.Norm) * 0.5f + 0.5f, 1.0f);
	const float2 topCoord = input.TexCoord * TexScale.xy;
	const float2 botCoord = input.TexCoord * TexScale.zw;
	const float4 top = detailTop.Sample( sampleWrap, topCoord );
	const float4 bot = Flatten( detailBot.Sample( sampleWrap, botCoord ) );

	const float trans = 1.0f - mask.Sample(sampleWrap, input.TexCoord ).x;

	const float3 tex = lerp(top.xyz, bot.xyz, trans.xxx).xyz;
//const float3 tex = float3(1.0f, 1.0f, 1.0f);
//const float3 tex = top.xyz;

	const float3 localNorm = normalize(input.LocalNorm);

	#if 0
	const float worldNorm = mul(localNorm, (float3x3)LocalToWorld);
    const float fDiffuse = saturate(dot(localNorm, LightVector.xyz ) ) * Light.x;


	#else
	const float3 worldPos = mul(float4(input.LocalPos, 1.0f), LocalToWorld).xyz;
	const float3 localTan = normalize(cross(localNorm, float3(1.0f, 0.0f, 0.0f)));
	const float3 localBin = normalize(cross(localTan, localNorm));

	float3 bumpNorm0;
	const float dirt0 = RollingPS0( input.LocalPos, 
						localNorm, localBin, localTan, 
						float4(0.01f, 0.05f, 0.0f, 0.0f),
						bumpNorm0 );
	float3 bumpNorm1;
	const float dirt1 = RollingPS1( input.LocalPos, 
						localNorm, localBin, localTan, 
						float4(0.05f, 0.01f, 0.07f, 0.005f),
						bumpNorm1 );

	float3 bumpNorm = lerp(bumpNorm0, bumpNorm1, trans.x);

//return float4(tex, 1.0f);

	const float aafactor = UnitLocalToScreen( input.LocalPos, localTan, localBin );
	const float detail = saturate( aafactor * DetailParams.x );

	bumpNorm = lerp(localNorm, bumpNorm, detail);

	bumpNorm = mul(bumpNorm, (float3x3)LocalToWorld);

    float fDiffuse = saturate(dot(bumpNorm, LightVector.xyz ) ) * Light.x;
//return float4(fDiffuse.xxx, 1.0f);

	#endif

    if(ChessBoard.x > 0)
    {
        float2 gridPos = input.LocalPos.xy * ChessBoard.y;
        gridPos = gridPos > 0 ? gridPos : -gridPos + 0.5f.xx;
        gridPos = frac(gridPos);
        gridPos = gridPos > 0.5f.xx ? 1.0f : 0.0f;
        fDiffuse *= gridPos.x + gridPos.y == 1.0f ? 1.0f : ChessBoard.z;
    }

	return PS_Reproject( input.Pos, input.LastPos, float4(tex * fDiffuse, 1.0f) );
}


#endif // PROC_TERRAIN_FX
