
#ifndef REFSHADER_FX
#define REFSHADER_FX

TextureCube <float4> cubeMap; // : register( t0 );
SamplerState cubeSampler : register( s0 );

//--------------------------------------------------------------------------------------
// Constant Buffer Variables
//--------------------------------------------------------------------------------------
cbuffer cbNeverChanges : register( b0 )
{
    matrix WorldToCamera;
	float4 WorldCamPos;
};

cbuffer cbChangeOnResize : register( b1 )
{
    matrix CameraToNDC;
};

cbuffer cbChangesEveryFrame : register( b2 )
{
    matrix LocalToWorld;
    
	float4 Light;
};

#include "PointLight.fx"

//--------------------------------------------------------------------------------------
struct VS_INPUT
{
    float2 Tex : TEXCOORD0;
};

struct PS_INPUT
{
    float4 Pos : SV_POSITION;
    float2 Tex : TEXCOORD0;
	float3 WorldPos : TEXCOORD1;
	float3 Norm : TEXCOORD2;
};


//--------------------------------------------------------------------------------------
// Vertex Shader
//--------------------------------------------------------------------------------------
PS_INPUT VS( VS_INPUT input )
{

    PS_INPUT output = (PS_INPUT)0;
	float3 surface = SurfaceFunc(input.Tex);
	output.Norm = normalize(cross(surface`(input.Tex.x), surface`(input.Tex.y)));
    output.WorldPos = mul( float4(surface, 1.0f), LocalToWorld );
    output.Pos = mul( float4(output.WorldPos, 1.0f), WorldToCamera );
    output.Pos = mul( output.Pos, CameraToNDC );
    output.Tex = input.Tex;
    
    return output;
}

//--------------------------------------------------------------------------------------
// Pixel Shader
//--------------------------------------------------------------------------------------
float4 PS( PS_INPUT input) : SV_Target
{

	float2 uv = input.Tex;

	float3 norm = input.Norm;

	norm = mul(norm, (float3x3)LocalToWorld);

	norm = normalize(norm);

//return float4(norm * 0.5 + 0.5.xxx, 1.0);

	float4 pos = float4(input.WorldPos, 1.0f);
	float4 eyeVec = pos - WorldCamPos;
	eyeVec.xyz = reflect(eyeVec.xyz, norm);

	float4 env = cubeMap.Sample( cubeSampler, eyeVec.xyz );

	float3 color = PointLight(pos, norm, WorldCamPos);
//	env.xyz = color;
//	env.xyz *= color;
//	env.xyz = env * 0.2 + color;
	env.xyz = lerp(env.xyz, color, 0.75);

	return float4(env.xyz, 1.0f);
}

#endif // REFSHADER_FX
