
#ifndef REPROJ_FX
#define REPROJ_FX

#include "ReprojRegisters.fx"

cbuffer cbReprojection : register( REG_REPROJECTION_CONST )
{
    float4      m_ParamLine0;   // Blend, unused, Jitter Table Offset x, Jitter Table Offset y.
    float4      m_WindowSize;   // scale/Width, scale/Height, 1.0 / Length, 0.5 / Length
};


Texture2D <float4> previousColor : register( REG_PREVIOUS_COLOR_TEX );
//Texture2DMS <float, 4> previousDepth : register( t3 );

SamplerState samplePrevious : register( REG_PREVIOUS_COLOR_SAMP );

Texture2D<float2> randomOffset : register( REG_RANDOM_TEX );
SamplerState sampleOffset : register( REG_RANDOM_SAMP );

float2 RandomTableIndex(const float2 coord)
{
    return frac(coord /* * m_WindowSize.zz */ + m_WindowSize.ww + m_ParamLine0.zw);
}

float4 VS_Reproject(const float3 localPos, const matrix lastLocalToNDC)
{
	return mul( float4(localPos, 1), lastLocalToNDC );
}

// interpPos has not only had the w-divide applied, but has also already been scaled into viewport coords.
// interpLastPos is still pre-divide homogeneous.
float4 PS_Reproject(const float4 interpPos, const float4 interpLastPos, const float4 currColor)
{
	float4 outputColor = currColor;
	const float4 lastPos = interpLastPos / interpLastPos.w;
	if(lastPos.x * lastPos.x < 1.0f && lastPos.y * lastPos.y < 1.0f)
	{
		float2 colorCoord = lastPos.xy * float2(0.5f, -0.5f) + 0.5f;

		const float4 prevColor = previousColor.Sample(samplePrevious, colorCoord);
		outputColor.xyz = lerp(outputColor.xyz, prevColor.xyz, m_ParamLine0.x);
	}
	return outputColor;
}

#endif // REPROJ_FX
