
#ifndef SVD_FX
#define SVD_FX

void oneSidedJacobiRotation(float2x2 A, const int j, const int k, const float sinTheta, const float cosTheta) 
{
    for (int i = 0; i < 2; i++) 
	{
        const float Aij = A[i][j];
		const float Aik = A[i][k];

        A[i][j] = cosTheta * Aij + sinTheta * Aik;
        A[i][k] = -sinTheta * Aij + cosTheta * Aik;
    }
}

void GivensRotation2x2(float ajj, float ajk, float akk, 
						out float sinTheta, 
						out float cosTheta, 
						out bool needToRotate) 
{
    float product = ajj * akk; //default value that may be overriden if computing small singular values to low relative accuracy            }

    float c, s, t, tau;

    // initialize to neutral values so if calling routine trys to use these values when no need to rotate 
	// it will have no ill effects.
    sinTheta = 0.0f;
    cosTheta = 1.0f;
	needToRotate = false;

	const float eps = 1.0e-4f;
	if(ajk * ajk >= eps * eps * product)
	{
        tau = (ajj - akk) / (2.0f * ajk);
        t = sign(tau) / (abs(tau) + sqrt(1.0f + tau * tau));
        c = 1.0f / sqrt(1.0f + t * t);
        s = c * t;
        sinTheta = s;
        cosTheta = c;
        needToRotate = true;
    }
}

void Jacobi2x2SVD(float2x2 A/*, out float2x2 U*/, out float2 sigma/*, out float2x2 V*/) 
{
    const int size = 2;

	float2x2 V = {
			1.0f, 0.0f, // row 0
			0.0f, 1.0f  // row 1
	};

    int2 indices = int2(0,1);

    float ATAjj = A[0][0] * A[0][0] + A[1][0] * A[1][0];
    float ATAjk = A[0][0] * A[0][1] + A[1][0] * A[1][1];
    float ATAkk = A[0][1] * A[0][1] + A[1][1] * A[1][1];
    float s;
	float c;
    bool needToRotate;

    GivensRotation2x2(ATAjj, ATAjk, ATAkk, s, c, needToRotate);
    if (needToRotate) 
	{
        oneSidedJacobiRotation(A, 0, 1, s, c);
        oneSidedJacobiRotation(V, 0, 1, s, c);
    }

	// Manually unrolling the following loop.
	//for (int i = 0; i < 2; i++) 
	//{
	//	for (int j = 0; j < 2; j++) 
	//	{
	//		sigma[i] += A[j][i] * A[j][i];
	//	}
	//	sigma[i] = sqrt(sigma[i]);
	//}
	float sig0 = sqrt(A[0][0] * A[0][0] + A[1][0] * A[1][0]);
	float sig1 = sqrt(A[0][1] * A[0][1] + A[1][1] * A[1][1]);
	sigma.x = max(sig0, sig1);
	sigma.y = min(sig0, sig1);
#if 0
	if (sigma[0] < sigma[1]) 
	{
        indices[0] = 1;
        indices[1] = 0;
        float temp = sigma[0];
        sigma[0] = sigma[1];
        sigma[1] = temp;

        //dont need to do this even if doing anisotropic filtering. Use the U vectors instead.
        temp = V[0, 0];
        V[0, 0] = V[0, 1];
        V[0, 1] = temp;
        temp = V[1, 0];
        V[1, 0] = V[1, 1];
        V[1, 1] = temp;
    }
    for (int i = 0; i < size; i++) 
	{
        for (int j = 0; j < size; j++) 
		{
            if (sigma[i] != 0) 
			{ 
				U[j, i] = A[j, indices[i]] / sigma[i]; 
			} //columns 0 and 1 are the vectors that represent the axes along which the texture has maximum and minimum stretch.
                //amount of stretch is sigma[0], sigma[1]
            else 
			{ 
				U[j, i] = 0; 
			}
        }
    }
#endif
}


#endif // SVD_FX
