
#define _USE_MATH_DEFINES
#include <math.h>

#include "DXUT.h"

#undef min // use __min instead
#undef max // use __max instead

#include "Accelerator.h"

Accelerator::Accelerator()
    :   m_fMaxRate(1.0f),
        m_fMaxElapsed(0.1f),
        m_fAccel(1.0f)
{
    ResetBase(0.0f);
}

void Accelerator::Reset(float f)
{
    ResetBase(f);
}

void Accelerator::ResetBase(float f)
{
    m_fRate = 0.0f;
    m_fGoal = f;
}

void Accelerator::SetAcceleration(float f) 
{ 
    assert(f > 0); // highly unexpected acceleration parameter.
    m_fAccel = f; 
}

float Accelerator::Update(float currVal, float fElapsed)
{
    float diff = m_fGoal - currVal;
    diff = Normalize(diff);

    fElapsed = Clamp(fElapsed, 0.0f, m_fMaxElapsed);

    if(abs(diff) > m_fRate * m_fRate * 0.5f / m_fAccel)
    {
        m_fRate += m_fAccel * fElapsed * SignOf(diff);
        m_fRate = Clamp(m_fRate, -m_fMaxRate, m_fMaxRate);

        currVal += m_fRate * fElapsed;
    }
    else
    {
        m_fRate += m_fAccel * fElapsed * -SignOf(m_fRate);

        float del = m_fRate * fElapsed;
        if(abs(del) < abs(diff))
        {
            currVal += del;
        }
        else
        {
            m_fRate = 0;
            currVal = m_fGoal;
        }
    }
    currVal = Normalize(currVal);

    return currVal;
}

float AcceleratorYaw::Normalize(float f) const
{
    /// keep values between [-M_PI..M_PI]
    if(f > M_PI)
        f -= 2.0f * M_PI;
    else if( f < -M_PI)
        f += 2.0f * M_PI;
    return f;
}

float AcceleratorPitch::Normalize(float f)
{
    return Clamp(f, -M_PI * 0.5f, M_PI * 0.5f);
}
