
#ifndef ACCELEERATOR_H
#define ACCELEERATOR_H

class Accelerator
{
public:
    Accelerator();

    virtual ~Accelerator() {}

    virtual void Reset(float newValue);

    float Update(float currVal, float fElapsed);

    void SetGoal(float f) { m_fGoal = f; }
    float GetGoal() const { return m_fGoal; }

    void SetAcceleration(float f);
    float GetAcceleration() const { return m_fAccel; }

    void SetMaxRate(float f ) { m_fMaxRate = f; }
    float GetMaxRate() const { return m_fMaxRate; }

    void SetMaxElapsed(float f) { m_fMaxElapsed = f; }
    float GetMaxElapsed() const { return m_fMaxElapsed; }

    static float SignOf(float f)
    {
        return f > 0.0f ? 1.0f : -1.0f;
    }
    static float Clamp(float val, float min, float max)
    {
        return val < min ? min : (val > max ? max : val);
    }

protected:

    virtual float Normalize(float f) const { return f; }

private:

    void ResetBase(float f);

    float m_fRate;
    float m_fAccel;

    float m_fGoal;

    float m_fMaxRate;
    float m_fMaxElapsed;
};

class AcceleratorYaw : public Accelerator
{
protected:
    /// keep values between [-M_PI..M_PI]
    virtual float Normalize(float f) const;
};

class AcceleratorPitch : public Accelerator
{
protected:
    /// keep values between [-M_PI/2..M_PI/2]
    virtual float Normalize(float f);
};

#endif // ACCELEERATOR_H
