

#include <windows.h>
#include <d3d11.h>
#include <d3dx11.h>
#include <d3dcompiler.h>
#include <D3DX11.h>

#define _USE_MATH_DEFINES
#include <math.h>

#include <DXUT.h>
#include "CFlyCamera.h"

#include "BlitEnvMap.h"

extern CFlyCamera          g_Camera;               // A model viewing camera


void BlitEnvMap::SetConstants(ID3D11DeviceContext* immediateContext, ID3D11ShaderResourceView* tex)
{
	immediateContext->PSSetShaderResources( 0, 1, &tex );

	ConstantBuffer buff;
	buff.scaleOffDst = D3DXVECTOR4(
		(m_DstRect.hi.x - m_DstRect.lo.x) * 2.0f,
		(m_DstRect.hi.y - m_DstRect.lo.y) * 2.0f,
		m_DstRect.lo.x * 2.0f - 1.0f,
		m_DstRect.lo.y * 2.0f - 1.0f);

    D3DXMATRIX worldToNDC = *g_Camera.GetViewMatrix() * *g_Camera.GetProjMatrix();
	D3DXMATRIX ndcToWorld;
	D3DXMatrixInverse(&ndcToWorld, NULL, &worldToNDC);
	D3DXMatrixTranspose( &buff.mNDCToWorld, &ndcToWorld);

	D3DXVECTOR3 origin = D3DXVECTOR3(0.0, 0.0, 0.0);
	D3DXMATRIX cameraToWorld;
	D3DXMatrixInverse(&cameraToWorld, NULL, g_Camera.GetViewMatrix());
	D3DXVECTOR4 eyePos;
	D3DXVec3Transform(&eyePos, &origin, &cameraToWorld);
    //eyePos.y += 0.005f;
    buff.eyePosition = eyePos;

	immediateContext->UpdateSubresource( m_pConstantBuffer, 0, NULL, &buff, 0, 0 );
    immediateContext->VSSetConstantBuffers( 0, 1, &m_pConstantBuffer );
    immediateContext->PSSetConstantBuffers( 0, 1, &m_pConstantBuffer );

}

HRESULT BlitEnvMap::InitConstantBuffers(ID3D11Device* pd3dDevice)
{
	HRESULT hr = S_OK;

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(ConstantBuffer);
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;
    hr = pd3dDevice->CreateBuffer( &bd, NULL, &m_pConstantBuffer );
    if( FAILED( hr ) )
        return hr;

	return hr;
}

