
#ifndef BOUNDS_H
#define BOUNDS_H

class CullVolume;

class Bounds
{
public:
    enum CullStatus
    {
        kAllIn,
        kSplit,
        kAllOut
    };
    virtual CullStatus CullTest(const CullVolume& culler) const = 0;

    virtual D3DXVECTOR3 Min() const = 0;
    virtual D3DXVECTOR3 Max() const = 0;

    virtual Bounds& Transform(const D3DXMATRIX& localToWorld) = 0;
};

class SphereBounds : public Bounds
{
public:
    D3DXVECTOR3 Position;
    FLOAT       Radius;

    SphereBounds() : Position(0.0f, 0.0f, 0.0f), Radius(0.0f) {}

    SphereBounds(const D3DXVECTOR3& pos, FLOAT radius)
        : Position(pos), Radius(radius) {}

    virtual CullStatus CullTest(const CullVolume& culler) const;

    virtual D3DXVECTOR3 Min() const { return Position - D3DXVECTOR3(Radius, Radius, Radius); }
    virtual D3DXVECTOR3 Max() const { return Position + D3DXVECTOR3(Radius, Radius, Radius); }

    virtual Bounds& Transform(const D3DXMATRIX& localToWorld);
};

class AABB : public Bounds
{
public:
    AABB() : lo(FLT_MAX, FLT_MAX, FLT_MAX), hi(-FLT_MAX, -FLT_MAX, -FLT_MAX) {}
    AABB(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi) : lo(lo), hi(hi) {}

    virtual CullStatus CullTest(const CullVolume& culler) const;

    virtual D3DXVECTOR3 Min() const { return lo; }
    virtual D3DXVECTOR3 Max() const { return hi; }

    virtual Bounds& Transform(const D3DXMATRIX& localToWorld);

protected:
    D3DXVECTOR3 lo;
    D3DXVECTOR3 hi;
};

class OBB : public Bounds
{
public:
    D3DXVECTOR3 Corners[8];

    OBB() {}
    /// Initialize from AABB
    OBB(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi);
    OBB(const AABB& aabb);

    virtual CullStatus CullTest(const CullVolume& culler) const;

    virtual D3DXVECTOR3 Min() const;
    virtual D3DXVECTOR3 Max() const;
    
    virtual Bounds& Transform(const D3DXMATRIX& localToWorld);

private:
    void Init(const D3DXVECTOR3& lo, const D3DXVECTOR3& hi);
};

#endif //  BOUNDS_H
