
#pragma once
#ifndef ORBIT_CAMERA_H
#define ORBIT_CAMERA_H

#include "DXUTcamera.h"
#include "CameraModeSelect.h"

//--------------------------------------------------------------------------------------
// Simple model viewing camera class that rotates around the object.
//--------------------------------------------------------------------------------------
class CFlyCamera : public CBaseCamera
{
public:
    CFlyCamera();
    ~CFlyCamera();

    // Call these from client and use Get*Matrix() to read new matrices
    virtual LRESULT HandleMessages( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
    virtual VOID    FrameMove( FLOAT fElapsedTime );


    // Functions to change behavior
    virtual VOID    SetDragRect( RECT& rc );
    VOID            Reset();
    VOID            SetViewParams( D3DXVECTOR3* pvEyePt, D3DXVECTOR3* pvLookatPt );
    VOID            SetOffProjParams( FLOAT left, FLOAT right, FLOAT bot, FLOAT top, FLOAT fNearPlane, FLOAT fFarPlane );
    VOID            SetProjection(const D3DXMATRIX* proj);
    VOID            SetButtonMasks( int nRotateModelButtonMask = MOUSE_LEFT_BUTTON, int nZoomButtonMask = MOUSE_WHEEL,
                                    int nRotateCameraButtonMask = MOUSE_RIGHT_BUTTON )
    {
    }
    VOID            SetAttachCameraToModel( BOOL bEnable = false )
    {
    }
    VOID            SetWindow( int nWidth, int nHeight, FLOAT fArcballRadius=0.9f )
    {
		m_nWidth = nWidth; m_nHeight = nHeight;
    }
    VOID            SetRadius( FLOAT fDefaultRadius=5.0f, FLOAT fMinRadius=1.0f, FLOAT fMaxRadius=FLT_MAX )
    {
    }
    VOID            SetModelCenter( D3DXVECTOR3 vModelCenter )
    {
    }
    VOID            SetLimitPitch( BOOL bLimitPitch )
    {
    }
    VOID            SetViewQuat( D3DXQUATERNION q )
    {
    }
    VOID            SetWorldQuat( D3DXQUATERNION q )
    {
        
    }

    CameraModeSelect::eCameraModes GetFlyMode() const { return m_eCameraMode; }
    void SetFlyMode(CameraModeSelect::eCameraModes mode);

    /// These mouse positions are in units [-1..1][-1..1] in the main window.
    const D3DXVECTOR2&  GetMousePosition() const { return mousePos; }
    const D3DXVECTOR2&  GetLastMousePosition() const { return lastPos; }

    VOID                SetLastWorldToNDC(const D3DXMATRIX& w2ndc) { lastWorldToNDC = w2ndc; }
    const D3DXMATRIX&   GetLastWorldToNDC() const { return lastWorldToNDC; }

    bool ToggleBuildLog(const WCHAR* const fileName);
    bool ToggleRunFromLog(const WCHAR* const fileName);

    void SetSpeedGoal(float f);
    float GetSpeedGoal() const;

protected:

    VOID ResetMode();
	VOID ResetGoals();
	VOID SetMatrix();

	int m_nWidth;
	int m_nHeight;

    CameraModeSelect::eCameraModes m_eCameraMode;
    CameraMode*     m_pCameraMode;

	D3DXVECTOR2 mousePos;
	D3DXVECTOR2 lastPos;

    D3DXMATRIX lastWorldToNDC;
};


#endif // ORBIT_CAMERA_H
