
#define _USE_MATH_DEFINES
#include <math.h>

#include "DXUT.h"

#undef min // use __min instead
#undef max // use __max instead

#include "CameraCircle.h"
#include "../Objects/Scape.h"

CameraCircle::CameraCircle()
{
    ResetAll();
}

void CameraCircle::Reset()
{
    CameraDrive::Reset();
    ResetAll();
}

D3DXVECTOR3 CameraCircle::MoveToCircle(const D3DXVECTOR3& pos, float rate) const
{
    D3DXVECTOR3 dir = pos;
    D3DXVec3Normalize(&dir, &dir);
    dir *= m_fRadius;

    return pos + rate * (dir - pos);
}

D3DXVECTOR3 CameraCircle::GetTravelDirection(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const
{
    D3DXVECTOR3 dir;
    D3DXVECTOR3 up(0.0f, 1.0f, 0.0f);
    D3DXVec3Cross(&dir, &eye, &up);

    D3DXVec3Normalize(&dir, &dir);
    return dir;
}

void CameraCircle::Advance(const CFlyCamera& camera, const FLOAT fSecs)
{
    const FLOAT travelYaw0 = GetYaw(GetTravelDirection(m_vEyePt, m_vLookAtPt));

    CameraWalk::Advance(camera, fSecs);


    static float posRate = 0.8f;
    m_vEyePt = MoveToCircle(m_vEyePt, posRate);

    const FLOAT travelYaw1 = GetYaw(GetTravelDirection(m_vEyePt, m_vLookAtPt));

    D3DXVECTOR3 dir = m_vLookAtPt - m_vEyePt;

    dir = RotateYaw(dir, travelYaw1 - travelYaw0);

    m_vLookAtPt = m_vEyePt + dir;

    BlendLookAtToTravel(fSecs);

    //m_vEyePt.y = FitToGround(m_vEyePt, 1.0f);
    //m_vLookAtPt.y = FitToGround(m_vLookAtPt, 0.25f);
}

void CameraCircle::ResetAll()
{
    extern POINT numModels;
    extern float modelSpacing;

    m_fRadius = numModels.x * modelSpacing * 0.7f * 0.5f;
    const float kMinRadius = 10.0f;
    if(m_fRadius < kMinRadius)
        m_fRadius = kMinRadius;

    m_vEyePt = D3DXVECTOR3(m_fRadius, 30.0f, 0.0f);
    m_vLookAtPt = m_vEyePt;
    m_vLookAtPt.z += 5.0f;
}