
#ifndef CAMERA_DRIVE_H
#define CAMERA_DRIVE_H

#include <vector>
#include "CameraMode.h"

class Object;

class CameraDrive : public CameraMode
{
public:
    CameraDrive();

    virtual void Reset();

    virtual void Init(const CFlyCamera& camera);

    virtual void Advance(const CFlyCamera& camera, const FLOAT fSecs);

    virtual D3DXVECTOR3 GetEyePt() const { return m_vEyePt; }
    virtual D3DXVECTOR3 GetLookAtPt() const { return m_vLookAtPt; }
    virtual D3DXVECTOR3 GetUpDir() const { return m_vUpDir; }

    LogMode GetLogMode() const { return m_LogMode; }
    virtual bool BeginBuildLog();
    virtual bool EndBuildLog(const WCHAR* const fileName);
    virtual bool BeginFromLog(const WCHAR* const fileName);
    virtual bool EndFromLog();

    virtual void SetSpeedGoal(float f);
    virtual float GetSpeedGoal() const;

protected:

    virtual VOID ResetGoals();
    virtual VOID UpdateGoals(const D3DXVECTOR2& mousePos, const D3DXVECTOR2& lastPos, const FLOAT fElapsedTime);
    virtual VOID AdvanceToGoals(const FLOAT fElapsedTime);
    virtual D3DXVECTOR3 UpdateManuals(const D3DXVECTOR3& eye, const D3DXVECTOR3& dir, const D3DXVECTOR3& bin, const FLOAT fElapsedTime);
    
    virtual D3DXVECTOR3 GetTravelDirection(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const;
    virtual D3DXVECTOR3 GetLookAtDirection(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const;
    virtual FLOAT GetFollowDistance(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const;

    virtual float FitToGround(D3DXVECTOR3& pos, const float radius) const;

    virtual void BlendLookAtToTravel(const FLOAT fSecs);

    D3DXVECTOR3 BlendToGoal(const D3DXVECTOR3& curr, const D3DXVECTOR3& goal, const FLOAT fElapsed, const D3DXVECTOR2& rate) const;

    FLOAT GetYaw(const D3DXVECTOR3& lookAtDir) const;
    D3DXVECTOR3 SetYaw(const FLOAT yaw, const FLOAT up) const;
    FLOAT BlendYaw(FLOAT lookAtYaw, const FLOAT travelYaw, const FLOAT fSecs, const FLOAT yawRate) const;
    D3DXVECTOR3 RotateYaw(const D3DXVECTOR3& dir, const FLOAT rads) const;
    FLOAT GetPitch(const D3DXVECTOR3& lookAtDir) const;
    FLOAT SetPitch(const FLOAT pitch) const;
    FLOAT BlendPitch(FLOAT lookAtPitch, const FLOAT travelPitch, const FLOAT fSecs, const FLOAT pitchRate) const;

    float QueryTerrainHeight(const D3DXVECTOR3& pos) const;
    float CheckCollision(const D3DXVECTOR3& eyePt, const D3DXVECTOR3& travel, const FLOAT elapsed) const;
    float CheckObjectHit(const Object* object, 
                         const D3DXVECTOR3& pos,
                         const D3DXVECTOR3& biAxis, const float extent,
                         const D3DXVECTOR3& axis, const float length) const;

    bool FromLog() const { return m_LogMode == kLogFrom; }
    bool ToLog() const { return m_LogMode == kLogBuild; }
    void AddToLog();
    bool SetFromLog();
    bool LoadLog(const WCHAR* const fileName);
    bool SaveLog(const WCHAR* const fileName) const;
    void SaveLogTime();
    FLOAT speedGoal;
	FLOAT speedCurr;
	FLOAT speedAcc;

	FLOAT yawGoal;
	FLOAT yawCurr;
	FLOAT yawAcc;

	FLOAT pitchGoal;
	FLOAT pitchCurr;
	FLOAT pitchAcc;

    BOOL inDrag;

    D3DXVECTOR3 m_vEyePt;
    D3DXVECTOR3 m_vLookAtPt;
    D3DXVECTOR3 m_vUpDir;

	D3DXVECTOR2 mousePos;
	D3DXVECTOR2 lastPos;

    struct LogEntry
    {
        D3DXVECTOR3 eyePt;
        D3DXVECTOR3 lookAtPt;
        D3DXVECTOR3 upDir;
    };
    std::vector<LogEntry> log;
    LogMode m_LogMode;
    int m_LogCurrent;
    int m_LogLoop;
    double m_LogTime;
};

#endif // CAMERA_DRIVE_H
