

#define _USE_MATH_DEFINES
#include <math.h>

#include "DXUT.h"

#undef min // use __min instead
#undef max // use __max instead

#include "CameraPan.h"

float CameraPan::m_fFrequency = 0.25f;
float CameraPan::m_fAmplitude = M_PI / 6.0;

CameraPan::CameraPan()
    :   m_dTotalTime(0.0)
{
}

void CameraPan::Advance( const CFlyCamera& camera, const FLOAT fSecs )
{
    CameraRandom::Advance(camera, fSecs);
    const float travelYaw = GetYaw( GetTravelDirection(GetEyePt(), GetLookAtPt()) );

    m_dTotalTime += fSecs * m_fFrequency * 2.0f * float(M_PI);
    m_dTotalTime = fmod(m_dTotalTime, M_PI * 2.0);
    const double sinTime = sin(m_dTotalTime);
    const float yaw = float(sinTime * m_fAmplitude) + travelYaw;

    const D3DXVECTOR3 lookAtDir = SetYaw(yaw, 0.0f) * GetFollowDistance(m_vEyePt, m_vLookAtPt);
    m_vLookAtPt.x = m_vEyePt.x + lookAtDir.x;
    m_vLookAtPt.z = m_vEyePt.z + lookAtDir.z;

    const float lookAtY = FitToGround(m_vLookAtPt, 0.25f);
    m_LookAtAccel.SetGoal(lookAtY);
    m_vLookAtPt.y = m_LookAtAccel.Update(m_vLookAtPt.y, fSecs);
}
