
#define _USE_MATH_DEFINES
#include <math.h>

#include "DXUT.h"

#undef min // use __min instead
#undef max // use __max instead

#include "CameraRandom.h"
#include "../Objects/Scape.h"
#include "CFlyCamera.h"

static float RndZeroToOne()
{
    return float(double(rand()) / double(RAND_MAX));
}

static float RndMinToMax(const float min, const float max)
{
    return min + RndZeroToOne() * (max - min);
}

CameraRandom::CameraRandom()
    :   secsOfTravel(0.0f),
        timeTillTurn(0.0f),
        currYaw(0.0f)
{
    m_HeightAccel.SetAcceleration(8.0f);
    m_HeightAccel.SetMaxRate(10.0f);
}

BOOL CameraRandom::NeedNewDirection(const FLOAT fSecs)
{
    return (secsOfTravel += fSecs) > TimeTillTurn();
}

BOOL CameraRandom::OutGoing(const D3DXVECTOR3& eyePt) const
{
    extern POINT numModels;
    extern float modelSpacing;
    
    const float outerEdge = numModels.x * modelSpacing * 0.25f;

    if(eyePt.x > outerEdge)
        return false;
    if(eyePt.z > outerEdge)
        return false;
    if(eyePt.x < -outerEdge)
        return false;
    if(eyePt.z < -outerEdge)
        return false;

    return true;
}
static FLOAT NormalizeYaw(FLOAT yaw)
{
    if(yaw > M_PI)
        yaw -= 2.0f * M_PI;
    else if(yaw < -M_PI)
        yaw += 2.0f * M_PI;
    return yaw;
}
FLOAT CameraRandom::NewDirection(const D3DXVECTOR3& eyePt)
{
    FLOAT newYaw = m_YawAccel.GetGoal();

    const BOOL outGoing = OutGoing(eyePt);
    if(!outGoing)
    {
        newYaw = GetYaw(-eyePt);
    }

    const float rnd = RndMinToMax(-1.0f, 1.0f);
    static float kPower = 1.0f;
    const float rndSqrt = pow(abs(rnd), kPower) * Accelerator::SignOf(rnd);
    const float rotRads = float(rndSqrt * M_PI * 0.25f);
    newYaw += rotRads;

    /// Normalize to [-M_PI..M_PI], just for neatness.
    newYaw = NormalizeYaw(newYaw);

    secsOfTravel = 0.0f;

    const float kMinSecs = 0.5f;
    const float kMaxSecs = 1.0f;
    timeTillTurn = RndMinToMax(kMinSecs, kMaxSecs);

    return newYaw;
}

VOID CameraRandom::CheckTravelDirection(const D3DXVECTOR3& eyePt, const FLOAT fSecs)
{
    if((GetSpeedGoal() != 0) && NeedNewDirection(fSecs))
    {
        float goalYaw = NewDirection(eyePt);
        m_YawAccel.SetGoal(goalYaw);
    }
}

void CameraRandom::SetAccel()
{
    const float kMinSpeed = 1.0f;
    const float kMaxSpeed = 10.0f;
    float parm = (speedCurr - kMinSpeed) / (kMaxSpeed - kMinSpeed);
    parm = Accelerator::Clamp(parm, 0.0f, 1.0f);

    const float kMinAccel = 2.0f;
    const float kMaxAccel = 8.0f;
    const float accel = kMinAccel + parm * (kMaxAccel - kMinAccel);
    m_HeightAccel.SetAcceleration(accel);

    const float kMinYaw = 0.1f;
    const float kMaxYaw = 1.0f;
    const float yawAccel = kMinYaw + parm * parm * (kMaxYaw - kMinYaw);
    m_YawAccel.SetAcceleration(yawAccel);
}

void CameraRandom::Advance(const CFlyCamera& camera, const FLOAT fSecs)
{
    SetAccel();

    CheckTravelDirection(*camera.GetEyePt(), fSecs);

    currYaw = m_YawAccel.Update(currYaw, fSecs);

    CameraDrive::Advance(camera, fSecs);


    const float height = CheckCollision(m_vEyePt, GetTravelDirection(m_vEyePt, m_vLookAtPt), fSecs);

    m_HeightAccel.SetGoal(height);
    m_vEyePt.y = m_HeightAccel.Update(m_vEyePt.y, fSecs);

    D3DXVECTOR3 lookAtDir = SetYaw(currYaw, 0.0f) * GetFollowDistance(m_vEyePt, m_vLookAtPt);
    m_vLookAtPt.x = m_vEyePt.x + lookAtDir.x;
    m_vLookAtPt.z = m_vEyePt.z + lookAtDir.z;

    const float lookAtY = FitToGround(m_vLookAtPt, 0.0f) + 0.9f;
    m_LookAtAccel.SetGoal(lookAtY);
    m_vLookAtPt.y = m_LookAtAccel.Update(m_vLookAtPt.y, fSecs);
}

D3DXVECTOR3 CameraRandom::GetTravelDirection(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const
{
    return SetYaw(currYaw, 0.0f);
}

