
#define _USE_MATH_DEFINES
#include <math.h>

#include "DXUT.h"

#undef min // use __min instead
#undef max // use __max instead

#include "CameraWalk.h"
#include "../Objects/Scape.h"

CameraWalk::CameraWalk()
    :   CameraDrive(),
        m_HeightAccel()
{
    m_HeightAccel.SetAcceleration(8.0f);
    m_HeightAccel.SetMaxRate(10.0f);
}

VOID CameraWalk::Advance(const CFlyCamera& camera, const FLOAT fElapsedTime)
{
    SetHeightAccel();

    CameraDrive::Advance(camera, fElapsedTime);

    const float height = CheckCollision(m_vEyePt, GetTravelDirection(m_vEyePt, m_vLookAtPt), fElapsedTime);

    m_HeightAccel.SetGoal(height);
    m_vEyePt.y = m_HeightAccel.Update(m_vEyePt.y, fElapsedTime);

    const float lookAtY = FitToGround(m_vLookAtPt, 0.0f) + 0.95f;
    m_LookAtAccel.SetGoal(lookAtY);
    m_vLookAtPt.y = m_LookAtAccel.Update(m_vLookAtPt.y, fElapsedTime);
}

D3DXVECTOR3 CameraWalk::GetTravelDirection(const D3DXVECTOR3& eye, const D3DXVECTOR3& lookAt) const
{
    D3DXVECTOR3 travel = lookAt - eye;
    travel.y = 0;
    D3DXVec3Normalize(&travel, &travel);
    return travel;
}

void CameraWalk::SetHeightAccel()
{
    const float kMinSpeed = 1.0f;
    const float kMaxSpeed = 10.0f;
    const float kMinAccel = 2.0f;
    const float kMaxAccel = 8.0f;
    float parm = (speedCurr - kMinSpeed) / (kMaxSpeed - kMinSpeed);
    parm = Accelerator::Clamp(parm, 0.0f, 1.0f);
    float accel = kMinAccel + parm * (kMaxAccel - kMinAccel);
    m_HeightAccel.SetAcceleration(accel);
}

