
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <share.h>
#include <assert.h>

#include "CompiledShader.h"
#include "FileDateCheck.h"

static bool CompiledName(const WCHAR* fileName, const WCHAR* profile, WCHAR* compName)
{
	const int len = wcslen(fileName);

	int iDot = len;
	while(--iDot >= 0)
	{
		if(fileName[iDot] == L'.')
			break;
	}
	const bool error = iDot < 0;
	if(error)
	{
		iDot = len;
	}

	for(int i = 0; i < iDot; ++i)
	{
		compName[i] = fileName[i];
	}
	compName[iDot + 0] = L'.';
	compName[iDot + 1] = profile[0];
	compName[iDot + 2] = profile[1];
	compName[iDot + 3] = L'o';
	compName[iDot + 4] = 0;
	assert(wcscmp(fileName, compName));

	return error;
}


CompiledShader::CompiledShader(const WCHAR* fileName,
				const WCHAR* profile,
				const WCHAR* entryPoint)
{

	WCHAR tempFile[MAX_PATH];
	CompiledName(fileName, profile, tempFile);

	bool error = false;

//	FileDateCheck check(fileName, tempFile);

//	if(CompiledOutOfDate(fileName, tempFile))
	if(FileDateCheck(fileName, tempFile).OutOfDate())
	{
		WCHAR command[MAX_PATH];

		WCHAR flags[MAX_PATH];
		flags[0] = 0;
		wcscpy_s(flags, L"/Ges");
#if defined(DEBUG) || defined(_DEBUG)
//		wcscat_s(flags, L" /Od /Zi");
#endif // DEBUG

#define MAF_ALTERNATE_COMPILER
#ifdef MAF_ALTERNATE_COMPILER
        const WCHAR* fxc = L"Shaders\\fxc_special";
#else // MAF_ALTERNATE_COMPILER
		const WCHAR* fxc = L"Shaders\\fxc";
#endif // MAF_ALTERNATE_COMPILER

		swprintf_s(command, L"%s /FeShaders\\errors.txt /Fo%s /T%s /E%s %s %s",
			fxc,
			tempFile,
			profile,
			entryPoint,
			flags,
			fileName);
		//swprintf_s(command, L"Shaders\\fxc /Fctemp.txt /T%s /E%s %s %s",
		//	profile,
		//	entryPoint,
		//	flags,
		//	fileName);

//		_wsystem(L"Shaders\\fxc-deriv /? > zzz.txt");
		int retVal = _wsystem(command);
		error = retVal != 0;
	}

	/// Check the return code, will be zero on success. We might
	/// fail, but have an old temp.byt lying around.
	if(!error)
	{
		int fd;
		_wsopen_s( &fd, tempFile, _O_RDONLY, _SH_DENYNO, 0 );

		if(fd >= 0)
		{
			this->byteCodeLength = _lseek(fd, 0, SEEK_END);

			this->byteCodes = new char[this->byteCodeLength];

			_lseek(fd, 0, SEEK_SET);

			_read(fd, this->byteCodes, this->byteCodeLength);

			_close(fd);
		}
	}
}

CompiledShader::~CompiledShader()
{
	Release();
}
