
#include <windows.h>
#include <d3dx11.h>
#include <D3DX10math.h>
#include <assert.h>

#include "CullVolume.h"

void CullVolume::Setup(const D3DXMATRIX& l2n)
{
    planes[0] = D3DXVECTOR4(
                    l2n.m[0][3] - l2n.m[0][0],
                    l2n.m[1][3] - l2n.m[1][0],
                    l2n.m[2][3] - l2n.m[2][0],
                    l2n.m[3][3] - l2n.m[3][0]);
    planes[1] = D3DXVECTOR4(
                    l2n.m[0][3] + l2n.m[0][0],
                    l2n.m[1][3] + l2n.m[1][0],
                    l2n.m[2][3] + l2n.m[2][0],
                    l2n.m[3][3] + l2n.m[3][0]);

    planes[2] = D3DXVECTOR4(
                    l2n.m[0][3] - l2n.m[0][1],
                    l2n.m[1][3] - l2n.m[1][1],
                    l2n.m[2][3] - l2n.m[2][1],
                    l2n.m[3][3] - l2n.m[3][1]);
    planes[3] = D3DXVECTOR4(
                    l2n.m[0][3] + l2n.m[0][1],
                    l2n.m[1][3] + l2n.m[1][1],
                    l2n.m[2][3] + l2n.m[2][1],
                    l2n.m[3][3] + l2n.m[3][1]);

    planes[4] = D3DXVECTOR4(
                    l2n.m[0][3] - l2n.m[0][2],
                    l2n.m[1][3] - l2n.m[1][2],
                    l2n.m[2][3] - l2n.m[2][2],
                    l2n.m[3][3] - l2n.m[3][2]);
    planes[5] = D3DXVECTOR4(
                    l2n.m[0][2],
                    l2n.m[1][2],
                    l2n.m[2][2],
                    l2n.m[3][2]);

    for(int i = 0; i < 6; ++i)
    {
        D3DXVECTOR4 norm(planes[i].x, planes[i].y, planes[i].z, 0.0f);
        double len = D3DXVec4Length(&norm);
        if(len > 0)
        {
            len = 1.0 / len;
            planes[i].x = float(planes[i].x * len);
            planes[i].y = float(planes[i].y * len);
            planes[i].z = float(planes[i].z * len);
            planes[i].w = float(planes[i].w * len);
        }
    }
}

//bool CullVolume::IsCulled(const Object::SphereBound& worldSphere) const
//{
//#if 0
//    const D3DXVECTOR3& pos = worldSphere.Position;
//    for(int i = 0; i < 6; i++)
//    {
//        const float dot = pos.x * planes[i].x
//                        + pos.y * planes[i].y
//                        + pos.z * planes[i].z
//                        + planes[i].w
//                        + worldSphere.Radius; 
//
//        if(dot < 0)
//            return true;
//    }
//    return false;
//#else
//    const float dist = CheckDist(worldSphere.Position);
//    return dist + worldSphere.Radius < 0;
//#endif
//}

D3DXVECTOR2 CullVolume::CheckDist(const D3DXVECTOR3* const worldPositions, int num) const
{
    D3DXVECTOR2 ret(FLT_MAX, FLT_MAX);
    assert(num > 0);
    for(int iPlane = 0; iPlane < 6; ++iPlane)
    {
        float minDot = worldPositions[0].x * planes[iPlane].x
                     + worldPositions[0].y * planes[iPlane].y
                     + worldPositions[0].z * planes[iPlane].z
                                  + planes[iPlane].w;
        float maxDot = minDot;

        for(int iPos = 1; iPos < num; ++iPos)
        {
            const D3DXVECTOR3& worldPos = worldPositions[iPos];
            const float dot = worldPos.x * planes[iPlane].x
                            + worldPos.y * planes[iPlane].y
                            + worldPos.z * planes[iPlane].z
                                         + planes[iPlane].w;

            minDot = __min(dot, minDot);
            maxDot = __max(dot, maxDot);
        }
        ret.x = __min(ret.x, minDot);
        ret.y = __min(ret.y, maxDot);
    }
    return ret;
}

float CullVolume::CheckDist(const D3DXVECTOR3& worldPos) const
{
    float minDot = worldPos.x * planes[0].x
                 + worldPos.y * planes[0].y
                 + worldPos.z * planes[0].z
                 + planes[0].w;
    for(int i = 1; i < 6; i++)
    {
        const float dot = worldPos.x * planes[i].x
                        + worldPos.y * planes[i].y
                        + worldPos.z * planes[i].z
                        + planes[i].w;

        minDot = __min(dot, minDot);
    }
    return minDot;
}
