
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <share.h>
#include <assert.h>

#include <algorithm>

#include <iostream>
#include <fstream>

#include "FileDateCheck.h"

using namespace std;

static const streamsize kBuffLen = 1024;

FileDateCheck::FileDateCheck()
	: outOfDate(false)
{
}

FileDateCheck::FileDateCheck(const WCHAR* srcFile, const WCHAR* dstFile)
	: outOfDate(false)
{
	Check(srcFile, dstFile);
}

FileDateCheck::~FileDateCheck()
{
	ClearDependencies();
}

bool FileDateCheck::Check(const WCHAR* srcFile, const WCHAR* dstFile)
{
	SetFolder(srcFile);

	GetDependencies(srcFile);

	CheckDependencies(dstFile);

	ClearDependencies();

	return outOfDate;
}

void FileDateCheck::SetFolder(const WCHAR* srcFile)
{
	_wsplitpath_s( srcFile, NULL, 0, folder, MAX_PATH, NULL, 0, NULL, 0 );

}

bool FileDateCheck::ExtractIncludeName(const WCHAR* lineBuff, WCHAR* nameBuff) const
{
	const streamsize kBuffLen = 1024;

	// Find the opening delimiter.
	const WCHAR* p = lineBuff;
	while(*p && (*p != L'<') && (*p != L'"'))
	{
		++p;
	}
	if(!*p)
	{
		/// Never found our file token.
		return false;
	}
	// Figure out which closing delimiter we'll be looking for.
	WCHAR del = 0;
	if(*p == L'<')
		del = L'>';
	else if(*p == L'"')
		del = L'"';
	else
	{
		assert(false); //
		return false;
	}
	++p; // advance past the opening delimiter.
	wcscpy_s(nameBuff, kBuffLen, folder);
	WCHAR* dstP = nameBuff + wcslen(folder);
	while(*p != del)
	{
		*dstP++ = *p++;
	}
	*dstP = 0;

	return true;
}

bool FileDateCheck::AlreadyDependent(const WCHAR* depName) const
{
	for(NameConstIter iter = dependencies.begin(); iter != dependencies.end(); ++iter)
	{
		const WCHAR* iterName = *iter;
		if(!wcscmp(iterName, depName))
			return true;
	}
	return false;
}

void FileDateCheck::GetDependencies(const WCHAR* srcFile)
{
	// We're obviously dependent on this source file.
	dependencies.push_back(_wcsdup(srcFile));

	// Now examine this source file looking for more included dependencies.
	wifstream src(srcFile);

	WCHAR lineBuff[kBuffLen];

	if(src.good())
	{
		// Read each line in the file looking for #includes
		while(!src.eof() && src.good() && !src.bad())
		{
			src.getline(lineBuff, kBuffLen);

			// If it's an include, extract the file name
			const WCHAR* kInc = L"#include";
			WCHAR* foundInc = wcsstr(lineBuff, kInc);
			if(foundInc)
			{
				WCHAR childName[kBuffLen];
				// If the file isn't already in dependencies, recur on it.
				if(ExtractIncludeName(foundInc + wcslen(kInc), childName))
				{

					/// If we were expecting more than 2 or 3 dependencies, we could sort the
					/// list and use a faster search.
					if(!AlreadyDependent(childName))
					{
						GetDependencies(childName);
					}
				}
			}
		}
	}
//    src.close();
}

void FileDateCheck::CheckDependencies(const WCHAR* dstFile)
{
	FILETIME dstTime;
	if(!GetTimeStamp(dstFile, dstTime))
	{
		/// dstFile doesn't exist, so it is clearly out of date.
		outOfDate = true;
		return;
	}

	for(NameIter iter = dependencies.begin(); iter != dependencies.end(); ++iter)
	{
		WCHAR* dependencyName = *iter;
		FILETIME depTime;
		if(GetTimeStamp(dependencyName, depTime))
		{
			const LONG compare = CompareFileTime(&depTime, &dstTime);

			if(compare >= 0)
			{
				outOfDate = true;
				return;
			}
		}
	}
}

void FileDateCheck::ClearDependencies()
{
	for(NameIter iter = dependencies.begin(); iter != dependencies.end(); ++iter)
	{
		WCHAR* name = *iter;
		free(name);
	}
	dependencies.clear();
}

bool FileDateCheck::GetTimeStamp(const WCHAR* fileName, FILETIME& fileTime) const
{
	HANDLE fileHandle = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if(fileHandle == INVALID_HANDLE_VALUE)
		return false;

	if(!GetFileTime(fileHandle, NULL, NULL, &fileTime))
		return false;

	CloseHandle(fileHandle);

	return true;
}
