#include "pch.h"


#include <windows.h>
//#include <d3d11.h>
//#include <d3dx11.h>
//#include <d3dcompiler.h>
//#include <D3DX10math.h>

#define _USE_MATH_DEFINES
#include <math.h>

//#include "MAFMath.h"

#include "Utility.h"
#include "BlitComposite.h"
#include "SaveRenderState.h"

BlitComposite::ShowLevelMode BlitComposite::s_showLevelMode(kShowNone);
UINT BlitComposite::s_showLevelMask(0xff);


BlitComposite::BlitComposite()
	:   BlitQuad(),
        m_iNumLayers(0),
        m_fDecay(0.6f)
{
    for(int i = 0; i < kMaxLayers; ++i)
    {
        m_aMsaa[i] = false;
        m_aTextures[i] = NULL;
    }
}

BlitComposite::~BlitComposite()
{
}

HRESULT BlitComposite::InitConstantBuffers(ID3D11Device* pd3dDevice)
{
	HRESULT hr = S_OK;

    // Create the constant buffers
    D3D11_BUFFER_DESC bd;
    ZeroMemory( &bd, sizeof(bd) );

    bd.Usage = D3D11_USAGE_DEFAULT;
    bd.ByteWidth = sizeof(ConstantBuffer);
    bd.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    bd.CPUAccessFlags = 0;
    hr = pd3dDevice->CreateBuffer( &bd, NULL, &m_pConstantBuffer );
    if( FAILED( hr ) )
        return hr;

	return hr;
}

void BlitComposite::SetJitterOffsets(const XMFLOAT2 offs[kMaxLayers])
{
    for(int i = 0; i < kMaxLayers; ++i)
    {
        m_aJitterOffsets[i] = XMFLOAT2(offs[i].x / m_aSrcSizes[i].x, offs[i].y / m_aSrcSizes[i].y);
    }
}

void BlitComposite::SetConstants(ID3D11DeviceContext* immediateContext, ID3D11ShaderResourceView* baseTex)
{
	immediateContext->PSSetShaderResources( 0, 1, &baseTex );
	immediateContext->PSSetShaderResources( 1, kMaxLayers, m_aTextures );

	ConstantBuffer buff;
	buff.m_ScaleOffDst = XMLoadFloat4(&XMFLOAT4(
		(m_DstRect.hi.x - m_DstRect.lo.x) * 2.0f,
		(m_DstRect.hi.y - m_DstRect.lo.y) * 2.0f,
		m_DstRect.lo.x * 2.0f - 1.0f,
		m_DstRect.lo.y * 2.0f - 1.0f));

	for(int i = 0; i < kMaxLayers; ++i)
	{
		const BlitRect& dstRect = m_aDstRects[i];

		if(dstRect.hi.x > dstRect.lo.x)
		{
			buff.m_aScaleOffDomainToTex[i] = XMLoadFloat4(&XMFLOAT4(
				1.0f / (dstRect.hi.x - dstRect.lo.x),
				-1.0f / (dstRect.hi.y - dstRect.lo.y),
				-dstRect.lo.x / (dstRect.hi.x - dstRect.lo.x) - m_aJitterOffsets[i].x,
				dstRect.hi.y / (dstRect.hi.y - dstRect.lo.y) + m_aJitterOffsets[i].y));

            buff.m_aPixOffset[i] = XMLoadFloat4(&XMFLOAT4(
                0.5f / m_aSrcSizes[i].x,
                0.5f / m_aSrcSizes[i].y,
                -0.5f / m_aSrcSizes[i].x,
                -0.5f / m_aSrcSizes[i].y ));
		}
		else
		{
			buff.m_aScaleOffDomainToTex[i] = XMLoadFloat4(&XMFLOAT4(0.0f, 0.0f, 0.5f, 0.5f));

            buff.m_aPixOffset[i] = XMLoadFloat4(&XMFLOAT4(0.0f, 0.0f, 0.0f, 0.0f));
		}
	}
    buff.m_aPixOffset[kBaseLayer] = XMLoadFloat4(&XMFLOAT4(
        0.5f / m_aSrcSizes[kBaseLayer].x,
        0.5f / m_aSrcSizes[kBaseLayer].y,
        -0.5f / m_aSrcSizes[kBaseLayer].x,
        -0.5f / m_aSrcSizes[kBaseLayer].y));

    buff.m_aScaleOffDomainToTex[kBaseLayer] = XMLoadFloat4(&XMFLOAT4(
        1.0f, -1.0f, -m_aJitterOffsets[kBaseLayer].x, 1.0f + m_aJitterOffsets[kBaseLayer].y));

	buff.m_Decay = XMLoadFloat4(&XMFLOAT4(m_fDecay, 1.0f, 0.0f, 0.0f));

	immediateContext->UpdateSubresource( m_pConstantBuffer, 0, NULL, &buff, 0, 0 );
    immediateContext->VSSetConstantBuffers( 0, 1, &m_pConstantBuffer );
    immediateContext->PSSetConstantBuffers( 0, 1, &m_pConstantBuffer );

}

void BlitComposite::UnSetConstants(ID3D11DeviceContext* immediateContext)
{
	ID3D11ShaderResourceView* nullArray[kMaxLayers + 1] = { NULL };
	immediateContext->PSSetShaderResources( 0, kMaxLayers + 1, nullArray );
}


void BlitComposite::SetTextures(ID3D11ShaderResourceView* texList[kMaxLayers], const XMFLOAT2 sizes[kMaxLayers])
{
	for(int i = 0; i < kMaxLayers; ++i)
	{
		m_aTextures[i] = texList[i];
        m_aSrcSizes[i] = sizes[i];
	}
}

void BlitComposite::SetDstRects(BlitRect* rects, int numRects)
{
	numRects = numRects <= kMaxLayers ? numRects : kMaxLayers;
	for(int i = 0; i < numRects; ++i)
	{
		m_aDstRects[i] = rects[i];
	}
    for(int i = numRects; i < kMaxLayers; ++i)
    {
        m_aDstRects[i] = BlitRect(0.0f, 0.0f, 0.0f, 0.0f);
    }
}

static D3D_SHADER_MACRO Macro(const CHAR* name, const CHAR* def)
{
    D3D_SHADER_MACRO macro;
    macro.Name = name;
    macro.Definition = def;
    return macro;
 }
void BlitComposite::AppendPixelDefines(std::vector<D3D_SHADER_MACRO>& defines)
{
    static const LPCSTR nameInner = "MAF_MS_INNER";
    static const LPCSTR nameMid = "MAF_MS_MID";
    static const LPCSTR nameBase = "MAF_MS_BASE";
    static const LPCSTR nameNumLayers = "MAF_NUM_LAYERS";
    static const LPCSTR nameShowLevels = "MAF_SHOW_LEVELS";
    static const LPCSTR nameColorLevels = "MAF_COLOR_LEVELS";
    static const LPCSTR valTrue = "TRUE";
    static const LPCSTR valNumber[] = { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" };
    if(m_aMsaa[0])
    {
        defines.push_back(Macro(nameInner, valTrue));
    }
    if(m_aMsaa[1])
    {
        defines.push_back(Macro(nameMid, valTrue));
    }
    if(m_aMsaa[kBaseLayer])
    {
        defines.push_back(Macro(nameBase, valTrue));
    }
    const int showLevelMask = __min(GetShowLevelMask(), 3);
    switch(s_showLevelMode)
    {
    case kShowDarken:
        defines.push_back(Macro(nameShowLevels, valNumber[showLevelMask]));
        break;
    case kShowColor:
        defines.push_back(Macro(nameColorLevels, valNumber[showLevelMask]));
        break;
    case kShowNone:
    default:
        break;
    }
    defines.push_back(Macro(nameNumLayers, valNumber[GetNumLayers()]));

    ForceRecompile();
}

void BlitComposite::SetShowLevelMode( ShowLevelMode m, const UINT mask/*=0xff*/ )
{
    s_showLevelMode = m; s_showLevelMask = mask;
}
