
#ifndef BLIT_COMPOSITE_H
#define BLIT_COMPOSITE_H

#include "BlitQuad.h"

class BlitComposite : public BlitQuad
{
public:
	BlitComposite();

	virtual ~BlitComposite();

	enum 
	{
		kMaxLayers = 3, // Changing this requires a change to BlitComposite.fx
        kBaseLayer = kMaxLayers - 1
	};
    enum ShowLevelMode
    {
        kShowNone,
        kShowDarken,
        kShowColor
    };

    // get the number of layers to composite.
    int GetNumLayers() const { return m_iNumLayers; }
    // this must be followed by a call to Init.
    void SetNumLayers(int n) { m_iNumLayers = n; }

    // Set the array of textures to be composited along with their sizes.
	void SetTextures(ID3D11ShaderResourceView* texList[kMaxLayers], const XMFLOAT2 srcSizes[kMaxLayers]);

    // Set the destination rect on the destination rendertarget for each of the textures to be composited.
	void SetDstRects(BlitRect* rects, int numRects);

    // Return whether to use the average of 4 samples when sampling the source texture.
    bool GetMSAA(int layer) const { return m_aMsaa[layer]; }

    // Set whether to use the average of 4 samples when sampling the source texture.
    void SetMSAA(int layer, bool on) { m_aMsaa[layer] = on; }

    // Debug visualization of where the different levels are and how they are being blended.
    static ShowLevelMode GetShowLevelMode() { return s_showLevelMode; }
    // Get the mask of which layers are affected by the show level mode.
    static UINT32 GetShowLevelMask() { return s_showLevelMask; }
    // Set the visualization of where the different levels are and how they are being blended.
    // Mask sets which layers will be affected.
    static void SetShowLevelMode(ShowLevelMode m, const UINT mask=0xff);

    // Get the blend function start
    float GetDecayStart() { return m_fDecay; }
    // Get the blend function end.
    void SetDecayStart(float f) { m_fDecay = f; }

    // Set the jitter offsets used for each layer this frame. Assumes SetTextures() already called.
    void SetJitterOffsets(const XMFLOAT2 offs[kMaxLayers]);

private:
	struct ConstantBuffer
	{
		XMVECTOR		m_ScaleOffDst;

		XMVECTOR    	m_aScaleOffDomainToTex[kMaxLayers];

        XMVECTOR        m_aPixOffset[kMaxLayers];

		XMVECTOR        m_Decay;
	};

protected:

    // override for this 
	virtual HRESULT InitConstantBuffers(ID3D11Device* pd3dDevice);
	virtual void SetConstants(ID3D11DeviceContext* immediateContext, ID3D11ShaderResourceView* baseTex);
	virtual void UnSetConstants(ID3D11DeviceContext* immediateContext);

    virtual void AppendPixelDefines(std::vector<D3D_SHADER_MACRO>& defines);

private:
	ID3D11ShaderResourceView*	m_aTextures[kMaxLayers];
	BlitRect					m_aDstRects[kMaxLayers];
    XMFLOAT2                    m_aSrcSizes[kMaxLayers];

    XMFLOAT2                    m_aJitterOffsets[kMaxLayers];

    bool                        m_aMsaa[kMaxLayers];

    int                         m_iNumLayers;

    float		                m_fDecay;

    static ShowLevelMode        s_showLevelMode;	
    static UINT                 s_showLevelMask;
};

#endif // BLIT_COMPOSITE_H
