
#ifndef CASCADE_H
#define CASCADE_H

class OffscreenRT;

class Cascade
{
public:
    static void Init();

	Cascade();
	~Cascade();

	void Release();

	HRESULT Create(ID3D11Device* pd3dDevice, int width, int height, int msaa, float dup);
    HRESULT CreateBase(ID3D11Device* pd3dDevice, int width, int height, int msaa);

    // Must be called before Push to do any good.
    const XMFLOAT2& GetJitter() const { return m_Jitter; }
    void SetJitter(const XMFLOAT2& jit) { m_Jitter = jit; }

	XMMATRIX Push(ID3D11DeviceContext* pd3dImmediateContext, const POINT& gazePos, 
        const XMMATRIX& origCamToNDC, const FLOAT nearClip, const FLOAT farClip);

	void Clear(ID3D11DeviceContext* pd3dImmediateContext, const float* const color);

	void Pop(ID3D11DeviceContext* pd3dImmediateContext);

    const XMMATRIX& GetLastWorldToCamera() const { return m_mLastWorldToCamera; }
    const XMMATRIX& GetLastCameraToNDC() const { return m_mLastCameraToNDC; }
    XMMATRIX GetLastWorldToNDC() const { return m_mLastWorldToCamera * m_mLastCameraToNDC; }
    void SetLastWorldToCamera(const XMMATRIX& w2c) { m_mLastWorldToCamera = w2c; }
    void SetLastCameraToNDC(const XMMATRIX& c2ndc) { m_mLastCameraToNDC = c2ndc; }

    int GetWidth() const;
    int GetHeight() const;

	ID3D11ShaderResourceView* GetTexture() const;
	ID3D11ShaderResourceView* GetPreviousTex() const;
	ID3D11ShaderResourceView* GetPreviousDS() const;
    const POINT& GetGazePos() const { return m_LastGazePos; }

	float GetDuplication() const { return m_fDuplication; }

    void PingPong();

    HRESULT Save(const WCHAR* fileName, ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext);

    static bool GetJitterEnabled() { return m_bJitterEnabled; }
    static void SetJitterEnabled(bool b) { m_bJitterEnabled = b; }

    float GetJitterScale() const { return m_fJitterScale; }
    void SetJitterScale(float f) { m_fJitterScale = f; }

    bool ShouldRefresh(float dt);

    void SetRefreshInterval(float dt) { m_fRefreshInterval = dt; }
    float GetRefreshInterval() const { return m_fRefreshInterval; }

private:

	OffscreenRT*	m_aOffscreenRT[2];

	float           m_fDuplication;
    int             m_iPing;

    XMFLOAT2     m_Jitter;
    POINT           m_LastGazePos;

    float           m_fRefreshCountDown;
    float           m_fRefreshInterval;
private:

    XMMATRIX m_mLastWorldToCamera;
    XMMATRIX m_mLastCameraToNDC;

private:
    // Jitter section
    static bool m_bJitterEnabled;

    float m_fJitterScale;
    int m_iJitterIndex;

    XMFLOAT2 NextJitterSample();
};

#endif // CASCADE_H
