
#include "pch.h"

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>
#include <share.h>
#include <assert.h>

#include "CompiledShader.h"
#include "FileDateCheck.h"

#include "SpawnCmd.h"

static bool CompiledName(const WCHAR* fileName, const WCHAR* profile, WCHAR* compName)
{
	const int len = int(wcslen(fileName));

	int iDot = len;
	while(--iDot >= 0)
	{
		if(fileName[iDot] == L'.')
			break;
	}
	const bool error = iDot < 0;
	if(error)
	{
		iDot = len;
	}

	for(int i = 0; i < iDot; ++i)
	{
		compName[i] = fileName[i];
	}
	compName[iDot + 0] = L'.';
	compName[iDot + 1] = profile[0];
	compName[iDot + 2] = profile[1];
	compName[iDot + 3] = L'o';
	compName[iDot + 4] = 0;
	assert(wcscmp(fileName, compName));

	return error;
}

static void ConvertCharToWChar(WCHAR* dst, const CHAR* src)
{
    size_t chars = 0;
    mbstowcs_s(&chars, dst, MAX_PATH, src, _TRUNCATE);
}
static void ExtractDefine(WCHAR* define, const D3D_SHADER_MACRO& def)
{
    WCHAR name[MAX_PATH];
    ConvertCharToWChar(name, def.Name);
    WCHAR definition[MAX_PATH];
    ConvertCharToWChar(definition, def.Definition);

    swprintf_s(define, MAX_PATH, L"/D%s=%s ", name, definition);
}
bool CompiledShader::Compile(const WCHAR* fileName,
				const WCHAR* profile,
				const WCHAR* entryPoint)
{

    WCHAR fullInput[MAX_PATH];
    swprintf_s(fullInput, L"%s%s", GetShaderFolder(), fileName);

    WCHAR outFile[MAX_PATH];
	CompiledName(fileName, profile, outFile);
    WCHAR fullOutput[MAX_PATH];
    swprintf_s(fullOutput, L"%s%s", GetShaderFolder(), outFile);

	bool error = false;

//	FileDateCheck check(fileName, tempFile);

//	if(CompiledOutOfDate(fileName, tempFile))
	if(GetForced() || FileDateCheck(fullInput, fullOutput).OutOfDate())
	{
		WCHAR command[1024];

		WCHAR flags[MAX_PATH];
		flags[0] = 0;
		wcscpy_s(flags, L"/Ges");
#if defined(DEBUG) || defined(_DEBUG)
//		wcscat_s(flags, L" /Od /Zi");
#endif // DEBUG

//#define MAF_ALTERNATE_COMPILER
#ifdef MAF_ALTERNATE_COMPILER
		const WCHAR* fxc = L"fxc-deriv";
		//	const WCHAR* fxc = L"fxc-drew";
#else // MAF_ALTERNATE_COMPILER
		const WCHAR* fxc = L"fxc";
#endif // MAF_ALTERNATE_COMPILER

		swprintf_s(command, L"%s%s.exe /Fe%serrors.txt /Fo%s /T%s /E%s %s ",
            GetShaderFolder(),
			fxc,
            GetShaderFolder(),
			fullOutput,
			profile,
			entryPoint,
			flags);

        for(size_t i = 0; i < m_Defines.size(); ++i)
        {
            WCHAR define[MAX_PATH];
            ExtractDefine(define, m_Defines[i]);
            wcscat_s(command, define);
        }

        wcscat_s(command, fullInput);
		//swprintf_s(command, L"Shaders\\fxc /Fctemp.txt /T%s /E%s %s %s",
		//	profile,
		//	entryPoint,
		//	flags,
		//	fileName);

//		_wsystem(L"Shaders\\fxc-deriv /? > zzz.txt");
        static int erm = 0;
        if(0)
        {
            WCHAR dbgCmd[MAX_PATH];
            swprintf_s(dbgCmd, L"echo hello handsome > zzz.txt");
            erm = _wsystem(dbgCmd);
            swprintf_s(dbgCmd, L"%s%s /? >> zzz.txt", GetShaderFolder(), fxc);
//            erm = _wsystem(dbgCmd);
            swprintf_s(dbgCmd, L"media\\MAFShaders\\fxc /? >> zzz.txt");
            erm = _wsystem(dbgCmd);
        }
#define MAF_USE_SPAWN
#ifndef MAF_USE_SPAWN
		int retVal = _wsystem(command);
        error = retVal != 0;
#else // MAF_USE_SPAWN
        WCHAR cmd[1024];
        wsprintf(cmd, L"%s%s.exe", GetShaderFolder(), fxc);
        error = !SpawnCmd::Spawn(cmd, command);
#endif //  MAF_USE_SPAWN
	}

	/// Check the return code, will be zero on success. We might
	/// fail, but have an old temp.byt lying around.
	if(!error)
	{
		int fd;
		_wsopen_s( &fd, fullOutput, _O_RDONLY, _SH_DENYNO, 0 );

		if(fd >= 0)
		{
			this->byteCodeLength = _lseek(fd, 0, SEEK_END);

			this->byteCodes = new char[this->byteCodeLength];

			_lseek(fd, 0, SEEK_SET);

			_read(fd, this->byteCodes, unsigned int(this->byteCodeLength));

			_close(fd);

            return true;
		}
	}
    return false;
}

CompiledShader::~CompiledShader()
{
	Release();
}

void CompiledShader::SetShaderFolder( const WCHAR* const folder )
{
    m_Folder[0] = NULL;
    if(folder && *folder)
    {
        assert(wcschr(folder, '/') == NULL); // only back slashes supported. sorry. 
        const size_t last = wcslen(folder) - 1;
        const size_t len = last + 2;
        wcscpy_s( m_Folder, len, folder );
        if(m_Folder[last] != '\\')
        {
            wcscat_s(m_Folder, L"\\");
        }
    }
    else
    {
        wcscpy_s( m_Folder, L".\\");
    }
}

const WCHAR* CompiledShader::GetShaderFolder() const
{
    return m_Folder;
}

void CompiledShader::AddDefine( const D3D_SHADER_MACRO& def )
{
    m_Defines.push_back(def);
}

void CompiledShader::ClearDefines()
{
    m_Defines.clear();
}

void CompiledShader::AddDefines( const std::vector<D3D_SHADER_MACRO>& defs )
{
    for(size_t i = 0; i < defs.size(); ++i)
    {
        m_Defines.push_back(defs[i]);
    }
}
