
#ifndef COMPILE_SHADER_H
#define COMPILE_SHADER_H

#include <vector>

class CompiledShader
{
public:

    CompiledShader()
        :	byteCodeLength(0),
            byteCodes(NULL),
            m_Forced(false)
    {
        m_Folder[0] = 0;
    }

	virtual ~CompiledShader();

    // Return true on success
    bool Compile(const WCHAR* fileName,
        const WCHAR* profile,
        const WCHAR* entryPoint);

	SIZE_T GetBufferSize() const { return byteCodeLength; }
	void* GetBufferPointer() const { return byteCodes; }

    const WCHAR* GetShaderFolder() const;
    void SetShaderFolder(const WCHAR* const folder);

    void AddDefine(const D3D_SHADER_MACRO& def);
    void AddDefines(const std::vector<D3D_SHADER_MACRO>& defs);
    void ClearDefines();

    // Sets a one time force of recompile. Only affects the next 
    // call to Compile, skipping the file date check and recompiling
    // regardless.
    bool GetForced() const { return m_Forced; }
    void SetForced(bool b) { m_Forced = b; }

private:
    CompiledShader(const CompiledShader& eff)
        :	byteCodeLength(eff.byteCodeLength)
    {
        byteCodes = new char[byteCodeLength];
        memcpy_s(byteCodes, byteCodeLength, eff.byteCodes, eff.byteCodeLength);

        SetShaderFolder(eff.m_Folder);
    }

    void Release()
    {
        if(byteCodes)
        {
            delete [] byteCodes;
            byteCodes = NULL;
            byteCodeLength = 0;
        }
        SetShaderFolder(NULL);
        m_Defines.clear();
    }

    SIZE_T		byteCodeLength;
    void*		byteCodes;

    WCHAR                           m_Folder[MAX_PATH];
    std::vector<D3D_SHADER_MACRO>   m_Defines;

    bool        m_Forced;
};


#endif // COMPILE_SHADER_H
