
#ifndef DBG_KEY_PRESS_H
#define DBG_KEY_PRESS_H

#include <vector>

class DbgKeyMgr
{
public:

    static void Update();

    static void Register(class DbgKeyPress* key) { _dbgKeys.push_back(key); }
    static void Unregister(class DbgKeyPress* key) { _dbgKeys.erase(find(_dbgKeys.begin(), _dbgKeys.end(), key)); }

    static bool ShiftDown() { return m_ShiftDown; }
    static bool ControlDown() { return m_ControlDown; }
    static bool AltDown() { return m_AltDown; }
private:
    static std::vector<class DbgKeyPress*> _dbgKeys;

    static bool m_ShiftDown;
    static bool m_ControlDown;
    static bool m_AltDown;
};

class DbgKeyPress
{
public:
    DbgKeyPress(int key)
        :   m_Key(key),
        m_AlreadyDown(false),
        m_Pressed(false)
    {
        DbgKeyMgr::Register(this);
    }

    virtual ~DbgKeyPress()
    {
        DbgKeyMgr::Unregister(this);
    }

    bool IsDown() const
    {
        return m_Pressed;
    }

    bool OnDown() const
    {
        return m_Pressed && !m_AlreadyDown;
    }

    void Update()
    {
#ifndef _DURANGO
        const bool hit = (GetAsyncKeyState(m_Key) & 0x8000) != 0;
        if(hit)
        {
            m_AlreadyDown = m_Pressed;
            m_Pressed = true;
        }
        else
        {
            m_Pressed = m_AlreadyDown = false;
        }
#endif // _DURANGO
    }

private:
    int m_Key;
    bool m_AlreadyDown;
    bool m_Pressed;

};


#endif // DBG_KEY_PRESS_H
