#include "pch.h"

#include <windows.h>
#include <d3d11.h>

#include "Utility.h"
#include "OffscreenRT.h"
#include "wincodec.h"

OffscreenRT::OffscreenRT()
	: m_iWidth(0), m_iHeight(0),
		m_pRTV(NULL), m_pDSV(NULL),
        m_pTex(NULL), m_pRTResourceView(NULL),
        m_pDSTex(NULL), m_pDSResourceView(NULL)
{
}

OffscreenRT::~OffscreenRT()
{
	Release();
}

HRESULT OffscreenRT::Create(ID3D11Device* pd3dDevice, int w, int h, int msaa)
{
	HRESULT hr = S_OK;

    /// mafinch - this 128 is placeholder. Actual max msaa count is device dependent, must be queried.
    msaa = __min(128, __max(1, msaa));

	m_iWidth = w;
	m_iHeight = h;

	DXGI_SAMPLE_DESC sampleDesc;
	/// mafinch - for multisampled, change Count to some higher value supported by the current hardware, e.g. 8.
	sampleDesc.Count = msaa;
	sampleDesc.Quality = 0;
    D3D11_TEXTURE2D_DESC rtDesc = 
    {
        m_iWidth,//UINT Width;
        m_iHeight,//UINT Height;
        1,//UINT MipLevels;
        1,//UINT ArraySize;
        DXGI_FORMAT_R8G8B8A8_UNORM_SRGB,//DXGI_FORMAT Format;
        sampleDesc,//DXGI_SAMPLE_DESC SampleDesc;
        D3D11_USAGE_DEFAULT,//D3D11_USAGE Usage;
        D3D11_BIND_RENDER_TARGET|D3D11_BIND_SHADER_RESOURCE,//UINT BindFlags;
        0,//UINT CPUAccessFlags;
        0//UINT MiscFlags;    
    };

    V_RETURN( pd3dDevice->CreateTexture2D( &rtDesc, NULL, &m_pRT ) );

	V_RETURN( pd3dDevice->CreateRenderTargetView( m_pRT, NULL, &m_pRTV ) );

    if(msaa > 1)
    {
        DXGI_SAMPLE_DESC singleDesc;
        singleDesc.Count = 1;
        singleDesc.Quality = 0;

        D3D11_TEXTURE2D_DESC texDesc = 
        {
            m_iWidth,//UINT Width;
            m_iHeight,//UINT Height;
            1,//UINT MipLevels;
            1,//UINT ArraySize;
            DXGI_FORMAT_R8G8B8A8_UNORM_SRGB,//DXGI_FORMAT Format;
            singleDesc,//DXGI_SAMPLE_DESC SampleDesc;
            D3D11_USAGE_DEFAULT,//D3D11_USAGE Usage;
            D3D11_BIND_RENDER_TARGET|D3D11_BIND_SHADER_RESOURCE,//UINT BindFlags;
            0,//UINT CPUAccessFlags;
            0//UINT MiscFlags;    
        };
        V_RETURN( pd3dDevice->CreateTexture2D( &texDesc, NULL, &m_pTex ) );

        V_RETURN( pd3dDevice->CreateShaderResourceView( m_pTex, NULL, &m_pRTResourceView) );
    }
    else
    {
        V_RETURN( pd3dDevice->CreateShaderResourceView( m_pRT, NULL, &m_pRTResourceView) );
    }

	D3D11_TEXTURE2D_DESC dsDesc = 
	{
		m_iWidth,
		m_iHeight,
		1,
		1,
		DXGI_FORMAT_R32_TYPELESS,
		sampleDesc,
		D3D11_USAGE_DEFAULT,
		D3D11_BIND_DEPTH_STENCIL|D3D11_BIND_SHADER_RESOURCE,
		0,
		0
	};

	V_RETURN( pd3dDevice->CreateTexture2D( &dsDesc, NULL, &m_pDS ) );

    {
        const UINT texMipSlice = 0;
        const D3D11_DEPTH_STENCIL_VIEW_DESC dsvDesc =
        {
            DXGI_FORMAT_D32_FLOAT,
            msaa > 1 ? D3D11_DSV_DIMENSION_TEXTURE2DMS : D3D11_DSV_DIMENSION_TEXTURE2D,
            0,
            texMipSlice
        };

        V_RETURN( pd3dDevice->CreateDepthStencilView( m_pDS, &dsvDesc, &m_pDSV ) ); 
    }

#if 0
    if(false) // msaa > 1)
    {
        DXGI_SAMPLE_DESC singleDesc;
        singleDesc.Count = 1;
        singleDesc.Quality = 0;

        D3D11_TEXTURE2D_DESC texDesc = 
        {
            width,//UINT Width;
            height,//UINT Height;
            1,//UINT MipLevels;
            1,//UINT ArraySize;
            DXGI_FORMAT_R32_FLOAT,//DXGI_FORMAT Format;
            singleDesc,//DXGI_SAMPLE_DESC SampleDesc;
            D3D11_USAGE_DEFAULT,//D3D11_USAGE Usage;
            D3D11_BIND_RENDER_TARGET|D3D11_BIND_SHADER_RESOURCE,//UINT BindFlags;
            0,//UINT CPUAccessFlags;
            0//UINT MiscFlags;    
        };
        V_RETURN( pd3dDevice->CreateTexture2D( &texDesc, NULL, &pDSTex ) );

        V_RETURN( pd3dDevice->CreateShaderResourceView( pDSTex, NULL, &pDSResourceView) );
    }
    else
#endif
    {
        D3D11_SHADER_RESOURCE_VIEW_DESC dsResViewDesc;
        dsResViewDesc.Format = DXGI_FORMAT_R32_FLOAT;
        dsResViewDesc.ViewDimension = msaa > 1 ? D3D11_SRV_DIMENSION_TEXTURE2DMS : D3D11_SRV_DIMENSION_TEXTURE2D;
        dsResViewDesc.Texture2D.MostDetailedMip = 0;
        dsResViewDesc.Texture2D.MipLevels = 1;

        V_RETURN( pd3dDevice->CreateShaderResourceView( m_pDS, &dsResViewDesc, &m_pDSResourceView ) );
    }


	return S_OK;
}

void OffscreenRT::Release()
{
	SAFE_RELEASE(m_pRT);
	SAFE_RELEASE(m_pRTV);

	SAFE_RELEASE(m_pDS);
	SAFE_RELEASE(m_pDSV);

    SAFE_RELEASE(m_pDSTex);
    SAFE_RELEASE(m_pDSResourceView);

    SAFE_RELEASE(m_pTex);
	SAFE_RELEASE(m_pRTResourceView);
}

void OffscreenRT::Push(ID3D11DeviceContext* pd3dImmediateContext)
{
	m_SaveState.Save(pd3dImmediateContext);

	pd3dImmediateContext->OMSetRenderTargets(1, &m_pRTV, m_pDSV);

	UINT nViewPorts = 1;
	D3D11_VIEWPORT viewport;
	viewport.Width = (FLOAT)m_iWidth;
	viewport.Height = (FLOAT)m_iHeight;
	viewport.TopLeftX = 0;
	viewport.TopLeftY = 0;
	viewport.MaxDepth = 1.0f;
	viewport.MinDepth = 0.0f;
	pd3dImmediateContext->RSSetViewports( nViewPorts, &viewport );

}

void OffscreenRT::Clear(ID3D11DeviceContext* pd3dImmediateContext, const float * const color)
{
	if(m_pRTV && color)
		pd3dImmediateContext->ClearRenderTargetView( m_pRTV, color );
	
	if(m_pDSV)
		pd3dImmediateContext->ClearDepthStencilView( m_pDSV, D3D11_CLEAR_DEPTH, 1.0, 0 );
}

void OffscreenRT::Pop(ID3D11DeviceContext* pd3dImmediateContext)
{
	m_SaveState.Restore(pd3dImmediateContext);

    if(m_pTex)
    {
        D3D11_TEXTURE2D_DESC Desc;
        m_pTex->GetDesc( &Desc );
        pd3dImmediateContext->ResolveSubresource( m_pTex, D3D11CalcSubresource( 0, 0, 1 ), m_pRT,
                                                  D3D11CalcSubresource( 0, 0, 1 ), Desc.Format );
    }
    if(m_pDSTex)
    {
        D3D11_TEXTURE2D_DESC Desc;
        m_pDSTex->GetDesc( &Desc );
        pd3dImmediateContext->ResolveSubresource( m_pDSTex, D3D11CalcSubresource( 0, 0, 1 ), m_pDS,
                                                  D3D11CalcSubresource( 0, 0, 1 ), Desc.Format );
    }
}

#if 0
HRESULT OffscreenRT::Save(const WCHAR* fileName, ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext)
{
    HRESULT hr = D3DX11SaveTextureToFile(pd3dImmediateContext,
        m_pTex,
        D3DX11_IFF_BMP,
        fileName);

    return hr;

}
#else
#include "../DirectXTex/DirectXTex.h"
HRESULT OffscreenRT::Save(const WCHAR* fileName, ID3D11Device* pd3dDevice, ID3D11DeviceContext* pd3dImmediateContext)
{
    HRESULT hr = S_OK;

    ScratchImage scratch;
    V_RETURN( CaptureTexture(pd3dDevice, pd3dImmediateContext, m_pTex, scratch) );

    const Image* image = scratch.GetImages();
    if(image)
    {
        const DWORD flags = 0;
        V_RETURN( SaveToWICFile(*image, flags, GUID_ContainerFormatBmp, fileName) );
    }


    return hr;
}
#endif
