
#ifndef REPROJ_FX
#define REPROJ_FX

#include "ReprojRegisters.fx"

cbuffer cbReprojection : register( REG_REPROJECTION_CONST )
{
    float4      m_ParamLine0;   // Blend, Jitter Scale, Jitter Table Offset, unused.
    float4      m_WindowSize;   // Width, Height, 1.0 / Length, 0.5 / Length
};


Texture2D <float4> previousColor : register( REG_PREVIOUS_COLOR_TEX );
//Texture2DMS <float, 4> previousDepth : register( t3 );

SamplerState samplePrevious : register( REG_PREVIOUS_COLOR_SAMP );

Texture2D<float2> randomOffset : register( REG_RANDOM_TEX );
SamplerState sampleOffset : register( REG_RANDOM_SAMP );

float4 VS_Reproject(const float3 localPos, const matrix lastLocalToNDC)
{
	return mul( float4(localPos, 1), lastLocalToNDC );
}

float2 RandomTableIndex(const float2 coord)
{
    return coord * m_WindowSize.zz + m_WindowSize.w + m_ParamLine0.z;
}

static const float kGamma = 0.66f;

float4 PS_Reproject(const float4 interpPos, const float4 interpLastPos, const float4 currColor)
{

	float4 outputColor = currColor;
    outputColor.xyz = pow(outputColor.xyz, kGamma);
	const float4 lastPos = interpLastPos / interpLastPos.w;
	if(lastPos.x * lastPos.x < 1.0f && lastPos.y * lastPos.y < 1.0f)
	{
        const float2 colorCoord = lastPos.xy * float2(0.5f, -0.5f) + 0.5f;

		const float4 prevColor = previousColor.Sample(samplePrevious, colorCoord);
		outputColor.xyz = lerp(outputColor.xyz, prevColor.xyz, m_ParamLine0.x);
	}
	return outputColor;
}

float4 PS_Reproject(const float2 lastPos, const float4 currColor)
{
	float4 outputColor = currColor;
    outputColor.xyz = pow(outputColor.xyz, kGamma);
	if(lastPos.x * lastPos.x < 1.0f && lastPos.y * lastPos.y < 1.0f)
	{
        const float2 colorCoord = lastPos.xy * float2(0.5f, -0.5f) + 0.5f;
		const float4 prevColor = previousColor.Sample(samplePrevious, colorCoord);
		outputColor.xyz = lerp(outputColor.xyz, prevColor.xyz, m_ParamLine0.x);
    }
    return outputColor;
}

#endif // REPROJ_FX
