
#ifndef SAVE_RENDER_STATE_H
#define SAVE_RENDER_STATE_H

#include <assert.h>

class SaveRenderState
{
public:
	SaveRenderState()
		:	m_pOrigRTV(NULL),
			m_pOrigDSV(NULL),
			m_pRasterizerState(NULL),
			m_pDepthStencilStateStored11(NULL),
			m_nViewPorts(0),
			m_iStencilRef(0),
			m_bPushed(false)
	{
	}

	virtual ~SaveRenderState()
	{
	}

	void Save(ID3D11DeviceContext* pContext)
	{
		assert(!m_bPushed);

		// Save current render target and depthStencil
	    pContext->OMGetRenderTargets( 1, &m_pOrigRTV, &m_pOrigDSV );

		// Save the old viewport
		m_nViewPorts = D3D11_VIEWPORT_AND_SCISSORRECT_MAX_INDEX;
		pContext->RSGetViewports( &m_nViewPorts, m_aViewPortOld );

		// And depth stencil state.
		pContext->OMGetDepthStencilState( &m_pDepthStencilStateStored11, &m_iStencilRef );

        pContext->OMGetBlendState( &m_pBlendState, m_aBlendFactor, &m_uSampleMask );

		pContext->RSGetState( &m_pRasterizerState );

		m_bPushed = true;
	}
	void Restore(ID3D11DeviceContext* pContext)
	{
		assert(m_bPushed);

		pContext->RSSetState( m_pRasterizerState );

        pContext->OMSetBlendState( m_pBlendState, m_aBlendFactor, m_uSampleMask );

		pContext->OMSetDepthStencilState( m_pDepthStencilStateStored11, m_iStencilRef );

		// Restore the Old viewport
		pContext->RSSetViewports( m_nViewPorts, m_aViewPortOld );

		ID3D11RenderTargetView* aRTViews[ 1 ] = { m_pOrigRTV };
		pContext->OMSetRenderTargets( 1, aRTViews, m_pOrigDSV );

        SAFE_RELEASE(m_pBlendState);
		SAFE_RELEASE(m_pRasterizerState);
		SAFE_RELEASE(m_pDepthStencilStateStored11);
		SAFE_RELEASE(m_pOrigRTV);
		SAFE_RELEASE(m_pOrigDSV);

		m_bPushed = false;
	}

private:
	ID3D11RenderTargetView*		m_pOrigRTV;
	ID3D11DepthStencilView*		m_pOrigDSV;

	D3D11_VIEWPORT				m_aViewPortOld[D3D11_VIEWPORT_AND_SCISSORRECT_MAX_INDEX];
	UINT m_nViewPorts;

	ID3D11RasterizerState*      m_pRasterizerState;

	ID3D11DepthStencilState*	m_pDepthStencilStateStored11;
	UINT m_iStencilRef;

    ID3D11BlendState*           m_pBlendState;
    float                       m_aBlendFactor[4];
    UINT                        m_uSampleMask;

	bool						m_bPushed;
};

class SaveRenderStateAuto 
{
public:
    SaveRenderStateAuto(ID3D11DeviceContext* pContext)
        :   m_pContext(pContext)
    {
        m_SaveRenderState.Save(m_pContext);
    }

    ~SaveRenderStateAuto()
    {
        m_SaveRenderState.Restore(m_pContext);
    }

private:
    ID3D11DeviceContext*    m_pContext;
    SaveRenderState         m_SaveRenderState;
};

#endif // SAVE_RENDER_STATE_H
